<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Command;

use App\Entity\Core\ExportJob;
use App\Service\ConfigExportImport\ExportService;
use App\Entity\Core\File;
use App\Entity\Core\PgFile;
use App\Entity\Settings\SystemSettings;
use App\EventSubscriber\MqMessagebroadcastSubscriber;
use App\Service\SettingsHelper;
use DateTime;
use Doctrine\ORM\EntityManagerInterface;
use Exception;
use InvalidArgumentException;
use Ramsey\Uuid\Uuid;
use RecursiveDirectoryIterator;
use RecursiveIteratorIterator;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Filesystem\Exception\IOExceptionInterface;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use ZipArchive;

/**
 * Class ExportJobCommand
 */
class ExportJobCommand extends Command
{
    const COMPONENT = 'component';
    const ENTITY_TYPE = 'entityType';
    const HTTP_METHOD = 'operation';
    const REMOTE_ADDR = 'source';
    const DELTA_DATA = 'data';
    const TRANSACTION_DATE = 'transactionDate';
    const UUID = 'entityUuid';
    const USER_UUID = 'userId';
    const USER = 'user';

    /** @var Filesystem $filesystem */
    protected $filesystem;

    /** @var SettingsHelper */
    private $settingsHelper;

    private $entityManager;

    /** @var ExportService $exportService */
    private $exportService;

    /** @var ContainerInterface */
    private $container;

    public function __construct(
        SettingsHelper $settingsHelper,
        EntityManagerInterface $entityManager,
        ExportService $exportService,
        ContainerInterface $container,
        TokenStorageInterface $tokenStorage
    ) {
        $this->settingsHelper = $settingsHelper;
        $this->entityManager = $entityManager;
        $this->exportService = $exportService;
        $this->container = $container;
        $this->tokenStorage = $tokenStorage;
        parent::__construct();
    }

    protected function configure()
    {
        $this
            ->setName('fortisoar:jobs:export')
            ->setDescription('Exports records and configuration into a zip file')
            ->setDefinition([
                new InputOption(
                    'job-uuid',
                    'j',
                    InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY,
                    'The uuid of the export job that will be started.'
                ),
                new InputOption(
                    'user-token',
                    't',
                    InputOption::VALUE_OPTIONAL,
                    'The user token for the export context. If empty, the appliance user will be used.'
                ),
                new InputOption(
                    'token-type',
                    'tt',
                    InputOption::VALUE_OPTIONAL,
                    'Type of token HMAC or JWT.'
                )
            ]);

        $this->filesystem = new Filesystem();
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);

        $jobUuids = $input->getOption('job-uuid');
        if (count($jobUuids) == 0) {
            $io->error('Please specify a job uuid');
            return;
        }
        $jobUuid = $jobUuids[0];
        $rawToken = $input->getOption('user-token');
        $tokenType = $input->getOption('token-type');
        if ($rawToken && $tokenType) {
            $this->settingsHelper->setRefreshedToken($rawToken, $tokenType);
        }
        $repository = $this->entityManager->getRepository(ExportJob::class);
        /** @var ExportJob $exportJob */
        try {
            $exportJob =  $repository->findOneBy(['uuid' => $jobUuid]);
        } catch (InvalidArgumentException | \Exception $e) {
            $io->error($e->getMessage());
        }

        $status = 'In Progress';
        $this->updateExportJob($exportJob, $status, null, null);

        $filesystem = new Filesystem();
        $tmpDirectory = sys_get_temp_dir() . '/fsrexport-' . $jobUuid . '/';
        try {
            $filesystem->mkdir($tmpDirectory);
        } catch (IOExceptionInterface | \Exception $exception) {
            $message = sprintf('Error while creating temp directory: %s', $exception->getMessage());
            $this->updateExportJob($exportJob, 'Error', null, null, $message);
            $io->error($message);
        }

        try {
            $configExportData = $this->exportService->exportConfiguration($exportJob, $io, $tmpDirectory . 'data/', $tokenType);
        } catch (InvalidArgumentException | \Exception $exception) {
            $message = sprintf('Error while exporting file: %s', $exception->getMessage());
            $this->updateExportJob($exportJob, 'Error', null, null, $message);
            $io->error($message);
            $filesystem->remove($tmpDirectory);
        }
        if ($configExportData) {
            $metadata = $this->getExportMetadata();
            $metadataJson = json_encode($metadata, JSON_PRETTY_PRINT);
            $folderPath = $tmpDirectory . 'data/';
            $metadataFilePath = $folderPath . 'export.metadata.json';
            $filesystem->dumpFile($metadataFilePath, $metadataJson);
            $filename = preg_replace('/\.zip$/', '', $exportJob->getFileName());
            $zipFilename = $filename ? $filename . '.zip' : 'config.zip';
            $zipFilePath = $tmpDirectory . $zipFilename;
            try {
                $this->zipFolder($zipFilePath, $folderPath, $filename);
            } catch (InvalidArgumentException | \Exception $exception) {
                $message = sprintf('Error while building tar.gz file: %s', $exception->getMessage());
                $this->updateExportJob($exportJob, 'Error', null, null, $message);
                $io->error($message);
            }
            
            
            $pgFile = new PgFile();
            $fp = fopen($zipFilePath, 'rb');
            $fileSize = filesize($zipFilePath);
            $pgFile->setFile(base64_encode(fread($fp, $fileSize)));
            fclose($fp);
            $filesystem->remove($tmpDirectory);
            $pgFile->setUuid(Uuid::uuid4()->toString());
            $this->entityManager->persist($pgFile);
            $this->entityManager->flush();
            
            $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
            $mimetype = 'application/zip';
            $file = new File();
            $file->setFile("/api/3/pgfile/" . $pgFile->getUuid());
            $file->setFilename($zipFilename);
            $file->setMimeType($mimetype);
            $file->setSize($fileSize);
            $file->setUploadDate($timestamp);
            $file->setUuid(Uuid::uuid4()->toString());
            $this->entityManager->persist($file);
            $this->entityManager->flush();

            $status = 'Export Complete';
            $this->updateExportJob($exportJob, $status, null, "/api/3/files/" . $file->getUuid());

            $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
            $exportTemplate = $exportJob->getTemplate();
            $exportTemplate->setLastExportDate($timestamp);
            $this->entityManager->persist($exportTemplate);
            $this->entityManager->flush();

            $io->success(sprintf('Export completed successfully.'));
        }
    }

    private function getUserName($user)
    {
        if (method_exists($user, 'getFirstname')) {
            return $user->getFirstname() . ' ' . $user->getLastname();
        }
        return $user->getName();
    }

    private function getExportMetadata() {
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $version = $this->container->getParameter('product_version');
        $systemRepository = $this->entityManager->getRepository(SystemSettings::class);
        $systemSetting = $systemRepository->findAll();
        $global_value = $systemSetting[0]->getGlobalValues();
        $hostname = isset($global_value['hostname']) ? $global_value['hostname'] : gethostname();
        $token = $this->tokenStorage->getToken();
        $user = $token->getUser();
        $userName = $this->getUserName($user);
        
        $metadata = [
            'version' => $version,
            'date' => $timestamp->format(DateTime::ATOM),
            'exported_from' => $hostname ? $hostname : 'Unknown',
            'exported_by' => $userName
        ];
        return $metadata;
    }

    /**
     * @return array
     */
    protected function prepareChangeData(array $oldData, array $newData, string $entityId)
    {
        $deltaData = [];
        $deltaData['oldData'] = $oldData;
        $deltaData['newData'] = $newData;
        try {
            /** @var App\Entity\Actors\Actor $user */
            $user = $this->settingsHelper->getCurrentUser();
            if ($user) {
                $userUUID = $user->getUuid();
                $userName = $this->settingsHelper->extractUserName($user);
            } else {
                $userUUID = 'SYSTEM';
                $userName = 'SYSTEM';
            }
        } catch (InvalidArgumentException | Exception $e) {
            $userUUID = 'SYSTEM';
            $userName = 'SYSTEM';
        }
        $sourceIp = 'localhost';
        if (isset($_SERVER['HTTP_X_FORWARDED_FOR'])) {
            $sourceIp = $_SERVER['HTTP_X_FORWARDED_FOR'];
        } elseif (isset($_SERVER['REMOTE_ADDR'])) {
            $sourceIp = $_SERVER['REMOTE_ADDR'];
        }
        return [
            self::COMPONENT => 'crudhub',
            self::ENTITY_TYPE => 'ExportJob',
            self::UUID => $entityId,
            self::USER_UUID => $userUUID,
            self::USER => $userName,
            self::HTTP_METHOD => 'Update',
            self::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            self::DELTA_DATA => $deltaData,
            self::REMOTE_ADDR => $sourceIp
        ];
    }

    private function updateExportJob(ExportJob $exportJob, $status = null, $progressPercent = null, $file = "", $errorMessage = null)
    {
        $oldData = [
            '@id' => '/api/3/export_jobs/' . $exportJob->getUuid()
        ];
        $newData = [
            '@id' => '/api/3/export_jobs/' . $exportJob->getUuid()
        ];
        $oldData['id'] = $exportJob->getId();
        $newData['id'] = $exportJob->getId();
        if ($status !== null) {
            $oldStatus = $exportJob->getStatus();
            $exportJob->setStatus($status);
            $oldData['status'] = $oldStatus;
            $newData['status'] = $status;
        }
        if ($progressPercent !== null) {
            $oldProgressPercent = $exportJob->getProgressPercent();
            $exportJob->setProgressPercent($progressPercent);
            $oldData['progressPercent'] = $oldProgressPercent;
            $newData['progressPercent'] = $progressPercent;
        }
        if ($file !== "") {
            $oldFile = $exportJob->getFile();
            $exportJob->setFile($file);
            $oldData['file'] = $oldFile;
            $newData['file'] = $file;
        }
        if ($errorMessage !== null) {
            $oldErrorMessage = $exportJob->getErrorMessage();
            $exportJob->setErrorMessage($errorMessage);
            $oldData['errorMessage'] = $oldErrorMessage;
            $newData['errorMessage'] = $errorMessage;
        }
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $exportJob->setModifyDate($timestamp);
        $this->entityManager->persist($exportJob);
        $this->entityManager->flush();
        $body = $this->prepareChangeData($oldData, $newData, $exportJob->getUuid());
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode([$body]), MqMessagebroadcastSubscriber::DEFALUT_KEY);
    }

    private function zipFolder($zipFilePath, $folderPath, $filename) {
        $zip = new ZipArchive();
        $zip->open($zipFilePath, ZipArchive::CREATE | ZipArchive::OVERWRITE);
        /** @var SplFileInfo[] $files */
        $files = new RecursiveIteratorIterator(
            new RecursiveDirectoryIterator($folderPath),
            RecursiveIteratorIterator::LEAVES_ONLY
        );

        foreach ($files as $name => $file)
        {
            if (!$file->isDir())
            {
                $filePath = $file->getRealPath();
                $relativePath = $filename . '/' . substr($filePath, strlen($folderPath));
                $zip->addFile($filePath, $relativePath);
            }
        }
        $zip->close();
    }
}
