<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Controller;

use App\Entity\Query\Query;
use App\Providers\QueryProvider;
use App\Security\Voter\FsrVoter;
use Doctrine\Inflector\Inflector;
use App\Constants\AggregateOperators;
use Doctrine\Inflector\InflectorFactory;
use Symfony\Component\HttpFoundation\Request;
use App\Providers\PermissionsDictionaryBuilder;
use App\Providers\AggregateDataProviderInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use ApiPlatform\Core\Api\ResourceClassResolverInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use ApiPlatform\Core\DataProvider\CollectionDataProviderInterface;
use ApiPlatform\Core\Serializer\SerializerContextBuilderInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use ApiPlatform\Core\Metadata\Resource\Factory\ResourceMetadataFactoryInterface;
use App\Service\IriHelper;
use App\Service\MetadataHelper;
use App\Service\UtilityService;
use Doctrine\ORM\Mapping\ClassMetadata;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Contracts\Cache\TagAwareCacheInterface;

class AdvancedQueryController extends BaseController
{
    const DEFAULT_KEYS = ['id', '@id', '@type'];
    const SELECTIVE_FIELD_KEY = '__selectFields';
    protected $security = ["roles", "teams", "import_jobs", "export_templates", "proxies", "module_permissions", "field_permissions"];
    protected $application = [
        "tags", "system_settings", "generic_template", "system_view_templates", "picklists", "picklist_names", "modules",
        "attribute_metadatas", "staging_model_metadatas", "model_metadatas",
    ];
    protected $workflows = ["workflows", "rules", "workflow_collections", "rule_collections"];

    /** @var  QueryProvider */
    private $queryProvider;

    /** @var IriHelper */
    private $iriHelper;

    /** @var UtilityService */
    private $utilityService;

    private $collectionDataProvider;

    private $serializerContextBuilder;

    protected $inflector;

    protected $fsrVoter;

    protected $resourceMetadataFactory;
    
    protected $aggregateDataProvider;

    public function __construct(ResourceClassResolverInterface $resourceClassResolver, QueryProvider $queryProvider, 
        CollectionDataProviderInterface $collectionDataProvider, IriHelper $iriHelper, UtilityService $utilityService, SerializerContextBuilderInterface $serializerContextBuilder, 
        ContainerInterface $containers, FsrVoter $fsrVoter, ResourceMetadataFactoryInterface $resourceMetadataFactory, 
        AggregateDataProviderInterface $aggregateDataProvider, TagAwareCacheInterface $cache, MetadataHelper $metadataHelper)
    {
        $this->queryProvider = $queryProvider;
        $this->iriHelper = $iriHelper;
        $this->utilityService = $utilityService;
        $this->collectionDataProvider = $collectionDataProvider;
        $this->serializerContextBuilder = $serializerContextBuilder;
        $this->fsrVoter = $fsrVoter;
        $this->aggregateDataProvider = $aggregateDataProvider;
        $this->resourceMetadataFactory = $resourceMetadataFactory;
        $this->inflector = InflectorFactory::create()->build();
        parent::__construct($resourceClassResolver, $containers, $cache, $metadataHelper);
    }

    public function applyQuery(Request $request, $resourceShortName)
    {
        $perm = $this->utilityService->getModulePermission($resourceShortName);
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $perm)) {
            throw new AccessDeniedException();
        }
        $resource = $this->getResourceFromClassName($resourceShortName);
        $content = $request->getContent();
        $query = json_decode($content, true) ?? [];
        $data = $this->evaluateQuery($request, $query, $resource);
        return JsonResponse::create($data);
    }

    public function applyNamedQuery()
    {
    }

    protected function evaluateQuery($request, $query, $resource)
    {

        $this->queryProvider->setInitialQuery($query);
        $queryLimit = $request->get('$limit', 30);
        $queryPage = $request->get('$page', 1);
        $attributes = [];
        $attributes['resource_class'] = $resource;
        $attributes['collection_operation_name'] = "get";
        $attributes['filters'] = ['$limit' => $queryLimit, '$page' => $queryPage];
        $attributes['_fsr_dynamic_context'] = false;

        $context = $this->serializerContextBuilder->createFromRequest($request, true, $attributes);

        /** @var AggregateDataProviderInterface $aggregateDataProvider */
        $data = $this->isAggregateQuery($query) ?
            $this->aggregateDataProvider->getAggregate($resource, $request, $query) :
            $this->collectionDataProvider->getCollection($resource, $attributes['collection_operation_name'], $attributes);
        $serializer = $this->container->get('serializer');

        $context = $this->serializerContextBuilder->createFromRequest($request, true, $attributes);
        $data1 = $serializer->normalize($data, 'jsonld', $context);
        return $data1;
    }

    protected function isAggregateQuery($query)
    {
        $isAggregateQuery = false;

        if (array_key_exists(Query::KEYWORD_AGGREGATES, $query)) {
            foreach ($query[Query::KEYWORD_AGGREGATES] as $aggregate) {
                if (!in_array($aggregate['operator'], [AggregateOperators::SELECT, AggregateOperators::FIELDS])) {
                    $isAggregateQuery = true;
                    break;
                }
            }
        }

        return $isAggregateQuery;
    }

    public function queryAssociationAction(Request $request, $resourceShortName, $id, $field)
    {
        $resourceClass = $this->getResourceFromClassName($resourceShortName);
        $module = $this->metadataHelper->getModuleFromClassName($resourceClass);

        $this->fsrVoter->denyUnlessAccesssGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module);
        $item = $this->iriHelper->findOrThrowNotFound('/api/3/' . $resourceShortName . '/' . $id);

        $this->utilityService->ValidateOwnerShip($request, $item);

        // Check Field Level Permissions
        $this->fsrVoter->denyUnlessAccesssGranted(PermissionsDictionaryBuilder::PERM_FIELD_READ . '.' . $module . '.' . $field);

        /** @var ManagerRegistry $managerRegistry */
        $managerRegistry = $this->container->get('doctrine');
        $manager = $managerRegistry->getManagerForClass($resourceClass);
        $classMetadata = $manager->getClassMetadata($resourceClass);
        if ($classMetadata->hasAssociation($field)) {
            /** @var ClassMetadata $classMetadata */
            $association = $classMetadata->associationMappings[$field];
            if (in_array($association['type'], [ClassMetadata::MANY_TO_MANY, ClassMetadata::ONE_TO_MANY, ClassMetadata::TO_MANY])) {
                if ($association['isOwningSide']) {
                    $inversedField = $association['inversedBy'];
                } else {
                    $inversedField = $association['mappedBy'];
                }
                $targetEntity = $association['targetEntity'];
                $associationResource = $targetEntity;
                $associationModule = $this->metadataHelper->getModuleFromClassName($associationResource);
                $this->fsrVoter->denyUnlessAccesssGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $associationModule);

                $associationFilter = [
                    'field' => $inversedField,
                    'operator' => 'eq',
                    'value' => $item->getUuid()
                ];
                if ($this->containers->hasParameter('ignore_rbac_for_related_record') && $this->containers->getParameter('ignore_rbac_for_related_record')) {
                    $GLOBALS['__associatedFieldRequest__'] = true;
                }
                $request->overrideGlobals();
                $resource = $this->getResourceFromClassName($associationModule);
                $content = $request->getContent();
                $associationQuery = json_decode($content, true) ?? [];
                $query = $associationQuery;
                $associationQueryLogic = $query['logic'];
                $associationQueryFilters = $query['filters'];
                $query['logic'] = 'AND';
                $query['filters'] = [
                    $associationFilter, 
                    [
                        'logic' => $associationQueryLogic,
                        'filters' => $associationQueryFilters
                    ]
                ];
                $data = $this->evaluateQuery($request, $query, $resource);
                return JsonResponse::create($data);
            }
        }
        throw new BadRequestHttpException("There is no field with the given name in the resource class");
    }
}
