<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Controller;

use App\Service\MetadataHelper;
use App\Security\Voter\FsrVoter;
use App\Service\BulkRequestService;
use Symfony\Component\HttpFoundation\Request;
use ApiPlatform\Core\Api\IriConverterInterface;
use App\Providers\PermissionsDictionaryBuilder;
use ApiPlatform\Core\Exception\ExceptionInterface;
use Symfony\Component\Serializer\SerializerInterface;
use ApiPlatform\Core\Api\ResourceClassResolverInterface;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Config\Definition\Exception\Exception;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Contracts\Cache\TagAwareCacheInterface;

class BulkCreateController extends BaseController
{
    /** @var  FsrVoter */
    private $fsrVoter;

    /** @var  EntityManagerInterface */
    protected $entityManager;

    protected $bulkRequestService;

    protected $metadataHelper;

    const OPTION_KEY = 'options';
    const CLONE_KEY = 'cloneField';
    const RELATIONSHIP_KEY = 'relationships';

    public function __construct(FsrVoter $fsrVoter, ResourceClassResolverInterface $resourceClassResolver,
        ContainerInterface $containers, IriConverterInterface $iriConverter,
        EntityManagerInterface $entityManager,
        SerializerInterface $serializer, BulkRequestService $bulkRequestService,
        MetadataHelper $metadataHelper, TagAwareCacheInterface $cache) {
        $this->fsrVoter = $fsrVoter;
        $this->iriConverter = $iriConverter;
        $this->entityManager = $entityManager;
        $this->serializer = $serializer;
        $this->bulkRequestService = $bulkRequestService;
        parent::__construct($resourceClassResolver, $containers, $cache, $metadataHelper);
    }

    public function bulkCreate(Request $request, $moduletype)
    {
        $resourceClass = $this->getResourceFromClassName($moduletype);
        $module = $this->metadataHelper->getModuleFromClassName($resourceClass);
        if (!($this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.' . $module))) {
            throw new AccessDeniedException();
        }

        // TODO: validate if payload is an array or single object
        $objectData = json_decode($request->getContent());
        $headers = $request->headers->all();
        $queryParams = $request->query->all();
        if (array_key_exists(self::SELECTIVE_FIELDS, $objectData)) {
            $selectedFields = $objectData->{self::SELECTIVE_FIELDS};
            $queryParams[self::SELECTIVE_FIELDS] = implode(',', $selectedFields);
        }
        if (array_key_exists("cloneIds", $objectData)) {
            $objectDataJson = json_decode($request->getContent(), true);
            $processedResponse = $this->prepareClonePayload($objectDataJson, $moduletype, $resourceClass);
            $objectsToClone = $processedResponse[0];
            $failedCloneObjects = $processedResponse[1];
            if (sizeof($failedCloneObjects) > 0) {
                array_push($failedObj, $failedCloneObjects);
            }
            $objectData->{'data'} = $objectsToClone;
        }
        return $this->bulkRequestService->prepareRecordSubRequest($resourceClass, $objectData, '/api/3/' . $moduletype, $headers, $queryParams);
    }

    private function prepareClonePayload(array $objectDataJson, string $moduletype, string $resourceClass): array
    {
        $data = [];
        $failedObj = [];
        if (array_key_exists(self::OPTION_KEY, $objectDataJson) &&
            array_key_exists(self::CLONE_KEY, $objectDataJson[self::OPTION_KEY]) &&
            ($objectDataJson[self::OPTION_KEY][self::CLONE_KEY] !== null)) {
            $cloneFieldName = $objectDataJson[self::OPTION_KEY][self::CLONE_KEY];
        }
        foreach ($objectDataJson['cloneIds'] as $cId) {
            if (isset($cId)) {
                $explodeId = explode('/', $cId);
                $id = end($explodeId);
                // check if resourceClass matches the one in iri
                $resourceInIRI = prev($explodeId);
                if (!($resourceInIRI == $moduletype)) {
                    throw new BadRequestHttpException('The resource in the record IRIs does not match the resource in API route');
                }
                $object = null;
                try {
                    $resourceContext = ['resource_class' => $resourceClass];
                    $object = $this->iriConverter->getItemFromIri($cId, $resourceContext + ['fetch_data' => true]);
                    if (!$object) {
                        array_push($failedObj, $id);
                    } else {
                        if (array_key_exists(self::OPTION_KEY, $objectDataJson) && array_key_exists(self::RELATIONSHIP_KEY, $objectDataJson[self::OPTION_KEY]) && $objectDataJson[self::OPTION_KEY][self::RELATIONSHIP_KEY]) {
                            $resourceContext["relationships"] = true;
                        }
                        $normalized_data = $this->serializer->normalize($object, self::NORMALIZER_METHOD, $resourceContext);
                        $cloneFieldName = $cloneFieldName && $normalized_data[$cloneFieldName] ? $cloneFieldName : 'name';
                        $repository = $this->entityManager->getRepository($resourceClass);
                        $newValue = $normalized_data[$cloneFieldName] . ' (1)';
                        $number = 1;
                        while ($repository->findOneBy([$cloneFieldName => $newValue]) != null) {
                            $newValue = $normalized_data[$cloneFieldName] . ' (' . ++$number . ')';
                        }
                        if (array_key_exists($cloneFieldName, $normalized_data) && $normalized_data[$cloneFieldName]) {
                            $normalized_data[$cloneFieldName] = $newValue;
                        }
                        unset($normalized_data['id']);
                        unset($normalized_data['uuid']);
                        if (isset($normalized_data['tenantRecordId'])) {
                            unset($normalized_data['tenantRecordId']);
                        }
                        unset($normalized_data['@id']);
                    }
                    $data[] = $normalized_data;
                } catch (Exception | ExceptionInterface $exception) {
                    if ($exception instanceof ExceptionInterface) {
                        $err = $exception->__toString();
                    } else {
                        $err = [$exception->getMessage()];
                    }
                    array_push($failedObj, $err);
                    continue;
                }
            } else {
                array_push($failedObj, $cId);
            }
        }
        return array($data, $failedObj);
    }
}
