<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Controller;

use GuzzleHttp\Client;
use App\Security\Voter\FsrVoter;
use App\Service\SamlActorService;
use App\Entity\Authorization\Team;
use Doctrine\ORM\EntityManagerInterface;
use App\Service\AccessibleTeamIdsProvider;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use ApiPlatform\Core\Api\IriConverterInterface;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class GatewayActionController extends AbstractController
{

    protected $samlActorService;

    protected $entityManager;

    protected $tokenStorage;

    protected $fsrVoter;

    protected $iriConverter;

    protected $accessibleTeamIdsProvider;

    public function __construct(
        SamlActorService $samlActorService,
        TokenStorageInterface $tokenStorage,
        FsrVoter $fsrVoter,
        IriConverterInterface $iriConverter,
        AccessibleTeamIdsProvider $accessibleTeamIdsProvider,
        EntityManagerInterface $entityManager
    ) {
        $this->samlActorService = $samlActorService;
        $this->tokenStorage = $tokenStorage;
        $this->fsrVoter = $fsrVoter;
        $this->iriConverter = $iriConverter;
        $this->accessibleTeamIdsProvider = $accessibleTeamIdsProvider;
        $this->entityManager = $entityManager;
    }

    const USER_UUID = 'userId';
    const USER = 'user';
    const REMOTE_ADDR = 'source';

    public function sendRequest($client, $gatewayRequest)
    {
        try {
            $response = $client->send($gatewayRequest);
            $statusCode = $response->getStatusCode();
            if ($statusCode >= 400) {
                $errMsg = "Failed to get response from audit log";
                return JsonResponse::create($errMsg, $statusCode);
            } else {
                $jsonDecodeResponse = json_decode($response->getBody()->getContents(), true);
                return JsonResponse::create($jsonDecodeResponse, $statusCode);
            }
        } catch (BadRequestHttpException | \Exception $exception) {
            $errVar =$exception->getMessage();
            error_log($errVar);
            $errMsg['message'] = $errVar['Error'] ? isset($errVar['Error']) : $errVar['error'];
            $errMsg = $errMsg['message'] ? $errMsg : $errVar;
            return JsonResponse::create($errMsg, 400);
        }
    }

    public function gatewayRouteAction(Request $request)
    {
        $objectData = json_decode($request->getContent(), true);
        $objectData['isAdmin'] = false;
        if (!isset($objectData['entityUuid'])) {
            $this->checkAuthorization($request->getMethod(), $objectData);
        }
        if ($this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . 'security')) {
            $objectData['isAdmin'] = true;
        }
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities';
        $token = $this->tokenStorage->getToken();
        $objectData['teams'] = $this->getAccessibleTeamIris();
        $objectData['current_user'] = $this->iriConverter->getIriFromItem($this->getUser());
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function checkAuthorization($method, $objectData = [])
    {
        if (isset($objectData['userId']) && $objectData['userId'] == $this->getUser()->getUuid()) {
            if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . 'activities')) {
                throw new AccessDeniedException();
            }
            return;
        }
        if ($method === 'POST' && !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . 'security')) {
            throw new AccessDeniedException();
        }
    }

    /**
     * @return array
     */
    protected function getAccessibleTeamIris()
    {
        $teamResource = Team::class;
        $teamBaseIri = $this->iriConverter->getIriFromResourceClass($teamResource);
        $accessibleTeamIds = $this->accessibleTeamIdsProvider->getDataForCurrentUser();
        $accessibleTeamIris = array_map(function ($teamId) use ($teamBaseIri) {
            return sprintf('%s/%s', $teamBaseIri, $teamId);
        }, $accessibleTeamIds);
        return json_decode(json_encode($accessibleTeamIris), true);
    }

    public function gatewayRouteIdAction(Request $request, $auditId)
    {

        $this->checkAuthorization($request->getMethod());
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/';
        if (isset($auditId) && $auditId != null) {
            $gatewayUri = $gatewayUri . $auditId;
        } else {
            $errMsg['message'] = "Invalid ID";
            return JsonResponse::create($errMsg, 404);
        }
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('get', $gatewayUri, $headers, $request->getContent());
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayDeleteRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities?';
        $query = $request->query->all();
        foreach ($query as $key => $value) {
            $gatewayUri = $gatewayUri . $key . '=' . $value . '&';
        }
        $gatewayUri = substr($gatewayUri, 0, -1);
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('delete', $gatewayUri, $headers);
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function checkAuditDeleteAuthorization($method)
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_DELETE . '.' . 'activities')) {
            throw new AccessDeniedException();
        }
    }

    public function gatewayDeleteActivitiesRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/delete';
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'];
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayGetRouteAction(Request $request)
    {

        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/ttl';
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('get', $gatewayUri, $headers);
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayGetAuditOperations(Request $request)
    {

        $this->fsrVoter->denyUnlessAccesssGranted('read.activities');
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/operations?' . $request->getQueryString();
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('get', $gatewayUri, $headers);
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayPostRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $objectData = json_decode($request->getContent(), true);
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/ttl';
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'];
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewayttlDeleteRouteAction(Request $request)
    {
        $this->checkAuthorization($request->getMethod());
        $this->checkAuditDeleteAuthorization($request->getMethod());
        $objectData = json_decode($request->getContent(), true);
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/audit/activities/ttl';
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'];
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('delete', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }
    /**
     * @return string
     */
    private function extractUserName()
    {
        if (method_exists($this->getUser(), "getfirstname")) {
            if (method_exists($this->getUser(), "getlastname")) {
                $userName = $this->getUser()->getfirstname() . ' ' . $this->getUser()->getlastname();
            } else {
                $userName = $this->getUser()->getfirstname();
            }
        } else {
            $userName = $this->getUser()->getname();
        }
        return $userName;
    }

    public function gatewaySyslogGetAction(Request $request)
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . 'security')) {
            throw new AccessDeniedException();
        }
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog';
        $token = $this->tokenStorage->getToken();
        $headers = array(
            "Content-type" => "application/json",
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('get', $gatewayUri, $headers);
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewaySyslogPostAction(Request $request)
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.' . 'security')) {
            throw new AccessDeniedException();
        }

        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'];
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog';
        $token = $this->tokenStorage->getToken();
        $originalHeader = $request->headers->all();
        $headers = array(
            "Content-type" => "application/json",
            "is-cluster-call" => $originalHeader["is-cluster-call"],
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('post', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewaySyslogPutAction(Request $request)
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.' . 'security')) {
            throw new AccessDeniedException();
        }
        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'];
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog';
        $token = $this->tokenStorage->getToken();
        $originalHeader = $request->headers->all();
        $headers = array(
            "Content-type" => "application/json",
            "is-cluster-call" => $originalHeader["is-cluster-call"],
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('put', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }

    public function gatewaySyslogDeleteAction(Request $request, $uuid)
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_DELETE . '.' . 'security')) {
            throw new AccessDeniedException();
        }

        $objectData = json_decode($request->getContent(), true);
        $objectData[self::USER_UUID] = $this->getUser()->getUuid();
        $objectData[self::USER] = $this->extractUserName();
        $objectData[self::REMOTE_ADDR] = $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'];
        $gatewayUri = $this->getParameter('gateway.uri');
        $gatewayUri = $gatewayUri . '/config/syslog/';
        if (isset($uuid) && $uuid != null) {
            $gatewayUri = $gatewayUri . $uuid;
        } else {
            $errMsg['message'] = "Invalid UUId";
            return JsonResponse::create($errMsg, 404);
        }
        $token = $this->tokenStorage->getToken();
        $originalHeader = $request->headers->all();
        $headers = array(
            "Content-type" => "application/json",
            "is-cluster-call" => $originalHeader["is-cluster-call"],
        );
        $headers['Authorization'] = "Bearer " . $token->getCredentials();
        $gatewayRequest = new GuzzleRequest('delete', $gatewayUri, $headers, json_encode($objectData));
        $client = new Client();
        return $this->sendRequest($client, $gatewayRequest);
    }
}
