<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Controller;

use stdClass;
use App\Utility\Resolver;
use App\Entity\Custom\Person;
use App\Security\Voter\FsrVoter;
use App\Encoder\CorrectSimpleJWS;
use App\Service\SamlActorService;
use App\Entity\Settings\SystemSettings;
use App\Providers\JwtLogoutKeyProvider;
use App\Service\AccessibleTeamIdsProvider;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use GuzzleHttp\Exception\BadResponseException;
use Symfony\Component\HttpFoundation\Response;
use App\Encoder\Restricted\RestrictedUserToken;
use App\Providers\PermissionsDictionaryBuilder;
use GuzzleHttp\Psr7\Response as GuzzleResponse;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Security\Core\Authorization\Voter\VoterInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class ProxyController extends AbstractController
{

    protected $containers;

    protected $samlActorService;

    protected $fsrVoter;

    protected $tokenStorage;

    protected $jwtLogoutProvider;

    private $accessibleTeamIdsProvider;

    public function __construct(
        AccessibleTeamIdsProvider $accessibleTeamIdsProvider,
        ContainerInterface $containers,
        SamlActorService $samlActorService,
        FsrVoter $fsrVoter,
        TokenStorageInterface $tokenStorage,
        JwtLogoutKeyProvider $jwtLogoutProvider
    ) {
        $this->accessibleTeamIdsProvider = $accessibleTeamIdsProvider;
        $this->containers = $containers;
        $this->samlActorService = $samlActorService;
        $this->fsrVoter = $fsrVoter;
        $this->tokenStorage = $tokenStorage;
        $this->jwtLogoutProvider = $jwtLogoutProvider;
    }

    const QUERY_STRING = 'query_string';
    const URL = 'url';
    const BODY = 'body';
    const LOGIN = 'login';
    const LOGOUT = 'logout';
    const SAML = 'saml';
    const USERID = 'userId';
    const SAML_FAILURE_MESSAGE = 'We are sorry, but the server encountered an error while handling your request. Please contact your administrator for assistance.';

    public function proxyRouteAction($routeIdentifier, $route, Request $request, $customHeader = [], $payload = null)
    {
        $token = $this->tokenStorage->getToken();

        $client = $this->findHandlerForRoute($routeIdentifier);
        $permissionGroups = $this->findPermissionGroupsForRoute($routeIdentifier);

        $matchedPermissionGroups = 0;

        if (get_class($this->tokenStorage->getToken()) != RestrictedUserToken::class) {
            foreach ($permissionGroups as $permissionName => $permissionGroup) {
                /** @var VoterInterface[] $additionalVoters */
                $additionalVoters = [];
                $permissionsRequired = [];
                // Every User should be able to retrieve the config token section so that UI can handle it accordingly.
                // $ignoreAuthorization = false;
                $operator = $permissionGroup['operator'] ?? null;

                //Check method first
                if ($request->getMethod() != $permissionGroup['action']) {
                    continue;
                }

                $uuid = null;

                //This was matched because it had no permissions required
                if (empty($permissionGroup['permissions'])) {
                    $matchedPermissionGroups++;
                }

                //Check permissions
                foreach ($permissionGroup['permissions'] as $permission) {
                    //Check incoming route
                    $routeRegex = sprintf('/%s/', str_replace('/', '\/', $permissionGroup['route']));
                    preg_match($routeRegex, $route, $matches);

                    // $ignoreAuthorization = false;
                    // if (array_key_exists('ignoreAuthorization', $permissionGroup)) {
                    //     $ignoreAuthorization = $permissionGroup['ignoreAuthorization'];
                    // }

                    if (empty($matches)) {
                        continue;
                    }

                    //This was matched because it matched the regex
                    $matchedPermissionGroups++;

                    // if ($ignoreAuthorization) {
                    //     continue;
                    // }

                    //Check query strings to make sure it matches
                    if (isset($permissionGroup['params'])) {
                        foreach ($permissionGroup['params'] as $key => $value) {
                            if ($request->query->get($key) != $value) {
                                continue 2; //Continue to next permission group
                            }
                        }
                    }

                    //Build Permissions
                    $permissionsRequired[] = $permission;

                    //Determine where uuid is
                    if (!array_key_exists('uuidLocation', $permissionGroup)) {
                        continue;
                    }
                    $uuidLocation = $permissionGroup['uuidLocation'];

                    // Check if in the URL
                    if (isset($uuidLocation[self::URL])) {
                        $matchNumber = $uuidLocation[self::URL]['matchNumber'];
                        if (count($matches) >= $matchNumber) {
                            $uuid = $matches[$matchNumber];
                        }
                    }

                    // Check if in the BODY
                    if (isset($uuidLocation[self::BODY])) {
                        $path = $uuidLocation[self::BODY]['path'];
                        $bodyAsArray = json_decode($request->getContent(), true);
                        $uuid = Resolver::get($bodyAsArray, $path);
                    }

                    // Check if in the Query String
                    if (isset($uuidLocation[self::QUERY_STRING])) {
                        $queryKey = $uuidLocation[self::QUERY_STRING]['key'];
                        $uuid = $request->query->get($queryKey);
                    }
                    //Build additional voters
                    $additionalVoters[] = $permissionGroup['additionalVoters'];
                }

                //Keep checking for more matched permission groups
                if ($matchedPermissionGroups <= 0) {
                    continue;
                }

                // Commenting for Possible Remebering if regression reported
                // if (array_key_exists('ignoreAuthorization', $permissionGroup)) {
                //     $ignoreAuthorization = $permissionGroup['ignoreAuthorization'];
                // }
                // if ($ignoreAuthorization) {
                //     break;
                // }

                if ($operator && $operator == 'AND') {
                    $permissionMatchCount = 0;
                    foreach ($permissionsRequired as $perm) {
                        if ($this->fsrVoter->isGranted($perm)) {
                            $permissionMatchCount++;
                        }
                    }
                    if ($permissionMatchCount < count($permissionsRequired)) {
                        throw new AccessDeniedException();
                    } else {
                        break;
                    }
                } else {
                    $matched = false;
                    foreach ($permissionsRequired as $perm) {
                        if ($this->fsrVoter->isGranted($perm)) {
                            $matched = true;
                            break;
                        }
                    }
                    if (!empty($additionalVoters)) {
                        if ($token->getUsername() == $uuid) {
                            break;
                        }
                    }
                    if (!$matched) {
                        throw new AccessDeniedException();
                    } else {
                        break;
                    }
                }
            }

            if ($matchedPermissionGroups == 0) {
                throw new AccessDeniedException();
            }
        }

        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), $route, $request->getQueryString());

        if ($payload) {
            $content = $payload;
        } else {
            $content = $request->getContent();
        }

        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        unset($headersExceptForHost['cookie']);
        $headersExceptForHost['X-REMOTE_ADDR'] = isset($_SERVER['HTTP_X_FORWARDED_FOR']) ? $_SERVER['HTTP_X_FORWARDED_FOR'] : $_SERVER['REMOTE_ADDR'];
        if (get_class($this->tokenStorage->getToken()) != RestrictedUserToken::class) {
            $headersExceptForHost['X-USER'] = $this->getUser()->getUuid();
        } else {
            $headersExceptForHost['X-USER-EMAIL'] = $this->tokenStorage->getToken()->getEmail();
        }
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, array_merge($headersExceptForHost, $customHeader), $content);

        /** @var GuzzleResponse $response */
        try {
            $response = $client->send($hmacRequest);
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse();
        } catch (\Exception $e) {
            throw new \Exception('Failure: ' . $e->getMessage());
        }
        return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
    }

    public function wfProxyRouteAction($routeIdentifier, $route, Request $request)
    {
        $teams = ["/api/3/teams/doesnotexist"];
        if (get_class($this->tokenStorage->getToken()) != RestrictedUserToken::class) {
            $teams = $this->getAccessibleTeamIris();
        }
        if (strpos($route, "api/workflows/log_list") !== false && $request->getMethod() == "POST") {
            $payload = json_decode($request->getContent(), true);
            $payload['PB_OWNERS'] = $teams;
            return $this->proxyRouteAction($routeIdentifier, $route, $request, [], json_encode($payload));
        } else if (strpos($route, "api/workflows/") !== false && $request->getMethod() == "GET") {
            $Customheader['X_PB_OWNERS'] = $teams;
            return $this->proxyRouteAction($routeIdentifier, $route, $request, $Customheader);
        }
        return $this->proxyRouteAction($routeIdentifier, $route, $request, []);
    }

    public function dasRouteAction($routeIdentifier, $route, Request $request)
    {
        $is_admin = $this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.security');
        $customHeader = array(
            "X-ADMIN-USER" => $is_admin
        );
        return $this->proxyRouteAction($routeIdentifier, $route, $request, $customHeader);
    }

    public function getLoggedInUserRoles()
    {
        $user = $this->getUser();
        $roles_uuid = [];
        foreach ($user->getRoles() as $role) {
            $roles_uuid[] = $role->getUuid();
        }
        return $roles_uuid;
    }

    public function integrationProxyRouteAction($routeIdentifier, $route, Request $request)
    {
        if ($request->getMethod() == "POST" || $request->getMethod() == "PUT" || $request->getMethod() == "DELETE") {
            if (strpos($route, "import-connector") === false) {
                $teams = $this->getAccessibleTeamIris();
                $roles = $this->getLoggedInUserRoles();
                $query = $request->getContent() ? json_decode($request->getContent()) : new stdClass();
                $query->rbac_info = new stdClass();
                $query->rbac_info->{'teams'} = $teams;
                $query->rbac_info->{'roles'} = $roles;
                $query->rbac_info->{'source'} = isset($_SERVER['HTTP_X_FORWARDED_FOR']) ? $_SERVER['HTTP_X_FORWARDED_FOR'] : $_SERVER['REMOTE_ADDR'];
                $query->rbac_info->{'user_iri'} = '/api/3/people/' . $this->getUser()->getUuid();
                return $this->proxyRouteAction($routeIdentifier, $route, $request, [], json_encode($query));
            }
        }
        return $this->proxyRouteAction($routeIdentifier, $route, $request);
    }

    public function postmanProxyRouteAction($routeIdentifier, $route, Request $request)
    {
        if ((strpos($route, "staging-mmd/") !== false && $request->getMethod() == "PUT") || (strpos($route, "publish/") !== false && $request->getMethod() == "POST")) {
            $customheader['X-USER'] = $this->getUser()->getUuid();
            return $this->proxyRouteAction($routeIdentifier, $route, $request, $customheader);
        }
        return $this->proxyRouteAction($routeIdentifier, $route, $request);
    }

    /**
     * @return array
     */
    protected function getAccessibleTeamIris()
    {
        $accessibleTeamIds = $this->accessibleTeamIdsProvider->getDataForCurrentUser();
        return json_encode($accessibleTeamIds);
    }

    public function routePublicSAMLAction($routeIdentifier, $route, Request $request)
    {
        if ($route !== self::LOGIN && $route !== self::LOGOUT) {
            return new Response(null, Response::HTTP_NOT_FOUND);
        }
        // Handle POST request for SOCaaS
        $request_body_str = $request->getContent();
        if(strpos($request_body_str, 'h_key') !== FALSE) {
            $host = $_SERVER['HTTP_HOST'];
            $html = "<html><body onload=\"javascript:window.location.href = 'https://$host/login/'\"></body></html>";
            return Response::create($html, 200, ['Content-Type' => 'text/html']);
        }
        $client = $this->findHandlerForRoute($routeIdentifier);
        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), $route, $request->getQueryString());
        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, $headersExceptForHost, $request->getContent());

        /** @var GuzzleResponse $response */
        try {
            $response = $client->send($hmacRequest, array(self::SAML => true));
            if ($response->getStatusCode() === 200 && $request->getMethod() === 'POST') {
                $html = $this->validateSAMLUser($response, $route, $headersExceptForHost);
                return Response::create($html, $response->getStatusCode(), ['Content-Type' => 'text/html']);
            }
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse();
            $respBody = json_decode($response->getBody()->getContents(), true);
            $error = $respBody['Error'] ??  $respBody['message'] ?? self::SAML_FAILURE_MESSAGE;
            return Response::create($this->getErrorHTML($error), $response->getStatusCode(), ['Content-Type' => 'text/html']);
        } catch (\Exception $e) {
            throw new \Exception('Failure: ' . $e->getMessage());
        }
        return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
    }

    public function routeAuthorizedSAMLAction($routeIdentifier, $route, Request $request)
    {
        $client = $this->findHandlerForRoute($routeIdentifier);
        $this->checkAuthorized($request->getMethod());
        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), $route, $request->getQueryString());
        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, $headersExceptForHost, $request->getContent());

        /** @var GuzzleResponse $response */
        try {
            $response = $client->send($hmacRequest, array(self::SAML => true));
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse();
        } catch (\Exception $e) {
            throw new \Exception('Failure: ' . $e->getMessage());
        }
        return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
    }

    public function routeLogoutAction(Request $request)
    {
        $client = $this->findHandlerForRoute(self::SAML);
        $userRequestUri = sprintf("%s/%s?%s", $client->getTargetUri(), self::LOGOUT, $request->getQueryString());
        //Remove host from the headers so guzzle rebuilds it with the port
        $headersExceptForHost = $request->headers->all();
        unset($headersExceptForHost['host']);
        $hmacRequest = new GuzzleRequest($request->getMethod(), $userRequestUri, $headersExceptForHost, $request->getContent());

        /** @var GuzzleResponse $response */
        $response = null;
        try {
            $this->logoutJwtAction($request);
            if (!$request->query->has('loggedout')) {
                $response = $client->send($hmacRequest, array(self::SAML => true));
                return Response::create($response->getBody()->getContents(), $response->getStatusCode(), ['Content-Type' => $response->getHeader('Content-Type')]);
            }
        } catch (BadResponseException $exception) {
            $response = $exception->getResponse()->getBody()->read(2048);
        } catch (\Exception $e) {
            throw new \Exception('Failure: ' . $e->getMessage());
        }
        return new Response($response, Response::HTTP_OK);
    }

    /**
     * @param $routeIdentifier
     * @return ProxyClient
     */
    public function findHandlerForRoute($routeIdentifier)
    {
        $handler = $this->findHandlerMapForRoute($routeIdentifier);

        return $this->containers->get($handler[$routeIdentifier]['handlerName']);
    }

    /**
     * @param $routeIdentifier
     * @return array
     */
    private function findHandlerMapForRoute($routeIdentifier)
    {
        $handlerMap = $this->containers->getParameter('app.proxy.handler.map');

        if (!array_key_exists($routeIdentifier, $handlerMap)) {
            throw new NotFoundHttpException();
        }
        return $handlerMap;
    }

    private function logoutJwtAction(Request $request)
    {

        $jwtKey = $request->headers->get('authorization');
        $haHeader = $request->headers->has('is-cluster-call') ? $request->headers->get('is-cluster-call'): false;
        if (!$haHeader) {
            $jws = CorrectSimpleJWS::load($this->tokenStorage->getToken()->getCredentials());
            $jti = $jws->getPayload()['jti'];
            $dasUri = $this->containers->getParameter('das.uri');
            $dasClient = $this->containers->get("app.proxy.client.das");
            $headers = array(
                "Content-type" => "application/json"
            );
            $dasRequests = new GuzzleRequest('delete', $dasUri . '/token/' . $jti . '?uuid=' . $this->getUser()->getUserId(), $headers);
            try {
                $dasResponse = $dasClient->send($dasRequests);
                $resCode = $dasResponse->getStatusCode();
                if ($resCode != 200) {
                    return -1;
                }
            } catch (BadRequestHttpException | \Exception $error) {
                error_log($error->getMessage());
                return -1;
            }
        }
        $this->jwtLogoutProvider->setJwtLogoutKey($jwtKey);
    }

    public function checkAuthorized($method)
    {
        if ($method === 'POST') {
            if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.' . 'security') && !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.' . 'security')) {
                throw new AccessDeniedException();
            }
        } else {
            if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . 'security')) {
                throw new AccessDeniedException();
            }
        }
    }
    /**
     * @param $routeIdentifier
     * @return array
     */
    private function findPermissionGroupsForRoute($routeIdentifier)
    {
        $handler = $this->findHandlerMapForRoute($routeIdentifier);
        return $handler[$routeIdentifier]['permissions'];
    }


    /**
     * @param String $userId
     * @param array $headers
     * @return GuzzleResponse $response
     */
    private function authenticateSAMLUser($userId, $headers)
    {
        $client = $this->findHandlerForRoute('auth');
        $body = "{ \"credentials\" : {\"loginid\" : \"$userId\", \"token\" : \"\" }, \"domain\" : \"\" , \"type\" : 6,
                    \"tenant\" : \"crudhub\", \"status\" : 1, \"access_type\": \"Concurrent\"}";
        $userRequestUri = sprintf("%s/%s", $client->getTargetUri(), 'users');
        /** @var GuzzleResponse $response */
        $hmacRequest = new GuzzleRequest('POST', $userRequestUri, $headers, $body);
        return $client->send($hmacRequest);
    }

     /**
     * @param String $userId
     * @param array $headers
     * @return GuzzleResponse $response
     */
    private function getSAMLUserDetail($userId, $headers)
    {
        $client = $this->findHandlerForRoute('auth');
        $userRequestUri = sprintf("%s/%s?loginid=%s", $client->getTargetUri(), 'users', $userId);
        /** @var GuzzleResponse $response */
        $headers = array(
            "Content-type" => "application/json"
        );
        $hmacRequest = new GuzzleRequest('GET', $userRequestUri, $headers);
        return $client->send($hmacRequest);
    }


    /**
     * @param GuzzleResponse $response
     * @param String $route
     * @param array $headers
     * @return String $html
     */
    private function validateSAMLUser($response, $route, $headers)
    {
        $userId = null;
        $jwtToken = null;
        $authStatus = null;
        if ($route === self::LOGIN) {
            $actorBody = json_decode($response->getBody()->getContents(), true);
            if (array_key_exists(self::USERID, $actorBody)) {
                $userId = $actorBody[self::USERID];

                # authenticationStatus field is received from FAC.
                $samlAttributes = $actorBody["attributes"];
                if (array_key_exists("authenticationStatus", $samlAttributes)) {
                    $authStatus = $samlAttributes["authenticationStatus"];
                }
                # Validate saml user provioning  
                $preProvisioned = $actorBody["preProvisioned"];
                if ($preProvisioned){
                    $actor = null; 
                    try{
                        $userResp = $this->getSAMLUserDetail($userId, $headers);
                        if ($userResp->getStatusCode() == 200 || $userResp->getStatusCode() == 201) {
                            $userRespBody = json_decode($userResp->getBody()->getContents(), true);
                            if(array_key_exists("usersresp", $userRespBody)){
                                $userResp = $userRespBody["usersresp"][0];
                                $actor = $this->samlActorService->getActor($userResp["uuid"]);
                            }
                        }
                    }catch (BadResponseException $exception) {
                        $response = $exception->getResponse();
                        $actor = null; 
                    }

                    if ($actor === null){
                        return $this->getErrorHTML('The SSO authorization settings are configured for pre-provisioned users. Please contact system administrator to ensure your user credentials are created in the system.');
                    }
                };
                
                /** @var GuzzleResponse $validateResp */
                try {
                    $validateResp = $this->authenticateSAMLUser($userId, $headers);
                    if ($validateResp->getStatusCode() == 200 || $validateResp->getStatusCode() == 201) {
                        $validateRespBody = json_decode($validateResp->getBody()->getContents(), true);
                        $jwtToken = $validateRespBody['token'];
                        $actorBody[self::USERID] = $validateRespBody['uuid'];
                        $actorBody["attributes"]["accessType"] = $validateRespBody['accessType'];
                        $this->samlActorService->saveSamlUser(Person::class, $actorBody);
                        return $this->getHTML($jwtToken, $route, $authStatus);
                    } else {
                        $respBody = json_decode($validateResp->getBody()->getContents(), true);
                        // DAS sends error message in key "Error" or "reason"
                        return $this->getErrorHTML($respBody['Error'] ?? $respBody['reason']);
                    }
                } catch (BadResponseException $exception) {
                    $response = $exception->getResponse();
                    $respBody = json_decode($response->getBody()->getContents(), true);
                    // DAS sends error message in key "Error" or "reason"
                    return $this->getErrorHTML($respBody['Error'] ?? $respBody['reason']);
                } catch (\Exception $e) {
                    throw new \Exception('Failure: ' . $e->getMessage());
                }
            } else {
                return $this->getErrorHTML('Invalid authentication request, no user found');
            }
        }
        return $this->getHTML($jwtToken, $route, $authStatus);
    }

    private function getHTML($jwtToken, $route, $authStatus)
    {
        $managerRegistry = $this->containers->get('doctrine');
        $manager = $managerRegistry->getManagerForClass(SystemSettings::class);
        $systemRepository = $manager->getRepository(SystemSettings::class);
        $systemSetting = $systemRepository->findAll();
        $global_value = $systemSetting[0]->getGlobalValues();
        $host = isset($global_value['hostname']) ? $global_value['hostname'] : $_SERVER['HTTP_HOST'];
        $html = "";
        if ($route === self::LOGIN) {
            # When SSO is configured from FSR to FAC and the user has not enabled 2FA, FSR will show nagging message to
            # user to enable 2FA.
            # Following values are received for authStatus field:
            # 1. "password only"
            # 2. "FortiToken only"
            # 3. "password and FortiToken"
            # 4. "password and Email token"
            # 5."password and SMS token"
            # 6. "federated"
            # Only values "password only" and "FortiToken only" identify the case when 2FA is not enabled.
            if ($authStatus != NULL and (strtolower($authStatus) == "password only" or strtolower($authStatus) == "fortitoken only")) {
                $html = "<html><body onload=\"javascript:window.location.href = 'https://$host/login/$jwtToken?saml=true&show_saml_2fa_warning=true'\"></body></html>";
            } else {
                $html = "<html><body onload=\"javascript:window.location.href = 'https://$host/login/$jwtToken?saml=true&show_saml_2fa_warning=false'\"></body></html>";
            }
        } elseif ($route === self::LOGOUT) {
            $html = "<html><body onload=\"javascript:window.location.href = 'https://$host/logout?loggedout=true'\"></body></html>";
        }
        return $html;
    }

    private function getErrorHTML($error)
    {
        $error = urlencode($error);
        $managerRegistry = $this->containers->get('doctrine');
        $manager = $managerRegistry->getManagerForClass(SystemSettings::class);
        $systemRepository = $manager->getRepository(SystemSettings::class);
        $systemSetting = $systemRepository->findAll();
        $global_value = $systemSetting[0]->getGlobalValues();
        $host = isset($global_value['hostname']) ? $global_value['hostname'] : $_SERVER['HTTP_HOST'];
        return "<html><body onload=\"javascript:window.location.href = 'https://$host/login/?error=$error'\"></body></html>";
    }
}
