<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Controller;

use stdClass;
use InvalidArgumentException;
use App\Service\MetadataHelper;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use App\Constants\CacheConstants;
use App\Entity\Workflow\Workflow;
use App\Controller\BaseController;
use App\Service\BulkRequestService;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\Serializer\SerializerInterface;
use ApiPlatform\Core\Api\ResourceClassResolverInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;

class UpsertController extends BaseController
{

    protected $entityManager;
    protected $queryProvider;
    protected $serializerContextBuilder;
    protected $fsrVoter;
    protected $eventDispatcher;
    protected $validator;
    protected $bulkRequestService;
    protected $utilityService;
    protected $serializer;

    public function __construct(
        ResourceClassResolverInterface $resourceClassResolver,
        ContainerInterface $containers,
        FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        MetadataHelper $metadataHelper,
        UtilityService $utilityService,
        BulkRequestService $bulkRequestService,
        TagAwareCacheInterface $cache,
        SerializerInterface $serializer
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->entityManager = $entityManager;
        $this->bulkRequestService = $bulkRequestService;
        $this->inflector = InflectorFactory::create()->build();
        $this->utilityService = $utilityService;
        $this->serializer = $serializer;
        parent::__construct($resourceClassResolver, $containers, $cache, $metadataHelper);
    }

    const DISALLOWED_MODULES = ['actors', 'files', 'images'];

    public function upsertAction(Request $request, $resourceShortName)
    {
        if (in_array($resourceShortName, self::DISALLOWED_MODULES)) {
            throw new BadRequestHttpException('Upsert Operation for ' . $resourceShortName . ' modules is not supported.');
        }
        $resource = $this->getResourceFromClassName($resourceShortName);
        $requestBody = json_decode($request->getContent());

        [$filterItems, $module] = $this->utilityService->getUniqueResourceFields($requestBody, $resourceShortName);

        $replaceIfExist = isset($requestBody->{'__replace'}) ? $requestBody->{'__replace'} : true;
        $this->bulkRequestService->updateReferenceToIriAndCheckForRelation($requestBody);

        $data_filters = $this->utilityService->prepareFilterData($requestBody, $filterItems);
        $object = [];
        if ($data_filters) {
            $entityRepo = $this->entityManager->getRepository($resource);
            $object = $this->utilityService->getItemWithWhere($entityRepo, $data_filters);
        }
        if (count($object) && !$replaceIfExist) {
            return $this->getSuccessResponse($object[0], $resource);
        }
        if (count($object)) {
            // Existing Object Found, trying to update the existing object
            if (!$this->fsrVoter->isGranted("update." . $module)) {
                throw new AccessDeniedException();
            }
            $object = $object[0];
            $this->utilityService->ValidateOwnerShip($request, $object);
            $groups = UtilityService::identifyChangeFields($requestBody);

            // As per change in computation of workflow and other component
            $normalizedPreviousData = $this->serializer->normalize(
                $object,
                'jsonld',
                ["asJson" => true, "relationships" => true, "groups" => $groups, "ignore_html_escape" => true]
            );
            $request->attributes->set('fsr_previous_data', $normalizedPreviousData);
            $request->attributes->set('fsr_change_fields', $groups);
            unset($requestBody->{'uuid'});
            $denormalizedObject = $this->get('serializer')->deserialize(
                json_encode($requestBody),
                $resource,
                'jsonld',
                ['object_to_populate' => $object, 'fetch_data' => false, '_is_update' => true]
            );
            $request->attributes->set('_api_item_operation_name', 'put');
            $request->attributes->set('previous_data', $object);
        } else {
            // Create the new record
            if (!$this->fsrVoter->isGranted("create." . $module)) {
                throw new AccessDeniedException();
            }
            $denormalizedObject = $this->serializer->deserialize(
                $request->getContent(),
                $resource,
                'jsonld',
                ['fetch_data' => false]
            );
            $request->attributes->set('_api_collection_operation_name', 'post');
            $request->attributes->set('_api_item_operation_name', 'post');
        }
        $request->attributes->set('_api_resource_class', $resource);
        $request->attributes->set('data', $denormalizedObject);
        $request->attributes->set('_format', 'jsonld');
        $request->overrideGlobals();
        return $denormalizedObject;
    }

    public function bulkUpsertAction(Request $request, $resourceShortName, $objectData = null)
    {
        $resourceClass = $this->getResourceFromClassName($resourceShortName);
        $objectDataClone = new stdClass();
        $headers = $request->headers->all();
        $queryParams = $request->query->all();

        if (is_null($objectData)) {
            $objectData = json_decode($request->getContent());
            if (array_key_exists(self::SELECTIVE_FIELDS, $objectData)) {
                $selectedFields = $objectData->{self::SELECTIVE_FIELDS};
                $queryParams[self::SELECTIVE_FIELDS] = implode(',', $selectedFields);
            }
            if (array_key_exists("data", $objectData)) {
                $objectDataClone = $objectData;
            } else if (array_key_exists("__data", $objectData)) {
                $objectDataClone->{'data'} = $objectData->{'__data'};
            } else {
                $objectDataClone->{'data'} = $objectData;
            }

            if ((array_key_exists("__unique", $objectData))) {
                foreach ($objectDataClone->{'data'} as $data) {
                    $data->{'__unique'} = $objectData->{'__unique'};
                }
            }

            if ((array_key_exists("__replace", $objectData))) {
                foreach ($objectDataClone->{'data'} as $data) {
                    $data->{'__replace'} = $objectData->{'__replace'};
                }
            }
        } else {
            // $objectDataClone->{'__unique'} = ['uuid'];
            foreach ($objectData as $data) {
                $data->{'__unique'} = ['uuid'];
            }
            $objectDataClone->{'data'} = $objectData;
        }

        return $this->bulkRequestService->prepareRecordSubRequest(
            $resourceClass,
            $objectDataClone,
            '/api/3/upsert/' . $resourceShortName,
            $headers,
            $queryParams
        );
    }

    public function bulkUpsertWorkflowAction(Request $request)
    {
        $request_data = json_decode($request->getContent());
        $requestDatas = property_exists($request_data, "__data") ? $request_data->__data : $request_data;
        $objectDatas = json_decode(json_encode($requestDatas), true);
        $objectDatas = $this->upsertWorkflows($objectDatas);
        return $this->bulkUpsertAction($request, 'workflows', json_decode(json_encode($objectDatas)));
    }

    public function bulkUpsertWorkflowCollectionsAction(Request $request)
    {
        $request_data = json_decode($request->getContent());
        $requestDatas = property_exists($request_data, "__data") ? $request_data->__data : $request_data;
        $objectDatas = json_decode(json_encode($requestDatas), true);
        $objectDatas = $this->upsertWorkflowCollections($objectDatas);
        return $this->bulkUpsertAction($request, 'workflow_collections', json_decode(json_encode($objectDatas)));
    }

    public function modify_workflow_steps_routes($existingItems, $newItems, $prefix)
    {
        foreach ($newItems as $index => $newitem) {
            foreach ($existingItems as $existingItem) {
                if ($existingItem->getUuid() == $newitem['uuid']) {
                    $newItems[$index]['@id'] = $prefix . $newitem['uuid'];
                    break;
                }
            }
        }
        return $newItems;
    }

    public function upsertWorkflowCollections($workflowCollections)
    {
        foreach ($workflowCollections as $index => $workflowCollection) {
            $workflows = $workflowCollection['workflows'];
            $workflowCollections[$index]['workflows'] = $this->upsertWorkflows($workflows);
        }
        return $workflowCollections;
    }


    public function upsertWorkflows($workflows)
    {
        $resource = Workflow::class;
        foreach ($workflows as $index => $workflow) {
            $data_filters = $this->utilityService->prepareFilterData($workflow, ['uuid']);
            $workflowRepository = $this->entityManager->getRepository(Workflow::class);
            $object = $this->utilityService->getItemWithWhere($workflowRepository, $data_filters);
            if (count($object)) {
                $workflow['@id'] = '/api/3/workflows/' . $workflow['uuid'];
                $workflow['steps'] =  $this->modify_workflow_steps_routes($object[0]->getSteps(), $workflow['steps'], '/api/3/workflow_steps/');
                $workflow['routes'] = $this->modify_workflow_steps_routes($object[0]->getRoutes(), $workflow['routes'], '/api/3/workflow_routes/');
            }
            $workflows[$index] = $workflow;
        }
        return $workflows;
    }
}
