<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

declare(strict_types=1);

namespace App\EventSubscriber;

use Exception;
use Psr\Log\LoggerInterface;
use Doctrine\DBAL\DBALException;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\ErrorHandler\Error\UndefinedMethodError;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\Serializer\Exception\UnexpectedValueException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Security\Core\Exception\AuthenticationException;
use Symfony\Component\HttpKernel\Exception\MethodNotAllowedHttpException;
use ApiPlatform\Core\Bridge\Symfony\Validator\Exception\ValidationException;
use Doctrine\ORM\EntityNotFoundException;

final class ExceptionSubscriber implements EventSubscriberInterface
{

    private $monolog;

    const KNOWN_EXCEPTION = [UndefinedMethodError::class, ValidationException::class, MethodNotAllowedHttpException::class,
    UnexpectedValueException::class, EntityNotFoundException::class];

    public function __construct(LoggerInterface $logger)
    {
        $this->monolog = $logger;
    }

    const UNIQUE_VIOLATION = 23505;
    const UNIQUE_VIOLATION_REGEX = "/Key \((.*?)\)=\((.*)\) already exists\./";

    const FOREIGN_KEY_VIOLATION = 23503;
    const FOREIGN_KEY_VIOLATION_REGEX_DELETE_REF = "/Key \((.*?)\)=\((.*)\) is still referenced from table \"(.*)\"\./";
    const FOREIGN_KEY_VIOLATION_REGEX_INVALID_FOREIGN_KEY = "/Key \((.*?)\)=\((.*)\) is not present in table \"(.*)\"\./";

    const MIN_MAX_CONSTRAINT_VIOLATION = 22001;

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::EXCEPTION => [
                ['authenticationException', 20],
                ['authorizationException', 18],
                ['minMaxConstraintViolations', 16],
                ['foreignKeyException', 14],
                ['uniqueConstraintException', 12],
                ['NotFoundHttpException', 10],
                ['badRequestHttpException', 9],
                ['genericHttpException', 8]
            ],
        ];
    }

    public function NotFoundHttpException(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();
        if ($exception instanceof NotFoundHttpException) {
            $confictMessage = 'Requested resource not found.';
            $error = [
                'type' => $this->getErrorTypeFromException($exception),
                'message' => $confictMessage,
            ];
            $response = new JsonResponse($error, Response::HTTP_NOT_FOUND);
            $event->setResponse($response);
        }
    }

    public function uniqueConstraintException(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();

        if ($exception instanceof DBALException) {
            $previous = $exception->getPrevious();
            if ($previous->getCode() == self::UNIQUE_VIOLATION) {
                preg_match(self::UNIQUE_VIOLATION_REGEX, $previous->getMessage(), $matches);
                $confictMessage = 'Unique violation: ';
                if (count($matches) == 3) {
                    $confictMessage = sprintf("Failed to create the record. The system administrator has defined a uniqueness constraint for the fields (%s) and a record already exists with the specified values for these fields!", $matches[1]);
                } else {
                    $explodedString = explode("Unique violation: 7 ERROR:", $previous->getMessage());
                    if (count($explodedString)) {
                        $confictMessage = trim($explodedString[1]);
                    }
                }
                $error = [
                    'type' => $this->getErrorTypeFromException($exception),
                    'message' => $confictMessage,
                ];
                $response = new JsonResponse($error, Response::HTTP_CONFLICT);
                $event->setResponse($response);
            }
        }
    }

    public function foreignKeyException(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();

        if ($exception instanceof DBALException) {
            $previous = $exception->getPrevious();
            if ($previous->getCode() == self::FOREIGN_KEY_VIOLATION) {
                $confictMessage = 'Foreign key violation';
                preg_match(self::FOREIGN_KEY_VIOLATION_REGEX_DELETE_REF, $previous->getMessage(), $matches);
                if (count($matches) == 4) {
                    $confictMessage = sprintf("%s on field '%s' with data '%s' conflicts with entity '%s'.", $confictMessage, $matches[1], $matches[2], $matches[3]);
                } else {
                    preg_match(self::FOREIGN_KEY_VIOLATION_REGEX_INVALID_FOREIGN_KEY, $previous->getMessage(), $matches);
                    if (count($matches) == 4) {
                        $confictMessage = sprintf("%s on field '%s' with data '%s' is not a valid entry of type '%s'.", $confictMessage, $matches[1], $matches[2], $matches[3]);
                    } else {
                        $confictMessage = $previous->getMessage();
                    }
                }
                $error = [
                    'type' => $this->getErrorTypeFromException($exception),
                    'message' => $confictMessage,
                ];
                $response = new JsonResponse($error, $this->getStatusCodeFromException($exception));
                $event->setResponse($response);
            }
        }
    }

    public function minMaxConstraintViolations(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();
        if ($exception instanceof DBALException) {
            $previous = $exception->getPrevious();
            if ($previous->getCode() == self::MIN_MAX_CONSTRAINT_VIOLATION) {
                $confictMessage = $previous->getMessage();
                $error = [
                    'type' => $this->getErrorTypeFromException($exception),
                    'message' => $confictMessage,
                ];
                $response = new JsonResponse($error, $this->getStatusCodeFromException($exception));
                $event->setResponse($response);
            }
        }
    }

    public function authorizationException(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();
        if ($exception instanceof AccessDeniedException) {
            $confictMessage = $exception->getMessage();
            if (empty($confictMessage)) {
                $confictMessage = 'Access denied.';
            }
            $error = [
                'type' => $this->getErrorTypeFromException($exception),
                'message' => $confictMessage,
            ];
            $response = new JsonResponse($error, Response::HTTP_FORBIDDEN);
            $event->setResponse($response);
        }
    }

    public function authenticationException(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();
        if ($exception instanceof AuthenticationException) {
            $confictMessage = $exception->getMessage();
            if (empty($confictMessage)) {
                $confictMessage = 'Unauthenticated.';
            }
            $error = [
                'type' => $this->getErrorTypeFromException($exception),
                'message' => $confictMessage,
            ];
            $response = new JsonResponse($error, Response::HTTP_UNAUTHORIZED);
            $event->setResponse($response);
        }
    }

    public function badRequestHttpException(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();
        if ($exception instanceof BadRequestHttpException) {
            $confictMessage = $exception->getMessage();
            if (empty($confictMessage)) {
                $confictMessage = 'Runtime Error';
            }
            $error = [
                'type' => $this->getErrorTypeFromException($exception),
                'message' => $confictMessage,
            ];
            $response = new JsonResponse($error, Response::HTTP_BAD_REQUEST);
            $event->setResponse($response);
        }
    }

    public function genericHttpException(ExceptionEvent $event): void
    {
        $request = $event->getRequest();
        if (strpos($request->getPathInfo(), '/api') !== 0) {
            return;
        }
        $exception = $event->getThrowable();
        $this->monolog->addError($exception->getMessage());
        if (method_exists($exception, 'getFile') && method_exists($exception, 'getLine')) {
            $this->monolog->addError($exception->getFile() . ' at line ' . $exception->getLine());
        }
        $confictMessage =  'Internal Server Error. Check log \'prod.log\' for more details';
        if (in_array(get_class($exception), self::KNOWN_EXCEPTION)) {
            $confictMessage = $exception->getMessage();
        } else if (get_class($exception) == HttpException::class) {
            if (strpos($exception->getMessage(), 'Invalid query/body parameters') === 0) {
                $confictMessage = $exception->getMessage();
            }
        }
        $error = [
            'type' => $this->getErrorTypeFromException($exception),
            'message' => $confictMessage,
        ];
        $response = new JsonResponse($error, Response::HTTP_BAD_REQUEST);
        $event->setResponse($response);
    }

    private function getStatusCodeFromException(Exception $exception): int
    {
        if ($exception instanceof HttpException) {
            return $exception->getStatusCode();
        }

        return 500;
    }

    private function getErrorTypeFromException($exception): string
    {
        $parts = explode('\\', get_class($exception));

        return end($parts);
    }
}
