<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Entity\Base\BaseEntity;
use App\Service\MetadataHelper;
use Doctrine\Inflector\Inflector;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\KernelEvents;
use ApiPlatform\Core\Api\IriConverterInterface;
use App\Entity\ModelMetadata\AttributeMetadata;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;

final class LinkedUnlinkSubscriber implements EventSubscriberInterface
{
    protected $iriConverter;

    protected $requeststack;

    protected $propertyAccessor;

    /** @var EntityManagerInterface */
    private $entityManager;

    /** @var MetadataHelper */
    private $metadataHelper;

    private $inflector;

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['LinkUnLinkObjects', 35]
        ];
    }

    public function __construct(IriConverterInterface $iriConverter, RequestStack $requeststack, 
        PropertyAccessorInterface $propertyAccessor, EntityManagerInterface $entityManager, MetadataHelper $metadataHelper)
    {
        $this->iriConverter = $iriConverter;
        $this->requeststack = $requeststack;
        $this->propertyAccessor = $propertyAccessor;
        $this->entityManager = $entityManager;
        $this->metadataHelper = $metadataHelper;
        $this->inflector = InflectorFactory::create()->build();
    }

    protected function handleLinkItems($object, $linkItems, $request, &$attributes)
    {
        $linkItems = $this->handleBadRelationData($linkItems, 'link', get_class($object), $request, $attributes);
        foreach ($linkItems as $attribute => $itemToLinks) {
            $attributeValues = $this->propertyAccessor->getValue($object, $attribute);
            $camelized = $this->inflector->classify($attribute);
            $singulars = $this->inflector->singularize($camelized);
            $addMethod = 'add' . $singulars;
            if (!is_null($attributeValues)) {
                // If sending value as string
                // eg alerts: "/api/3/alerts/123, /api/3/alerts/234"
                if (is_string($itemToLinks)) {
                    if (strlen(trim($itemToLinks)) === 0) {
                        continue;
                    }
                    $itemToLinks = array_map('trim', explode(',', $itemToLinks));
                }
                foreach ($itemToLinks as $itemToLinkIri) {
                    // Check if Entire object has been passed for item to be linked.
                    // in that case check for @id and use that for linking purpose.
                    if (gettype($itemToLinkIri) == 'object') {
                        $itemToLinkIri = $itemToLinkIri->{"@id"};
                    }
                    else if (gettype($itemToLinkIri) == 'array' && array_key_exists('@id', $itemToLinkIri)) {
                        $itemToLinkIri = $itemToLinkIri["@id"];
                    } else if (strpos($itemToLinkIri, '/api/3') === false) {
                        // This should be depricated, Doesn't make sense tyo give this much customization from standard implementation
                        if ($attribute == 'recordTags') {
                            $attribute = 'tags';
                        }
                        $itemToLinkIri = '/api/3/' . $attribute . '/' . $itemToLinkIri;
                    }
                    $itemToLinkObj = $this->iriConverter->getItemFromIri($itemToLinkIri, ['fetch_data' => false]);
                    $object->{$addMethod}($itemToLinkObj);
                }
            }
        }
    }

    public function handleUnLinkItems($object, $unLinkItems, $request, &$attributes)
    {
        $unLinkItems = $this->handleBadRelationData($unLinkItems, 'unlink', get_class($object), $request, $attributes);
        foreach ($unLinkItems as $attribute => $itemToUnLinks) {
            $attributeValues = $this->propertyAccessor->getValue($object, $attribute);
            if (!is_null($attributeValues)) {
                $camelized = $this->inflector->classify($attribute);
                $singulars = $this->inflector->singularize($camelized);
                $removeMethod = 'remove' . $singulars;
                // If sending value as string
                // eg alerts: "/api/3/alerts/123, /api/3/alerts/234"
                if (is_string($itemToUnLinks)) {
                    if (strlen(trim($itemToUnLinks)) === 0) {
                        continue;
                    }
                    $itemToUnLinks = array_map('trim', explode(',', $itemToUnLinks));
                }
                foreach ($itemToUnLinks as $itemToUnLinkIri) {
                    // Check if Entire object has been passed for item to be unlinked.
                    // in that case check for @id and use that for linking purpose.
                    if (gettype($itemToUnLinkIri) == 'object') {
                        $itemToUnLinkIri = $itemToUnLinkIri->{"@id"};
                    }
                    else if (gettype($itemToUnLinkIri) == 'array' && array_key_exists('@id', $itemToUnLinkIri)) {
                        $itemToUnLinkIri = $itemToUnLinkIri["@id"];
                    }
                    $itemToUnLinkObj = $this->iriConverter->getItemFromIri($itemToUnLinkIri, ['fetch_data' => false]);
                    $object->{$removeMethod}($itemToUnLinkObj);
                }
            }
        }
    }

    public function LinkUnLinkObjects(ViewEvent $event): void
    {
        $object = $event->getControllerResult();
        $request = $event->getRequest();
        $request_method = $request->getMethod();
        if (!($object instanceof BaseEntity) || $request_method == Request::METHOD_GET || $request_method == Request::METHOD_DELETE) {
            return;
        }
        $this->manipulateObject($request, $object);
    }

    public function manipulateObject($request, $object)
    {
        $request_body = json_decode($request->getContent());
        if ($request_body && (array_key_exists("__link", $request_body) || array_key_exists("__unlink", $request_body))) {
            $linkItems = (array)$request_body->{'__link'} ?? [];
            $unlinkItems = (array)$request_body->{'__unlink'} ?? [];
            $attributes = null;
            if (count($linkItems)) {
                $this->handleLinkItems($object, $linkItems, $request, $attributes);
            }
            if (count($unlinkItems)) {
                $this->handleUnLinkItems($object, $unlinkItems, $request, $attributes);
            }
        }
    }

    private function handleBadRelationData($relationItems, $relationName, $className, $request, &$classAttributes)
    {
        $hasBadRelationData = false;
        foreach ($relationItems as $innerKey => $innerValue) {
            if (is_string($innerValue)) {
                $innerValue = $this->parseBadRelationData($innerValue);
                if (is_object($relationItems)) {
                    $relationItems->$innerKey = $innerValue;
                } else {
                    $relationItems[$innerKey] = $innerValue;
                }
                $hasBadRelationData = true;
            }
            # if relation UUIDs have been used instead of IRIs, convert them to IRIs
            # checking first element should suffice
            if ($innerKey != 'recordTags' && count($innerValue) > 0 && is_string($innerValue[0]) && strpos($innerValue[0], '/api/') === false) {
                $innerValue = $this->convertUUIDToIRI($innerValue, $innerKey, $className, $classAttributes);
                if (is_object($relationItems)) {
                    $relationItems->$innerKey = $innerValue;
                } else {
                    $relationItems[$innerKey] = $innerValue;
                }
                $hasBadRelationData = true;
            }
        }
        if ($hasBadRelationData === true){
            $request->attributes->set('fsr_' . $relationName . '_data_modified', $relationItems);
        }
        return $relationItems;
    }

    private function parseBadRelationData($relationString)
    {
        $value = trim($relationString);
        if (strlen($value) > 0) {
            return explode(",", $value);
        } else {
            return [];
        }
    }

    private function convertUUIDToIRI($uuidList, $key, $className, &$attributes){
        $iriList = [];
        $relationType = null;
        //lazy load only if relations have UUID and send back to caller
        if (is_null($attributes)){
            //lazy load only if relations have UUID and send back to caller
            $entityType = $this->metadataHelper->getTypeFromClassName($className);
            $modelMetaDataRepository = $this->entityManager->getRepository(AttributeMetadata::class);
            $attribute = $modelMetaDataRepository->createQueryBuilder('o')->andWhere('smd.type = :entityType')->andWhere('o.name = :name')->leftJoin('o.sattrib', 'smd')->setParameter('entityType', $entityType)->setParameter('name', $key)->getQuery()->getOneOrNullResult();
            if (is_null($attribute)){
                # could be a system module
                $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
                $systemFixture = json_decode($systemFixturesFile);
                $modelMetaDataForType = $systemFixture->{$entityType};
                $attributes = $modelMetaDataForType->{'attributes'};
                foreach($attributes as $eachAttribute){
                    if ($eachAttribute->getName() == $key){
                        $relationType = $eachAttribute->getType();
                        break;
                    }
                }
            } else{
                $relationType = $attribute->getType();       
            }
        }
        foreach($uuidList as $uuid){
            array_push($iriList, '/api/3/'. $relationType . '/' . $uuid);
        }
        return $iriList;
    }
}
