<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use ApiPlatform\Core\Api\IriConverterInterface;
use App\Entity\Base\BaseEntity;
use App\Entity\Core\Picklist;
use App\Entity\Core\PicklistName;
use App\Entity\Custom\Agent;
use App\Entity\Custom\Router;
use App\Entity\Custom\Tenant;
use App\Entity\ModelMetadata\AttribModelMetadata;
use App\Entity\ModelMetadata\ModelMetadata;
use App\Entity\ModelMetadata\StagingModelMetadata;
use App\Utility\MsspUtility;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\DependencyInjection\ParameterBag\ParameterBagInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\SerializerInterface;
use Throwable;

final class MqMessagebroadcastSubscriber implements EventSubscriberInterface
{

    const AGENT_REPLICATION_ROUTINGKEY = 'route.crudhub.tenant.request.agentconfig';
    const TENANT_REPLICATION_ROUTINGKEY = "route.crudhub.tenant.request.tenantconfig";
    const ROUTER_REPLICATION_ROUTINGKEY = 'route.crudhub.tenant.request.routerconfig';

    const DELTA_DATA = 'data';
    const HTTP_METHOD = 'operation';
    const REMOTE_ADDR = 'source';
    const WEB_SOCKET_SESSION_ID = 'websocketId';
    const COMPONENT = 'component';
    const ENTITY_TYPE = 'entityType';
    const UUID = 'entityUuid';
    const USER_UUID = 'userId';
    const USER = 'user';
    const TRANSACTION_DATE = 'transactionDate';
    const PLAYBOOK_IRI = 'playbookIri';
    const PLAYBOOK_NAME = 'playbookName';
    const TITLE = 'title';
    const DEFALUT_KEY = 'cyops.crudhub.datanotify';
    const ELASTIC_SEARCH_OPERATION_KEY = 'esoperation';

    protected $iriConverter;

    protected $requeststack;

    protected $container;

    protected $tokenStorage;

    protected $serializer;

    protected $msspUtility;

    private $parameters;

    private $inflector;

    /** @var EntityManagerInterface */
    private $entityManager;
    

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['broadcastData', 15],
        ];
    }

    public function __construct(
        IriConverterInterface $iriConverter,
        RequestStack $requeststack,
        ContainerInterface $container,
        TokenStorageInterface $tokenStorage,
        SerializerInterface $serializer,
        MsspUtility $msspUtility,
        ParameterBagInterface $parameters,
        EntityManagerInterface $entityManager
    ) {
        $this->iriConverter = $iriConverter;
        $this->requeststack = $requeststack;
        $this->container = $container;
        $this->tokenStorage = $tokenStorage;
        $this->serializer = $serializer;
        $this->msspUtility = $msspUtility;
        $this->parameters = $parameters;
        $this->entityManager = $entityManager;
        $this->inflector = InflectorFactory::create()->build();
    }

    public function broadcastData(ViewEvent $event)
    {
        $request = $event->getRequest();
        $requestMethod = $request->getMethod();
        if ($requestMethod == Request::METHOD_GET) {
            return;
        }
        $currentData = $event->getRequest()->attributes->get('data');
        $previousData = $request->attributes->get('previous_data');
        if (($currentData && !($currentData instanceof BaseEntity)) || ($previousData && !($previousData instanceof BaseEntity))) {
            return;
        }

        $normalized_current_data = json_decode($event->getControllerResult(), true);
        $normalized_previous_data = (object) null;     
        $changeData = [];
        $data_to_broadcast = [];
        $isImporting = $request->attributes->get('fsr_subrequest_type') == 'import';

        if (is_null($previousData) && $currentData) {
            $entity = get_class($currentData);
            $requestOperation = $isImporting ? 'Import' : 'Create';
            $replicaionOp = 'POST';
            $data_to_broadcast[self::ELASTIC_SEARCH_OPERATION_KEY] = 'index';
        } else if ($requestMethod == Request::METHOD_DELETE) {
            $replicaionOp = 'DELETE';
            $data_to_broadcast[self::ELASTIC_SEARCH_OPERATION_KEY] = 'delete';
            $entity = get_class($previousData);
            $requestOperation = $isImporting ? 'Overwrite During Import' : 'Delete';
            $normalized_previous_data = $request->attributes->get('fsr_previous_data');
        } else if ($previousData && $currentData) {
            $replicaionOp = 'PUT';
            $entity = get_class($currentData);
            $data_to_broadcast[self::ELASTIC_SEARCH_OPERATION_KEY] = 'modify';
            $requestOperation = $isImporting ? 'Update During Import' : 'Update';
            $normalized_previous_data = $request->attributes->get('fsr_previous_data');
            $entityType = explode("/", $normalized_previous_data['@id'])[3];
            $requestBody = json_decode($request->getContent());
            if (!is_null($request->attributes->get('fsr_link_data_modified'))) {
                $requestBody->{'__link'} = $request->attributes->get('fsr_link_data_modified');
            }
            if (!is_null($request->attributes->get('fsr_unlink_data_modified'))) {
                $requestBody->{'__unlink'} = $request->attributes->get('fsr_unlink_data_modified');
            }
            $changeData = $this->identifyChangeData($requestBody, $normalized_previous_data, $entityType);
        }

        // for lookup fields of non-system modules, audit and notifier expect iris
        if (!($currentData instanceof Picklist) && !($currentData instanceof PicklistName) && !($currentData instanceof StagingModelMetadata)){
            $this->replaceLookupWithIri($normalized_current_data);
        }

        // Data to transmit to Audit, Notifier and Falcon Services
        $data_to_broadcast = array_merge($data_to_broadcast, $this->prepareChangeData(
            $normalized_current_data,
            $normalized_previous_data,
            $requestOperation,
            $changeData,
            $entity
        ));

        // Data to transmit to Postman service
        // This needs to be reimplemented in future
        // All services should work on standard message
        // Any alteration required in data should be done by component not the API layer service
        $diffObj = $this->msspUtility->prepareReplicationDiffData($normalized_current_data, $replicaionOp, $normalized_previous_data, $changeData);

        // All these needs to be standard over singile message
        if ($entity == Agent::class) {
            $data = $this->msspUtility->prepareReplicationData($diffObj, $replicaionOp, null, null);
            $this->msspUtility->sendReplicationData($data, self::AGENT_REPLICATION_ROUTINGKEY);
        } else if ($entity == Tenant::class) {
            $data = $this->msspUtility->prepareReplicationData($diffObj, $replicaionOp, null, null);
            $this->msspUtility->sendReplicationData($data, self::TENANT_REPLICATION_ROUTINGKEY);
        } else if ($entity == Router::class) {
            $data = $this->msspUtility->prepareReplicationData($diffObj, $replicaionOp, null, null);
            $this->msspUtility->sendReplicationData($data, self::ROUTER_REPLICATION_ROUTINGKEY);
        }

        if ($event->getRequest()->attributes->get('fsr_subrequest_type') === null) {
            if (!($entity == Agent::class || $entity == Tenant::class || $entity == Router::class)) {
                $this->msspUtility->replicateData($entity, $requestBody, [$diffObj], $replicaionOp);
            }
            $this->sendData([$data_to_broadcast]);
            return;
        }
        $audit_data = $request->attributes->get('fsr_audit_data') ?? [];
        $audit_data[] = $data_to_broadcast;
        $request->attributes->set('fsr_audit_data', $data_to_broadcast);
        $request->attributes->set('fsr_replication_data', $diffObj);
    }

    private function identifyChangeData($requestBody, $normalizedRecord, $entityType)
    {
        $attributes = null;
        foreach ($requestBody as $key => $value) {
            if ($key == '__link') {
                foreach ($requestBody->{'__link'} as $innerKey => $innerValue) {
                    if (is_string($innerValue)) {
                        $innerValue = explode(',', $innerValue);
                    } else if (is_array($innerValue)) {
                        $innerVal = [];
                        foreach ($innerValue as $item) {
                            if (is_object($item)) {
                                $innerVal[] = $item->{'@id'};
                            } else {
                                $innerVal[] = $item;
                            }
                        }
                        $innerValue = $innerVal;
                    }
                    if (!empty($requestBody->$innerKey)) {
                        $requestBody->$innerKey = array_unique(array_merge($requestBody->$innerKey, $innerValue));
                    } else {
                        $requestBody->$innerKey = array_unique(array_merge($normalizedRecord[$innerKey], $innerValue));
                    }
                }
            } else if ($key == '__unlink') {
                foreach ($requestBody->{'__unlink'} as $innerKey => $innerValue) {
                    if (is_string($innerValue)) {
                        $innerValue = explode(',', $innerValue);
                    } else if (is_array($innerValue)) {
                        $innerVal = [];
                        foreach ($innerValue as $item) {
                            if (is_object($item)) {
                                $innerVal[] = $item->{'@id'};
                            } else {
                                $innerVal[] = $item;
                            }
                        }
                        $innerValue = $innerVal;
                    }
                    if (!empty($requestBody->$innerKey)) {
                        $requestBody->$innerKey = array_unique(array_diff($requestBody->$innerKey, $innerValue));
                    } else {
                        $requestBody->$innerKey = array_unique(array_diff($normalizedRecord[$innerKey], $innerValue));
                    }
                }
            }
            // we allow relations to be provided as a comma separated string also
            // that should be converted to an array for notifier, audit, postman to correctly identify them as a relation field
            else if (array_key_exists($key, $normalizedRecord) && is_array($normalizedRecord[$key]) && $key != 'recordTags') {
                if (is_string($value)) {
                    $value = $this->parseBadRelationData($value);
                }
                if ($key != 'recordTags' && is_array($value) && sizeof($value) > 0 && is_string($value[0]) && strpos($value[0], '/api/') === false) {
                    $value = $this->convertUUIDToIRI($value, $key, $entityType, $attributes);
                }
                $requestBody->$key = $value;
            }
        }
        if (isset($requestBody->{"@id"})) {
            unset($requestBody->{"@id"});
        }
        return $requestBody;
    }

    private function parseBadRelationData($relationString)
    {
        $value = trim($relationString);
        if (strlen($value) > 0) {
            return explode(",", $value);
        } else {
            return [];
        }
    }

    private function replaceLookupWithIri(&$jsonData) {
        if (!is_null($jsonData) && is_array($jsonData)) {
            foreach($jsonData as $key => $value){
                if (is_array($value)) {
                    if (array_key_exists('@id', $value)){
                        $jsonData[$key]  = $value['@id'];
                    } else {
                        // it could be a collection
                        $modifiedValues = [];
                        foreach($value as $index => $item){
                            if (is_array($item) && array_key_exists('@id', $item)){
                                // this is an fsr entity
                                $this->replaceLookupWithIri($item);
                                array_push($modifiedValues, $item);
                            }
                        }
                        if (count($modifiedValues) > 0){
                            $jsonData[$key] = $modifiedValues;
                        }
                    }
                }
            }
        }
    }

    private function convertUUIDToIRI($uuidList, $key, $entityType, &$attributes)
    {
        $iriList = [];
        if (is_null($attributes)) {
            //lazy load only if relations have UUID and send back to caller
            $modelMetaDataRepository = $this->entityManager->getRepository(ModelMetadata::class);
            $modelMetaDataForType = $modelMetaDataRepository->findOneBy(['type' => $entityType]);
            if (is_null($modelMetaDataForType)) {
                # could be a system module
                $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
                $systemFixture = json_decode($systemFixturesFile);
                $modelMetaDataForType = $systemFixture->{$entityType};
                $attributes = $modelMetaDataForType->{'attributes'};
            } else {
                $attributes = $modelMetaDataForType->getAttributes();
            }
        }
        if (get_class($attributes[0]) === AttribModelMetadata::class) {
            foreach ($attributes as $eachAttribute) {
                if ($eachAttribute->getName() == $key) {
                    $relationType = $eachAttribute->getType();
                    break;
                }
            }
        } else {
            foreach ($attributes as $eachAttribute) {
                if ($eachAttribute->{'name'} == $key) {
                    $relationType = $eachAttribute->{'type'};
                    break;
                }
            }
        }

        foreach ($uuidList as $uuid) {
            array_push($iriList, '/api/3/' . $relationType . '/' . $uuid);
        }
        return $iriList;
    }


    public function sendData(array $data, $key = self::DEFALUT_KEY)
    {
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode($data), $key);
    }

    public function getUser()
    {
        return $this->tokenStorage->getToken()->getUser();
    }

    public function getUserName()
    {
        $user = $this->getUser();
        if (method_exists($user, 'getFirstname')) {
            return $user->getFirstname() . ' ' . $user->getLastname();
        }
        return $user->getName();
    }

    public static function extractEntityType($previousObject, $updatedObject, $identity)
    {
        if (is_array($updatedObject) && array_key_exists($identity, $updatedObject)) {
            return $updatedObject[$identity];
        } elseif (is_array($previousObject) && array_key_exists($identity, $previousObject)) {
            return $previousObject[$identity];
        }
        return null;
    }

    /**
     * @param $updatedObject
     * @param $previousObject
     * @param $method
     * @param array $changeData
     * @return array
     */
    public function prepareChangeData(
        $updatedObject,
        $previousObject,
        $method,
        $changeData = [],
        $entity = null,
        $title = null
    ) {
        $delta_data = [];
        $delta_data['oldData'] = $previousObject;
        $delta_data['newData'] = $updatedObject;
        $delta_data['changeData'] = $changeData;
        $userName = $this->getUserName();
        $userId = $this->getUser()->getUuid();
        $entityType = $this->extractEntityType($previousObject, $updatedObject, "@type");
        $entityId = $this->getEntityUuid($updatedObject, $previousObject);
        return [
            self::DELTA_DATA => $delta_data,
            self::HTTP_METHOD => $method,
            self::REMOTE_ADDR => $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'] ?? '127.0.0.1',
            self::WEB_SOCKET_SESSION_ID => isset($_SERVER['HTTP_WEBSOCKET_SESSIONID']) ? $_SERVER['HTTP_WEBSOCKET_SESSIONID'] : null,
            self::COMPONENT => 'crudhub',
            self::ENTITY_TYPE => $entityType,
            self::UUID => $entityId,
            self::USER_UUID => $userId,
            self::TRANSACTION_DATE => round(microtime(true) * 1000),
            self::USER => $userName,
            self::PLAYBOOK_IRI => isset($_SERVER['HTTP_X_CYOPS_PLAYBOOKIRI']) ? $_SERVER['HTTP_X_CYOPS_PLAYBOOKIRI'] : null,
            self::PLAYBOOK_NAME => isset($_SERVER['HTTP_X_CYOPS_PLAYBOOKNAME']) ? $_SERVER['HTTP_X_CYOPS_PLAYBOOKNAME'] : null,
            self::TITLE => $title
        ];
    }

    private function getEntityUuid($updatedObject, $previousObject)
    {
        if (is_array($updatedObject)) {
            return $updatedObject['uuid'];
        }
        return $previousObject['uuid'];
    }
}
