<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Traits\Ownable;
use App\Traits\Trackable;
use App\Entity\Custom\Tenant;
use App\Traits\PeerReplicable;
use App\Entity\Base\BaseEntity;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\HttpKernel\Exception\HttpException;
use ApiPlatform\Core\DataPersister\DataPersisterInterface;
use App\Service\UtilityService;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class MsspEventSubscriber implements EventSubscriberInterface
{

    protected $containers;

    protected $tokenStorage;

    protected $entityManager;

    protected $requeststack;

    protected $cache;

    private const MASTER_ROLE_NAME = 'master';

    protected $dataPersister;
    private $utilityService;

    public function __construct(
        ContainerInterface $containers,
        TokenStorageInterface $tokenStorage,
        EntityManagerInterface $entityManager,
        DataPersisterInterface $dataPersister,
        TagAwareCacheInterface $cache,
        UtilityService $utilityService
    ) {
        $this->containers = $containers;
        $this->tokenStorage = $tokenStorage;
        $this->entityManager = $entityManager;
        $this->dataPersister = $dataPersister;
        $this->cache = $cache;
        $this->utilityService = $utilityService;
    }

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['handleObjectForMsspChanges', 34]
        ];
    }

    public function handleObjectForMsspChanges(ViewEvent $event)
    {
        $request = $event->getRequest();
        $object = $event->getControllerResult();
        $request = $event->getRequest();
        $request_method = $request->getMethod();

        if (!($object instanceof BaseEntity) || $request_method == Request::METHOD_GET || $request_method == Request::METHOD_DELETE) {
            return;
        }

        $previous_data = $event->getRequest()->attributes->get('previous_data');
        $this->msspOperations($request, $object, $previous_data);
    }

    protected function msspOperations($request, $object, $previous_data, $event = null)
    {
        $isDistributedInstance = $this->utilityService->checkForMsspInstance();
        if (!$isDistributedInstance) {
            return;
        }
        if (is_null($previous_data)) {
            // POST Request
            $this->setTenantObject($object);
        } else {
            // PUT REQUEST
            $request_body = json_decode($request->getContent());
            $entityClass = get_class($object);
            //Set Owners to Records when tenant is changed.

            if ($this->isPeerReplicable($entityClass) && is_null($object->getTenant())) {
                $object->setTenant($previous_data->getTenant());
            }

            if (
                $this->isPeerReplicable($entityClass) && $this->isOwnable($entityClass) &&
                $previous_data->getTenant()->getUuid() != $object->getTenant()->getUuid()
            ) {
                $tenantObject = $object->getTenant();
                foreach ($tenantObject->getOwners() as $owner) {
                    $object->addOwner($owner);
                }
            }

            $conflict = $this->isConflcts($object, $previous_data, $request_body);
            if ($conflict) {
                $previous_data->setConflict($conflict);
                if ($event) {
                    $event->setControllerResult($previous_data);
                }
            }
        }
    }

    public function isConflcts($object, $previous_data, $request_body = null)
    {
        if (isset($request_body->{'__replicateData'})) {
            return false;
        }
        $entityClass = get_class($object);
        if ($this->isPeerReplicable($entityClass) && $this->isTrackable($entityClass)) {
            $instanceRole = $this->utilityService->getInstanceRole();
            $modifyUserUuidOfRequestedObj = $this->tokenStorage->getToken()->getUser()->getUuid();
            $modifyUserUuidOfExistingObj = $previous_data->getModifyUser() ? $previous_data->getModifyUser()->getUuid() : null;
            $modifyDateOfExistingObj = $previous_data->getModifyDate()->getTimestamp() ?? 0;
            $modifyDateOfRequestedObj = $request_body->{'modifyDate'} ?? PHP_INT_MAX;
            if ($instanceRole != self::MASTER_ROLE_NAME && ($modifyUserUuidOfRequestedObj != $modifyUserUuidOfExistingObj) && ($modifyDateOfExistingObj >= $modifyDateOfRequestedObj)) {
                return true;
            }
            return false;
        }
    }

    public function isPeerReplicable($entity_class)
    {
        return in_array(PeerReplicable::class, class_uses($entity_class));
    }

    public function isTrackable($entity_class)
    {
        return in_array(Trackable::class, class_uses($entity_class));
    }

    public function isOwnable($entity_class)
    {
        return in_array(Ownable::class, class_uses($entity_class));
    }

    public function getSelfTenantObject()
    {
        $resourceClass = Tenant::class;
        $tenantRepo = $this->entityManager->getRepository($resourceClass);
        $selfObj = $tenantRepo->findOneBy(["role" => "self"]);
        return $selfObj;
    }

    public function setTenantObject($object)
    {
        $entityClass = get_class($object);
        // Check if Peer Relicable is enabled for class
        if ($this->isPeerReplicable($entityClass)) {
            // Set Self Tenant ID for Objects if not already came as part of requests
            if (!$object->getTenant() || is_null($object->getTenant())) {
                $selfTenantObject = $this->getSelfTenantObject();
                if (is_null($selfTenantObject)) {
                    throw new HttpException(Response::HTTP_BAD_REQUEST, "Self TenantId not defined for instance");
                }
                $object->setTenant($selfTenantObject);
            }
            //Set Owners to Records
            if ($this->isOwnable($entityClass)) {
                $tenantObject = $object->getTenant();
                foreach ($this->tokenStorage->getToken()->getUser()->getTeams() as $owner) {
                    $object->addOwner($owner);
                }
                foreach ($tenantObject->getOwners() as $owner) {
                    $object->addOwner($owner);
                }
            }
            $object->setConflict(false);
        }

        return $object;
    }
}
