<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use Error;
use App\Entity\Actors\Actor;
use App\Entity\Custom\Person;
use App\Entity\Base\BaseEntity;
use App\Service\UtilityService;
use Doctrine\ORM\EntityManager;
use App\Entity\Settings\SystemSettings;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use GuzzleHttp\Exception\BadResponseException;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use ApiPlatform\Core\EventListener\EventPriorities;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class PeopleSubscriber implements EventSubscriberInterface
{

    /** @var EntityManager */
    private $entityManager;
    /** @var ContainerInterface */
    private $container;
    /** @var UtilityService  */
    private $utilityService;

    // keeping same constant as in DAS 
    const FSR_USER_CREATE = "User Create";

    public function __construct(
        EntityManagerInterface $entityManager,
        ContainerInterface $container,
        UtilityService $utilityService
    ) {
        $this->entityManager = $entityManager;
        $this->container = $container;
        $this->utilityService = $utilityService;
    }

    public static function getSubscribedEvents()
    {
        // Adding this on POST WRITE PRIORITY in order so data is persisted in database and then handling of data can be done.
        return [
            KernelEvents::VIEW => [
                ['handlePeopleCreate', EventPriorities::POST_WRITE],
                ['handlePeoplePreCreate', EventPriorities::PRE_WRITE]
            ]
        ];
    }

    // Send mail to user on create of People.
    public function handlePeopleCreate(ViewEvent $event)
    {
        $dasClient = $this->container->get("app.proxy.client.das");
        $integrationClient = $this->container->get("app.proxy.client.integrations");
        $request = $event->getRequest();
        $request_data = json_decode($request->getContent(), true);
        $method = $request->getMethod();
        $object = $event->getControllerResult();
        $previousData = $request->attributes->get('previous_data');

        if (!$method || $method == Request::METHOD_GET || $method == Request::METHOD_DELETE || !$object instanceof BaseEntity) {
            return;
        }
        $entityClass = get_class($object);
        if (!($entityClass == Person::class)) {
            return;
        }
        if (!is_null($previousData)) {
            return;
        }
        $dasUri = $this->container->getParameter('das.uri');
        $headers = array(
            "Content-type" => "application/json",
        );
        $keyFile = file_get_contents('/opt/cyops-api/src/keys/daspublic.key', true);
        $idKey = base64_encode($keyFile);
        $objectData = array();
        $objectData['idKey'] = $idKey;
        $objectData['uuid'] = $object->getUserId();
        $dasRequests = new GuzzleRequest('post', $dasUri . '/token?uuid=' . $object->getUserId(), $headers, json_encode($objectData));
        try {
            $dasResponse = $dasClient->send($dasRequests);
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $resCode = $dasResponse->getStatusCode();
        } catch (BadRequestHttpException | \Exception $error) {
            error_log($error->getMessage());
            $errVar = json_decode($error->getResponse()->getBody()->read(2048), true);
            $errMsg['message'] = $errVar['Error'];
            $errCode = $error->getCode();
            error_log($errMsg['message']);
            error_log($errCode);
            return;
        }
        $actorEmail = $object->getEmail();

        if ($resCode < 300) {
            $systemRepository = $this->entityManager->getRepository(SystemSettings::class);
            $systemSetting = $systemRepository->findAll();
            $global_value = $systemSetting[0]->getGlobalValues();
            $system_host = isset($global_value['hostname']) ? $global_value['hostname'] : $request->getHttpHost();

            $dataToReplace = [];
            $dataToReplace['{{vars.userId}}'] = $request_data['username'];
            $dataToReplace['{{vars.token}}'] = $jsonDecodeResponse['token'];
            $dataToReplace['{{vars.request.baseUri}}'] = $request->getScheme() . "://" . $system_host;
            $dataToReplace['{{@Server_fqhn}}'] = $request->getScheme() . "://" . $system_host;
            $dataToReplace['{{@cyops_hostname}}'] = $request->getScheme() . "://" . $system_host;
            $mailDetails = [];
            $mailDetails['to'] = $actorEmail;
            $defaultMessage = 'newpassword';
            $mailResponse = $this->utilityService->sendMail(
                '12f23d8d-505c-4b4b-8911-6595c99b07f0',
                $dataToReplace,
                $mailDetails,
                $integrationClient,
                $defaultMessage
            );
            if ($mailResponse === true) {
                $sucessMsg['message'] = "Reset token sent to user";
                error_log($sucessMsg['message']);
            } else {
                $sucessMsg['message'] = "Reset token generated,mail not sent due to smtp connector failure";
                error_log($sucessMsg['message']);
            }
        }
    }



    public function handlePeoplePreCreate(ViewEvent $event)
    {
        $dasClient = $this->container->get("app.proxy.client.das");
        $request = $event->getRequest();
        $method = $request->getMethod();
        $object = $event->getControllerResult();

        if (!$method || $method == Request::METHOD_GET || $method == Request::METHOD_DELETE || $method == Request::METHOD_PUT || !$object instanceof BaseEntity) {
            return;
        }

        $entityClass = get_class($object);
        if (!($entityClass == Person::class)) {
            return;
        }

        $request_data = json_decode($request->getContent(), true);

        if ($request_data['user']) {

            $dasUri = $this->container->getParameter('das.uri');
            $headers = array(
                "Content-type" => "application/json",
            );

            $access_type = $request_data['user']['access_type'] ?? 'Concurrent';

            if (!in_array($access_type, ['Concurrent', 'Named'])) {
                $access_type = 'Concurrent';
            }

            $body = [
                'credentials' => [
                    'loginid' => $request_data['user']['loginid']
                ],
                'status' => array_key_exists('status', $request_data['user']) ? $request_data['user']['status'] : 2,
                'type' => array_key_exists('user_type', $request_data['user']) ? $request_data['user']['user_type'] : 2,
                'domain' => $request_data['user']['domain'],
                'nfa' => array_key_exists('nfa', $request_data['user']) ? $request_data['user']['nfa'] : [],
                'bind_name' => $request_data['user']['bind_name'],
                'tenant' => $request_data['user']['tenant'],
                'access_type' => $access_type,
                'event_type' => self::FSR_USER_CREATE
            ];
            $dasRequests = new GuzzleRequest('post', $dasUri . '/users', $headers, json_encode($body));
            try {
                $dasResponse = $dasClient->send($dasRequests);
            } catch(BadResponseException | \Exception $exception) {
                error_log($exception->getMessage());
                if (method_exists($exception, 'getResponse')) {
                    throw new BadRequestHttpException($exception->getResponse()->getBody()->read(2048));
                }
                throw new BadRequestHttpException($exception->getMessage());
            }
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $actorRepo = $this->entityManager->getRepository(Actor::class);
            $user_id = $jsonDecodeResponse['uuid'];
            if ($actorRepo->findOneBy(["userId" => $user_id])) {
                $event->stopPropagation();
                throw new BadRequestHttpException(sprintf('The user with loginID %s already exists',  $request_data['user']['loginid']));
            }
            $object->setUserId($user_id);
        }
        return;
    }
}
