<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use Psr\Log\LoggerInterface;
use App\Entity\Workflow\Rule;
use App\Entity\Base\BaseEntity;
use App\Service\MetadataHelper;
use App\Service\WorkflowHelper;
use App\Entity\Workflow\Workflow;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class RuleEventSubscriber implements EventSubscriberInterface
{
    const RULE_PRIORITY = 2;

    const TRIGGER_WHEN_POST = 'post';
    const TRIGGER_ACTION_CREATE = 'create';
    const TRIGGER_ACTION_UPDATE = 'update';
    private $workflowHelper;
    private $entityManager;
    private $serializer;
    private $metadataHelper;
    private $logger;
    private $container;

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['handleTriggerType', 14]
        ];
    }

    public function __construct(
        WorkflowHelper $workflowHelper,
        EntityManagerInterface $entityManager,
        SerializerInterface $serializer,
        MetadataHelper $metadataHelper,
        LoggerInterface $logger,
        ContainerInterface $container
    ) {
        $this->workflowHelper = $workflowHelper;
        $this->entityManager = $entityManager;
        $this->serializer = $serializer;
        $this->metadataHelper = $metadataHelper;
        $this->logger = $logger;
        $this->container = $container;
    }

    public function handleTriggerType(ViewEvent $event)
    {
        $newData = $event->getRequest()->attributes->get('data');
        $method = $event->getRequest()->getMethod();
        $subRequestType = $event->getRequest()->attributes->get('fsr_subrequest_type');
        if (Request::METHOD_GET == $method || Request::METHOD_DELETE == $method || $subRequestType == 'agnostic_bulk' || $subRequestType == 'import' || is_null($newData) || !($newData instanceof BaseEntity) || ($newData instanceof Rule) || ($newData instanceof Workflow)) {
            return;
        }

        $oldData = $event->getRequest()->attributes->get('previous_data');
        $oldDataNormalized = $event->getRequest()->attributes->get('fsr_previous_data');
        $resourceClassName = get_class($newData);
        $triggerResource = $this->metadataHelper->getTypeFromClassName($resourceClassName);

        $when = self::TRIGGER_WHEN_POST;
        if (is_null($oldData) && Request::METHOD_POST == $method) {
            $action = self::TRIGGER_ACTION_CREATE;
        } else if ($oldData && !(Request::METHOD_DELETE == $method)) {
            $action = self::TRIGGER_ACTION_UPDATE;
        } else {
            return;
        }

        /* @var  RuleRepository */
        $ruleRepository = $this->entityManager->getRepository(Rule::class);
        $rules = $ruleRepository->getRulesByTrigger($when, $action, $triggerResource);
        if (empty($rules)) {
            return;
        }

        $newDataNormalized = json_decode($event->getControllerResult(), true);
        $arguments = $this->workflowHelper->prepareWorkflowRequestBody($event->getRequest(), $triggerResource, $newDataNormalized, $oldDataNormalized);

        $normalizedRules = array_map(function ($rule) use ($arguments) {
            $normalizedRule = $this->serializer->normalize(
                $rule,
                'jsonld',
                ['ignore_authorization' => true, "ignore_html_escape" => true]
            );
            $normalizedRule['args'] = $arguments;
            return $normalizedRule;
        }, $rules);

        $serializedRules = $this->serializer->encode($normalizedRules, 'json');
        $proxyClient = $this->container->get("app.proxy.client.workflow");
        $workflowRequestUri = $proxyClient->getTargetUri();
        $ruleExecutionUri = $workflowRequestUri . '/api/rules/start/';
        $request = new GuzzleRequest('POST', $ruleExecutionUri, ['Content-Type' => 'application/json', 'curl' => [CURLOPT_TCP_KEEPALIVE => 10, CURLOPT_TCP_KEEPIDLE => 10]], $serializedRules);
        $timeout = $this->container->getParameter('workflow.endpoint_timeout');
        $this->logger->info(sprintf('Waiting for up to %ds for a Sealab Rule Engine response...', $timeout));
        try {
            $response = $proxyClient->send($request, ['timeout' => $timeout]);
            $statusCode = $response->getStatusCode();
            if ($statusCode >= 400) {
                $this->logger->alert(sprintf('An error occurred while sending Rule to Sealab for execution. Returned with status: %s', $statusCode));
            } else {
                $this->logger->info(sprintf('Rule successfully sent to Sealab for execution. Returned with status: %s', $statusCode));
            }
        } catch (BadRequestHttpException | \Exception $exception) {
            $this->logger->alert('An error occurred sending data to Sealab.', ['exception' => $exception]);
        }
    }
}
