<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\Serializer\SerializerInterface;
use App\Constants\Triggers;
use App\Service\WorkflowHelper;
use App\Entity\Base\BaseEntity;
use App\Entity\Workflow\Workflow;
use App\Service\MetadataHelper;
use App\Service\UtilityService;
use DateTime;
use stdClass;
use Symfony\Component\DependencyInjection\ParameterBag\ParameterBagInterface;

class WorkflowEventSubscriber implements EventSubscriberInterface
{
    const FIELD_BASED_TRIGGER = 'fieldbasedtrigger';
    const ROUTE_PARAM = 'route';
    const FIELD_OR_OPERATOR = 'OR';
    const FIELD_AND_OPERATOR = 'AND';
    const LIKE_OPERATOR = 'like';
    const NOT_LIKE_OPERATOR = 'notlike';
    const IN_ALL_OPERATOR = 'in_all';
    const NIN_ALL_OPERATOR = 'nin_all';
    const IN_OPERATOR = 'in';
    const NIN_OPERATOR = 'nin';
    const ADDED_VALUE_IS = 'addedvalueis';
    const EQUALS_OPERATOR = 'eq';
    const NOT_EQUALS_OPERATOR = 'neq';
    const GREATER_THAN_EQUALS_OPERATOR = 'gte';
    const LESS_THAN_EQUALS_OPERATOR = 'lte';
    const GREATER_THAN_OPERATOR = 'gt';
    const LESS_THAN_OPERATOR = 'lt';
    const NULL_OPERATOR = 'isnull';
    const CHANGED_OPERATOR = 'changed';
    const RESOURCE_PARAM = 'resource';

    const NON_DATETIME_CONVERT_OPERATOR = [self::NULL_OPERATOR, self::CHANGED_OPERATOR];
    // field Constants
    const FIELD = 'field';
    const OPERATOR = 'operator';
    const VALUE = 'value';
    const FIELD_TYPE = 'type';
    const WORKFLOW_PRIORITY = 2;
    /* @var  EntityManagerInterface */
    private $workflowHelper;
    private $entityManager;
    private $serializer;
    private $parameters;
    private $metadataHelper;
    private $utilityService;

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['handleTriggerType', 13]
        ];
    }

    public function __construct(
        WorkflowHelper $workflowHelper,
        EntityManagerInterface $entityManager,
        SerializerInterface $serializer,
        MetadataHelper $metadataHelper,
        ParameterBagInterface $parameters,
        UtilityService $utilityService
    ) {
        $this->workflowHelper = $workflowHelper;
        $this->entityManager = $entityManager;
        $this->serializer = $serializer;
        $this->metadataHelper = $metadataHelper;
        $this->parameters = $parameters;
        $this->utilityService = $utilityService;
    }

    public function handleTriggerType(ViewEvent $event)
    {
        $newData = $event->getRequest()->attributes->get('data');
        $oldData = $event->getRequest()->attributes->get('previous_data');
        $fsr_previous_data = $event->getRequest()->attributes->get('fsr_previous_data');
        $newDataSerialized = $event->getControllerResult();
        $method = $event->getRequest()->getMethod();
        if (Request::METHOD_GET == $method) {
            return;
        }

        if ($method == Request::METHOD_DELETE) {
            if (is_null($oldData) || $oldData instanceof Workflow) {
                return;
            }
        } else {
            if (is_null($newData) || !($newData instanceof BaseEntity) || ($newData instanceof Workflow)) {
                return;
            }
        }
        $subRequestType = $event->getRequest()->attributes->get('fsr_subrequest_type');
        if ($subRequestType == 'agnostic_bulk' || $subRequestType == 'import') {
            // type agnostic and import need not call workflow
            return;
        }
        $resourceClassName = is_null($newData) ? get_class($oldData) : get_class($newData);
        $triggerResource = $this->metadataHelper->getTypeFromClassName($resourceClassName);

        if (is_null($oldData) && Request::METHOD_POST == $method) {
            $triggerType = Triggers::POST_CREATE;
        } else if ($oldData && !(Request::METHOD_DELETE == $method)) {
            $triggerType = Triggers::POST_UPDATE;
        } else if (Request::METHOD_DELETE == $method) {
            $triggerType = Triggers::POST_DELETE;
        }
        if (is_null($triggerType)) {
            return;
        }

        $owners = [];
        if (method_exists($newData, 'getOwners') && $newData->getOwners() != null) {
            foreach ($newData->getOwners() as $owner) {
                array_push($owners, $owner->getUuid());
            }
        }

        $oldDataNormalized = $fsr_previous_data;

        /* @var  WorkflowRepository */
        $workflowRepository = $this->entityManager->getRepository(Workflow::class);
        $workflows = $workflowRepository->getWorkflowsForTriggerType($triggerType, $owners, $triggerResource);

        if (count($workflows) == 0) {
            return;
        }

        $newDataNormalized = null;
        if ($triggerType == Triggers::POST_CREATE || $triggerType == Triggers::POST_UPDATE) {
            $newDataNormalized = json_decode($newDataSerialized, true);
        } else if ($triggerType == Triggers::POST_DELETE) {
            $newDataNormalized =  $oldDataNormalized;
        }

        $applicableWorkflows = [];
        foreach ($workflows as $workflow) {
            $triggerStep = $workflow->getTriggerStep();
            $arguments = $triggerStep->getArguments();

            # TODO: needs to be handled for API trigger
            $triggerParameter = self::RESOURCE_PARAM;

            //Validate trigger step
            if (!array_key_exists($triggerParameter, $arguments)) {
                continue;
            }

            if (array_key_exists(self::FIELD_BASED_TRIGGER, $arguments)) {
                $max_relation_count = 100;
                if ($this->parameters->has('max_relation_count')) {
                    $max_relation_count = $this->parameters->get('max_relation_count', 100);
                }
                if (!$this->evaluateTriggerFields($arguments, $newDataNormalized, $oldDataNormalized, $max_relation_count)) {
                    continue;
                }
            }
            //Normalize workflow
            //TODO: check if this normalization has any fields different from old
            $normalizedBody = $this->workflowHelper->normalize_workflow_jsonld($workflow);
            //Update arguments on normalized workflow so they are not persisted
            unset($arguments[$triggerParameter]);
            array_push($applicableWorkflows, $normalizedBody);
        }

        if (sizeof($applicableWorkflows) > 0) {
            $headerVars = $event->getRequest()->headers->all();
            $requestBody = $this->workflowHelper->prepareWorkflowRequestBody($event->getRequest(), $triggerResource, $newDataNormalized, $oldDataNormalized);
            $isBulkTrigger = $event->getRequest()->attributes->get('fsr_subrequest_type') === 'bulk' ? true: false;
            $allWorkflows = [];
            foreach ($applicableWorkflows as $applicableWorkflow) {
                $payload = $this->workflowHelper->prepareWorkflowTriggerPayload($triggerType, $requestBody, $applicableWorkflow, $headerVars);
                $data = $payload['payload'];
                $headers = $payload['headers'];
                if ($isBulkTrigger){
                    array_push($allWorkflows, $data);
                } else {
                    $eventContext = ['workflowId' => $workflow->getUuid()];
                    $this->workflowHelper->invokeWorkflowEndpoint($data, $headers, $eventContext);
                }
            }
            if ($isBulkTrigger){
                $event->getRequest()->attributes->set('fsr_workflow_triggers', $allWorkflows);
            }
        }
    }

    private function evaluateTriggerFields($arguments, $newData, $previousData, $max_relation_count)
    {
        return $this->evaluateObject($arguments[self::FIELD_BASED_TRIGGER], $newData, $previousData, $max_relation_count);
    }

    /**
     * @param queryObject
     * @param $currentData
     * @param $previousData
     * @return boolean
     */
    private function evaluateObject($queryObject, $currentData, $previousData, $max_relation_count)
    {
        $fieldExpressions = $queryObject['filters'];
        $fieldOperator = $queryObject['logic'] ? $queryObject['logic'] : self::FIELD_AND_OPERATOR;
        $evaluation = false;
        if ($fieldOperator === self::FIELD_AND_OPERATOR) {
            $evaluation = true;
        }
        foreach ($fieldExpressions as $fieldExpression) {
            if (array_key_exists('filters', $fieldExpression)) {
                $currEval = $this->evaluateObject($fieldExpression, $currentData, $previousData, $max_relation_count);
                $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                continue;
            }
            if ($currentData && !array_key_exists($fieldExpression[self::FIELD], $currentData) && !array_key_exists($fieldExpression[self::FIELD], $previousData)) {
                // This will be the case for M2M, For performance optimization, if M2M fields is not changed, it is not normalized.
                // M2M only supports 'isChanged' and 'added' operator. So field not present means there has been no change,
                // and condition should evaluate to False
                if ($fieldOperator === self::FIELD_AND_OPERATOR) {
                    return false;
                }
                continue;
            }
            $currDataVal = $currentData[$fieldExpression[self::FIELD]];
            $fieldExprVal = $fieldExpression[self::VALUE];
            $fieldType = isset($fieldExpression[self::FIELD_TYPE]) ? $fieldExpression[self::FIELD_TYPE] : null;
            $fieldExpressionOperator = $fieldExpression['operator'];

            if ($fieldType == 'datetime' && !in_array($fieldExpressionOperator, self::NON_DATETIME_CONVERT_OPERATOR)) {
                $dateFieldResult = $this->utilityService->parseDateField($fieldExprVal);
                $fieldExprVal = $dateFieldResult ? $dateFieldResult->getTimestamp() : $dateFieldResult;
            }
            if ($fieldExpression[self::FIELD] == 'recordTags') {
                $tags = [];
                foreach ($fieldExprVal as $item) {
                    $exploded_item = explode('/', $item);
                    array_push($tags, end($exploded_item));
                }
                $fieldExprVal = $tags;
            }
            $prevDataVal = $previousData[$fieldExpression[self::FIELD]];
            if (is_array($currDataVal) && isset($currDataVal['@id'])) {
                $currDataVal = $currDataVal['@id'];
            } else if (is_array($currDataVal)) {
                $finalVal = [];
                foreach ($currDataVal as $val) {
                    $item = isset($val['@id']) ? $val['@id'] : $val;
                    array_push($finalVal, $item);
                }
                $currDataVal = $finalVal;
            }
            if (is_array($prevDataVal) && isset($prevDataVal['@id'])) {
                $prevDataVal = $prevDataVal['@id'];
            } else if (is_array($prevDataVal)) {
                $finalVal = [];
                foreach ($prevDataVal as $val) {
                    $item = isset($val['@id']) ? $val['@id'] : $val;
                    array_push($finalVal, $item);
                }
                $prevDataVal = $finalVal;
            }
            // Backwards compatibility after tags operator change
            // TODO: Remove after a few releases
            if ($fieldExpression['operator'] == self::IN_OPERATOR && !$fieldExpression['useInOperator']) {
                $fieldExpression['operator'] = self::IN_ALL_OPERATOR;
            }
            switch ($fieldExpression['operator']) {

                case self::EQUALS_OPERATOR:
                    $currEval = strtolower($currDataVal) == strtolower($fieldExprVal);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NOT_EQUALS_OPERATOR:
                    $currEval = strtolower($currDataVal) !== strtolower($fieldExprVal);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::LIKE_OPERATOR:
                    $currEval = false;
                    $fieldExprVal = preg_quote($fieldExprVal);
                    $expression = str_replace('%', '.*', $fieldExprVal);
                    $expression = str_replace('_', '.{1}', $expression);
                    $regex = sprintf('/^%s$/mi', $expression);
                    preg_match_all($regex, $currDataVal, $matches, PREG_SET_ORDER, 0);
                    $currEval = count($matches) > 0;
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NOT_LIKE_OPERATOR:
                    $currEval = false;
                    $fieldExprVal = preg_quote($fieldExprVal);
                    $expression = str_replace('%', '.*', $fieldExprVal);
                    $expression = str_replace('_', '.{1}', $expression);
                    $regex = sprintf('/%s/mi', $expression);
                    preg_match_all($regex, $currDataVal, $matches, PREG_SET_ORDER, 0);
                    $currEval = count($matches) == 0;
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NULL_OPERATOR:
                    $currEval = (is_null($currDataVal) ? 'true' : 'false') === $fieldExprVal;
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::IN_ALL_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = $this->in_array($fieldExprVal, $currDataVal, true);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NIN_ALL_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = $this->nin_array($fieldExprVal, $currDataVal, true);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::IN_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = $this->in_array($fieldExprVal, $currDataVal);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NIN_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = $this->nin_array($fieldExprVal, $currDataVal);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::CHANGED_OPERATOR:
                    if (is_array($currDataVal) || is_array($prevDataVal)) {
                        $prevDataVal = $prevDataVal ? $prevDataVal : [];
                        # no of related records is large. since the key was present in data, detect as changed
                        if (
                            !is_null($currDataVal) && count($currDataVal) >= $max_relation_count &&
                            !is_null($prevDataVal) && count($prevDataVal) >= $max_relation_count
                        ) {
                            $currEval = true;
                        }
                        if (!is_null($currDataVal)) {
                            sort($currDataVal);
                            sort($prevDataVal);
                            $currEval = $currDataVal !== $prevDataVal;
                        } else {
                            $currEval = !is_null($prevDataVal);
                        }
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                        break;
                    } else {
                        $currEval = strtolower($currDataVal) !== strtolower($prevDataVal);
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                        break;
                    }
                case self::ADDED_VALUE_IS:
                    if (is_array($currDataVal)) {
                        $prevDataVal = $prevDataVal ? $prevDataVal : [];
                        $fieldExprVal = is_array($fieldExprVal) ? $fieldExprVal : [$fieldExprVal];
                        $currentMatch = array_intersect($fieldExprVal, $currDataVal);
                        $prevMatch = array_intersect($fieldExprVal, $prevDataVal);
                        if (count($currentMatch) > 0 && count($prevMatch) == 0) {
                            $currEval = true;
                        } else {
                            $currEval = false;
                        }
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                        break;
                    } else {
                        $evaluation = false;
                        break;
                    }
                case self::GREATER_THAN_EQUALS_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal >= $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                case self::LESS_THAN_EQUALS_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal <= $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                case self::GREATER_THAN_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal > $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                case self::LESS_THAN_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal < $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                default:
                    return $this->evaluateOperator($fieldOperator, false, $evaluation);
            }
        }
        return $evaluation;
    }

    /**
     * @param $operator
     * @param $prevEval
     * @param $currEval
     * @return boolean
     */
    private function evaluateOperator($operator, $prevEval, $currEval)
    {
        if ($operator === self::FIELD_OR_OPERATOR) {
            return $prevEval || $currEval;
        } elseif ($operator === self::FIELD_AND_OPERATOR) {
            return $prevEval && $currEval;
        } else {
            return false;
        }
    }

    

    public function in_array($needles, $haystack, $all = false)
    {
        if (!is_array($haystack)) {
            $haystack = [$haystack];
        }
        $diff = array_udiff($needles, $haystack, 'strcasecmp');
        return $all ? empty($diff) : count($diff) < count($needles);
    }

    public function nin_array($needles, $haystack, $all = false)
    {
        if (!is_array($haystack)) {
            $haystack = [$haystack];
        }
        $diff = array_uintersect($needles, $haystack, 'strcasecmp');
        return $all ? count($diff) > 0 : count($diff) == count($needles);
    }
    
}
