<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Serializer\Normalizer;

use ApiPlatform\Core\Api\IriConverterInterface;
use ApiPlatform\Core\Api\ResourceClassResolverInterface;
use ApiPlatform\Core\DataProvider\ItemDataProviderInterface;
use ApiPlatform\Core\Exception\InvalidArgumentException;
use ApiPlatform\Core\Exception\InvalidValueException;
use ApiPlatform\Core\Exception\ItemNotFoundException;
use ApiPlatform\Core\Metadata\Property\Factory\PropertyMetadataFactoryInterface;
use ApiPlatform\Core\Metadata\Property\Factory\PropertyNameCollectionFactoryInterface;
use ApiPlatform\Core\Metadata\Property\PropertyMetadata;
use ApiPlatform\Core\Metadata\Resource\Factory\ResourceMetadataFactoryInterface;
use ApiPlatform\Core\Serializer\AbstractItemNormalizer;
use App\Providers\PermissionsDictionaryBuilder;
use App\Security\Voter\FsrVoter;
use App\Service\MetadataHelper;
use App\Service\NormalizerUtilityService;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;
use Symfony\Component\Serializer\Exception\LogicException;
use Symfony\Component\Serializer\Exception\NotNormalizableValueException;
use Symfony\Component\Serializer\Exception\UnexpectedValueException;
use Symfony\Component\Serializer\Mapping\Factory\ClassMetadataFactoryInterface;
use Symfony\Component\Serializer\NameConverter\NameConverterInterface;
use Symfony\Component\Serializer\Normalizer\DenormalizerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\PropertyInfo\Type;

class EfficientItemDeNormalizer extends AbstractItemNormalizer implements NormalizerInterface, DenormalizerInterface
{

    protected $iriConverter;
    protected $classMetadata;
    protected $fsrVoter;
    protected $request;
    protected $normalizerUtilityService;
    protected $containers;
    protected $entityManager;
    protected $metadataHelper;
    protected $requestStack;

    private const REQUIREDATTRIBUTE = array("@id", "@context", "@type");
    private const SELF_TENANT = "/api/3/tenants/b3a700f7-00be-4ef9-90c6-3c8fe6e1be63";

    public function __construct(
        PropertyNameCollectionFactoryInterface $propertyNameCollectionFactory,
        PropertyMetadataFactoryInterface $propertyMetadataFactory,
        IriConverterInterface $iriConverter,
        ResourceClassResolverInterface $resourceClassResolver,
        PropertyAccessorInterface $propertyAccessor = null,
        NameConverterInterface $nameConverter = null,
        ClassMetadataFactoryInterface $classMetadataFactory = null,
        ItemDataProviderInterface $itemDataProvider = null,
        bool $allowPlainIdentifiers = false,
        array $defaultContext = [],
        iterable $dataTransformers = [],
        ResourceMetadataFactoryInterface $resourceMetadataFactory = null,
        ContainerInterface $containers,
        FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        RequestStack $requestStack,
        NormalizerUtilityService $normalizerUtilityService,
        MetadataHelper $metadataHelper
    ) {
        parent::__construct($propertyNameCollectionFactory, $propertyMetadataFactory, $iriConverter, $resourceClassResolver, $propertyAccessor, $nameConverter, $classMetadataFactory, null, false, $defaultContext, $dataTransformers, $resourceMetadataFactory);
        $this->iriConverter = $iriConverter;
        $this->classMetadata = $classMetadataFactory;
        $this->fsrVoter = $fsrVoter;
        $this->requestStack = $requestStack;
        $this->normalizerUtilityService = $normalizerUtilityService;
        $this->containers = $containers;
        $this->entityManager = $entityManager;
        $this->itemDataProvider = $itemDataProvider;
        $this->metadataHelper = $metadataHelper;
    }

    const FORMAT = 'jsonld';

    /**
     * {@inheritdoc}
     */
    public function supportsDenormalization($data, $type, $format = null): bool
    {
        return self::FORMAT === $format && parent::supportsDenormalization($data, $type, $format);
    }

    public function supportsNormalization($data, $format = null)
    {
        return false;
    }

    /**
     * {@inheritdoc}
     *
     * @throws NotNormalizableValueException
     */
    public function denormalize($data, $class, $format = null, array $context = [])
    {
        // Avoid issues with proxies if we populated the object
        if (isset($data['@id']) && !isset($context[self::OBJECT_TO_POPULATE])) {
            if (true !== ($context['api_allow_update'] ?? true)) {
                throw new NotNormalizableValueException('Update is not allowed for this operation.');
            }

            $context[self::OBJECT_TO_POPULATE] = $this->iriConverter->getItemFromIri($data['@id'], $context + ['fetch_data' => true]);
        }
        // type validation is disabled by default unless specifically enabled in context
        if (!isset($context[self::DISABLE_TYPE_ENFORCEMENT])) {
            $context[self::DISABLE_TYPE_ENFORCEMENT] = true;
        }
        //TODO: fix this correctly. temporary change to replace file form type objects with iris
        foreach ($data as $indexKey => $objectValue) {
            if (isset($data[$indexKey]) && is_array($data[$indexKey]) && array_key_exists('@type', $data[$indexKey]) && $data[$indexKey]['@type'] == 'File') {
                $data[$indexKey] = $data[$indexKey]['@id'];
            }
        }
        return parent::denormalize($data, $class, $format, $context);
    }

    private function supportsPlainIdentifiers(): bool
    {
        return $this->allowPlainIdentifiers && null !== $this->itemDataProvider;
    }

    /**
     * Denormalizes a relation.
     *
     * @throws LogicException
     * @throws UnexpectedValueException
     * @throws ItemNotFoundException
     *
     * @return object|null
     */
    protected function denormalizeRelation(string $attributeName, PropertyMetadata $propertyMetadata, string $className, $value, ?string $format, array $context)
    {
        $supportsPlainIdentifiers = $this->supportsPlainIdentifiers();

        if (!$propertyMetadata->getType()->isCollection() && is_array($value) && $value["@id"]) {
            return $this->iriConverter->getItemFromIri($value["@id"], $context + ['fetch_data' => false]);
        }

        if (\is_string($value)) {
            try {
                return $this->iriConverter->getItemFromIri($value, $context + ['fetch_data' => false]);
            } catch (InvalidArgumentException $e) {
                try {
                    $iri = $this->iriConverter->getIriFromResourceClass($className);
                    $valueIRI = sprintf('%s/%s', $iri, $value);
                    return $this->iriConverter->getItemFromIri($valueIRI, $context + ['fetch_data' => false]);
                } catch (InvalidArgumentException $e) {
                    return null;
                    // ignoring an invalid iri for backward compatibility
                    // throw new HttpException(Response::HTTP_BAD_REQUEST, (sprintf(
                    //     'IRI not supported (found "%s" in "%s" of "%s")',
                    //     $value,
                    //     $attributeName,
                    //     $className
                    // )), $e);
                } catch (\Exception $e) {
                    error_log(sprintf('Invalid IRI "%s".', $value));
                }
                if (!$supportsPlainIdentifiers) {
                    throw new UnexpectedValueException(sprintf('Invalid IRI "%s".', $value), $e->getCode(), $e);
                }
            } catch (\Exception $e) {
                return null;
            }
        }

        if ($propertyMetadata->isWritableLink()) {
            $context['api_allow_update'] = true;

            if (!$this->serializer instanceof DenormalizerInterface) {
                throw new LogicException(sprintf('The injected serializer must be an instance of "%s".', DenormalizerInterface::class));
            }

            try {
                $item = $this->serializer->denormalize($value, $className, $format, $context);
                if (!\is_object($item) && null !== $item) {
                    throw new \UnexpectedValueException('Expected item to be an object or null.');
                }

                return $item;
            } catch (InvalidValueException $e) {
                if (!$supportsPlainIdentifiers) {
                    throw $e;
                }
            } catch (\Exception $e) {
                if (!$supportsPlainIdentifiers) {
                    throw $e;
                }
            }
        }

        if (!\is_array($value)) {
            if (!$supportsPlainIdentifiers) {
                throw new UnexpectedValueException(sprintf('Expected IRI or nested document for attribute "%s", "%s" given.', $attributeName, \gettype($value)));
            }

            // $item = $this->iriConverter->getItemFromIri($value, $context + ['fetch_data' => true]);

            $item = $this->itemDataProvider->getItem($className, $value, null, $context + ['fetch_data' => true]);
            if (null === $item) {
                throw new ItemNotFoundException(sprintf('Item not found for resource "%s" with id "%s".', $className, $value));
            }

            return $item;
        }

        throw new UnexpectedValueException(sprintf('Nested documents for attribute "%s" are not allowed. Use IRIs instead.', $attributeName));
    }

    protected function getAllowedAttributes($classOrObject, array $context, $attributesAsString = false)
    {
        // get_class($classOrObject) returns false in case if its an class not an object
        // Adding ternary here else if returns false
        // purpose is to check if object retrieve its class else directly consume the class
        $entity = is_string($classOrObject) ? $classOrObject : get_class($classOrObject);
        $module = $this->metadataHelper->getModuleFromClassName($entity);
        $allowedAttributes = [];
        $currentRequest = $this->requestStack->getCurrentRequest();

        $is_update = false;
        if ((isset($context['_is_update']) && $context['_is_update'] == true) || ($currentRequest && $currentRequest->getMethod() == 'PUT')) {
            $is_update = true;
        }
        $context['ignore_authorization'] ?? $context['ignore_authorization'] = false;
        $context['ignore_field_authorization'] ?? $context['ignore_field_authorization'] = false;
        $context['ignore_relationship_authorization'] ?? $context['ignore_relationship_authorization'] = false;
        $options = $this->getFactoryOptions($context);
        $propertyNames = $this->propertyNameCollectionFactory->create($context['resource_class'], $options);
        $prev_data = array();
        if ($currentRequest) {
            $prev_data = $currentRequest->attributes->get('fsr_previous_data');
        }
        foreach ($propertyNames as $attributeName => $attributeMetadata) {
            $propertyMetadata = $this->propertyMetadataFactory->create($classOrObject, $attributeMetadata);
            $attribute_prev_value = $prev_data[$attributeMetadata] ?? null;
            if ($propertyMetadata->isWritable() || !$is_update
                    || ($attribute_prev_value == null || ($attributeMetadata == "tenant" && $attribute_prev_value == self::SELF_TENANT))) {
                if (
                    $context['ignore_authorization']
                    || $context['ignore_field_authorization']
                    || $context['ignore_relationship_authorization']
                    || $this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_FIELD_WRITE . '.' . $module . '.' . $attributeMetadata)
                ) {
                    $allowedAttributes[] = $attributeMetadata;
                }
            }
        }
        $allowedAttributes = array_merge($allowedAttributes, self::REQUIREDATTRIBUTE);
        return $allowedAttributes;
    }

    /**
     * Denormalizes a collection of objects.
     *
     * @throws InvalidArgumentException
     */
    protected function denormalizeCollection(string $attribute, PropertyMetadata $propertyMetadata, Type $type, string $className, $value, ?string $format, array $context): array
    {
        if (is_null($value)) {
            return [];
        }

        if (is_string($value) && strlen(trim($value)) === 0) {
            return [];
        }

        if (!\is_array($value)) {
            $value = explode(",", $value);
        }

        return parent::denormalizeCollection($attribute, $propertyMetadata, $type, $className, $value, $format, $context);
    }
}
