<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Providers\PermissionsDictionaryBuilder;
use App\Service\ConfigExportImport\PlaybookConfig;
use App\Security\Voter\FsrVoter;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Exception\RequestException;
use Symfony\Component\Filesystem\Exception\IOException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * Class ConnectorConfig
 * @package App\Service\ConfigExportImport
 */
class ConnectorConfig extends PortingConfig
{
    /** @var fsrVoter */
    private $fsrVoter;

    /** @var PlaybookConfig */
    private $playbookConfig;
    /**@var ImportService */
    
    private $importService;

    public function __construct(
        FsrVoter $fsrVoter,
        ImportService $importService,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        PlaybookConfig $playbookConfig
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->importService = $importService;
        $this->playbookConfig = $playbookConfig;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container);
    }

    /**
     *
     * @param array $data Playbook object containing collections & global vars
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        // at least one of these permissions
        if (
            (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS) &&
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS) &&
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS)) ||
            count($data) == 0
        ) {
            return [];
        }
        $options = [];
        $existingConnectors = $this->getConnectors($tokenType);
        $existingConnectorNames = array_column($existingConnectors, 'name');
        $connectorsInRepo = $this->loadConnectorRpms();
        $options = [];
        foreach ($data as $connector) {
            $existing = in_array($connector['name'], $existingConnectorNames);
            $individualOptions = [];
            $individualOptions['name'] = $connector['name'];
            $individualOptions['label'] = $connector['label'];
            $individualOptions['version'] = $connector['version'];
            $individualOptions['rpm_name'] = $connector['rpm_name'];
            $individualOptions['configurationCount'] = count($connector['configurations']);
            if (count($connector['configurations']) > 0) {
                $individualOptions['includeConfigurations'] = true;
            }
            if (array_key_exists('install_mode', $connector) && array_key_exists('installer_path', $connector)) {
                $individualOptions['install_mode'] = $connector['install_mode'];
                $individualOptions['installer_path'] = $connector['installer_path'];
            }
            $individualOptions['exists'] = $existing;
            if ($existing) {
                foreach ($existingConnectors as $existingConnector) {
                    if ($existingConnector['name'] == $connector['name']) {
                        $individualOptions['existingId'] = $existingConnector['id'];
                        $existingVersion = str_replace('.', '', $existingConnector['version']);
                        $version = str_replace('.', '', $connector['version']);
                        if ($existingVersion >= $version) {
                            $individualOptions['includeInstall'] = false;
                            $individualOptions['locked'] = true;
                        } else {
                            $individualOptions['includeInstall'] = !!$connector['rpm_name'];
                            $individualOptions['existingVersion'] = $existingConnector['version'];
                        }
                        break;
                    }
                }
            } else if ($connector['rpm_name'] || array_key_exists('install_mode', $connector)) {
                $individualOptions['includeInstall'] = true;
            }
            if ($individualOptions['includeInstall']) {
                $fullKey = sprintf('%s_%s', $connector['name'], $connector['version']);
                $individualOptions['rpmExists'] = array_key_exists($fullKey, $connectorsInRepo);
            }

            $options[] = $individualOptions;
        }
        return $options;

        return $options;
    }

    private function getConnectors($tokenType = 'JWT')
    {
        $token = $this->getTokenFromStorage();
        $endpoint = '/connector_details/?ordering=label&page_size=10000&$export=true';
        if ($token) {
            $uri = sprintf('%s/api/integration%s', $this->container->getParameter('crudhub.uri'), $endpoint);
            $response = $this->sendRequest($uri, 'post', null, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('data', $jsonDecodeResponse) ? $jsonDecodeResponse['data'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('integrations.uri') . $endpoint;
            return $this->sendIntegrationRequest($uri, 'post');
        }
    }

    private function getDataImports($tokenType = 'JWT', $configurationUuid)
    {
        $token = $this->getTokenFromStorage();
        $endpoint = '/data-import/?configuration=' . $configurationUuid;
        if ($token) {
            $uri = sprintf('%s/api/integration%s', $this->container->getParameter('crudhub.uri'), $endpoint);
            $response = $this->sendRequest($uri, 'get', null, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('data', $jsonDecodeResponse) ? $jsonDecodeResponse['data'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('integrations.uri') . $endpoint;
            return $this->sendIntegrationRequest($uri, 'get');
        }
    }


    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType)
    {
        $token = $this->getTokenFromStorage();
        foreach ($options as $option) {
            if (!$option['include']) {
                continue;
            }
            foreach ($data as $connector) {
                if ($connector['rpm_name'] === $option['rpm_name']) {
                    $connectorId = array_key_exists('existingId', $option) ? $option['existingId'] : null;
                    // Install & wait if not exists
                    if ($option['includeInstall']) {
                        $result = $this->installConnectorRpm($connector, $tokenType);
                        if (!$result) {
                            // Connector failed to install
                            throw new \Exception(sprintf('Connector %s took longer than 1 minute to install', $connector['rpm_name']));
                        }
                        $connectorId = $result['id'];
                    } else {
                        $iri = sprintf('connectors/%s/%s/?format=json', $connector['name'], $connector['version']);
                        if ($token) {
                            $jsonData = $this->sendRequest(sprintf('%s/api/integration/%s', $this->container->getParameter('crudhub.uri'), $iri), 'post', $connector, $tokenType);
                        } else {
                            $jsonData = $this->sendIntegrationRequest(sprintf('%s/%s', $this->container->getParameter('integrations.uri'), $iri), 'post', $connector);
                        }
                        $result = json_decode($jsonData, true);
                    }
                    // Now, update the roles for operations
                    if (array_key_exists('operation_roles', $connector)) {
                        $this->updateOperationRoles($connector['operation_roles'], $result['operations'], $tokenType);
                    }

                    if (!$option['includeConfigurations']) {
                        continue;
                    }
                    // First, delete existing configs
                    $this->deleteExistingConfigurations($connector['configurations'], $result['configuration']);

                    // Next, ingestion playbooks need to go in
                    $this->importIngestionPlaybooks($connector['ingestionPlaybooks'], $tokenType);

                    // Install configurations afterwards
                    $this->importConfigurations($connector['configurations'], $connectorId, $tokenType);

                    if (!array_key_exists('dataImports', $connector)) {
                        continue;
                    }

                    $this->importDataImports($connector['dataImports'], $tokenType);
                }
            }
        }
        return true;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType)
    {
        $connectorPath = $tmpDirectory . $key . '/';
        $defaultJsonPath = $connectorPath . 'data.json';
        $fileData = file_get_contents($defaultJsonPath);
        $data = json_decode($fileData, true);

        $token = $this->getTokenFromStorage();
        foreach ($options as $option) {
            if (!$option['include']) {
                continue;
            }
            foreach ($data as $connector) {
                if ($connector['install_mode'] == $option['install_mode'] && $connector['installer_path'] == $option['installer_path']) {
                    $connectorId = array_key_exists('existingId', $option) ? $option['existingId'] : null;
                    // Install & wait if not exists
                    if ($option['includeInstall'] && $connector['install_mode'] == 'rpm' && $option['rpmExists']) {
                        $result = $this->installConnectorRpm($connector, $tokenType);
                        if (!$result) {
                            // Connector failed to install
                            throw new \Exception(sprintf('Connector %s took longer than 1 minute to install', $connector['installer_path']));
                        }
                        $connectorId = $result['id'];
                    } else if ($option['includeInstall'] && $connector['install_mode'] == 'tgz') {
                        $fileParts = explode('.', $connector['installer_path']);
                        $uri = sprintf('%s/import-connector/%s/?replace=true&format=json', $this->container->getParameter('integrations.uri'), $fileParts[0]);
                        $fileStream = fopen($connectorPath . $option['installer_path'], 'rb');
                        $token = $this->tokenStorage->getToken();
                        $authorization = $token ? sprintf("Bearer %s", $token->getCredentials()) : null;
                        $client = $this->container->get("app.proxy.client.integrations");
                        $params = [
                            'verify' => false,
                            'body' => $fileStream,
                            'headers' => [
                                'Authorization' => $authorization
                            ]
                        ];
                        try {
                            $integrationResponse = $client->request('POST', $uri, $params);
                        } catch (IOException | \Exception $e) {
                            $this->io->warning(sprintf('Skipping import for %s (from file %) since the file is invalid or incompatible.', $option['name'], $option['installer_path']));
                            continue;
                        }
                        $jsonResponse = json_decode($integrationResponse->getBody()->getContents(), true);
                        $connectorId = $jsonResponse['id'];
                    } else {
                        $iri = sprintf('connectors/%s/%s/?format=json', $connector['name'], $connector['version']);
                        if ($token) {
                            $jsonData = $this->sendRequest(sprintf('%s/api/integration/%s', $this->container->getParameter('crudhub.uri'), $iri), 'post', $connector, $tokenType);
                        } else {
                            $jsonData = $this->sendIntegrationRequest(sprintf('%s/%s', $this->container->getParameter('integrations.uri'), $iri), 'post', $connector);
                        }
                        $result = json_decode($jsonData, true);
                    }
                    // Now, update the roles for operations
                    if (array_key_exists('operation_roles', $connector)) {
                        $this->updateOperationRoles($connector['operation_roles'], $result['operations'], $tokenType);
                    }

                    if (!$option['includeConfigurations']) {
                        continue;
                    }
                    // First, delete existing configs
                    $this->deleteExistingConfigurations($connector['configurations'], $result['configuration']);

                    // Next, ingestion playbooks need to go in
                    $this->importIngestionPlaybooks($connector['ingestionPlaybooks'], $tokenType);

                    // Install configurations afterwards
                    $this->importConfigurations($connector['configurations'], $connectorId, $tokenType);

                    if (!array_key_exists('dataImports', $connector)) {
                        continue;
                    }

                    $this->importDataImports($connector['dataImports'], $tokenType);
                }
            }
        }
        return true;
    }

    private function installConnectorRpm($connector, $tokenType) {
        $token = $this->getTokenFromStorage();
        if ($token) {
            $this->sendRequest(sprintf('%s/api/integration/install-connector/?format=json', $this->container->getParameter('crudhub.uri')), 'post', $connector, $tokenType);
        } else {
            $this->sendIntegrationRequest(sprintf('%s/install-connector/?format=json', $this->container->getParameter('integrations.uri')), 'post', $connector);
        }
        $status = 'In-Progress';
        $dep_installed = 'In-Progress';
        $counter = 0;
        while (($status == 'In-Progress' || $dep_installed == 'In-Progress') && $counter < 15) {
            sleep(4);
            $counter++;
            $iri = sprintf('connectors/%s/%s/?format=json', $connector['name'], $connector['version']);
            if ($token) {
                $jsonData = $this->sendRequest(sprintf('%s/api/integration/%s', $this->container->getParameter('crudhub.uri'), $iri), 'post', $connector, $tokenType);
            } else {
                $jsonData = $this->sendIntegrationRequest(sprintf('%s/%s', $this->container->getParameter('integrations.uri'), $iri), 'post', $connector);
            }
            $result = json_decode($jsonData, true);
            $status = $result['status'];
            $dep_installed = $result['requirements_installed'];
            if ($status == 'Failed') {
                // Connector failed to install
                throw new \Exception(sprintf('Connector %s failed to install. Please check connectors.log for more information.', $connector['rpm_name']));
            }
            if ($dep_installed == 'Failed') {
                // Connector failed to install
                $this->io->warning(sprintf('Connector %s installed, but dependencies failed to install. Please run dependency installation manually or check connectors.log for more information.', $connector['rpm_name']));
                return $result;
            }
        }

        return $counter == 15 ? false : $result;
    }

    private function updateOperationRoles($connectorOperationRoles, $operationRoles, $tokenType) {
        $token = $this->getTokenFromStorage();
        foreach ($operationRoles as $operation) {
            if (array_key_exists($operation['operation'], $connectorOperationRoles)) {
                $payload = [
                    'roles' => $connectorOperationRoles[$operation['operation']]
                ];
                $uri = sprintf('%s/api/integration/connectors/operations/%s/roles/?format=json', $this->container->getParameter('crudhub.uri'), $operation['id']);
                if ($token) {
                    $this->sendRequest($uri, 'post', $payload, $tokenType);
                } else {
                    $this->sendIntegrationRequest(str_replace('/api/integration/', '', $uri), 'post', $payload);
                }
            }
        }
    }

    private function deleteExistingConfigurations($configurations, $existingConfigurations) {
        $token = $this->getTokenFromStorage();
        foreach ($configurations as $configuration) {
            foreach ($existingConfigurations as $existingConfiguration) {
                if ($existingConfiguration['config_id'] == $configuration['config_id'] || $existingConfiguration['name'] == $configuration['name']) {
                    // Delete existing config
                    if ($token) {
                        $this->sendRequest(sprintf('%s/api/integration/configuration/%s/?format=json', $this->container->getParameter('crudhub.uri'), $existingConfiguration['id']), 'delete', null, $tokenType);
                    } else {
                        $this->sendIntegrationRequest(sprintf('%s/configuration/%s/?format=json', $this->container->getParameter('integrations.uri'), $existingConfiguration['id']), 'delete', null);
                    }
                    break;
                }
            }
        }
    }

    private function importDataImports($dataImports, $tokenType) {
        $token = $this->getTokenFromStorage();
        foreach ($dataImports as $dataImport) {
            try {
                if ($token) {
                    $scheduleId = $this->saveSchedule($dataImport['schedule'], $tokenType);
                    $dataImport['metadata']['scheduleId'] = $scheduleId;
                    $this->sendRequest(sprintf('%s/api/integration/data-import/?format=json', $this->container->getParameter('crudhub.uri')), 'post', $dataImport, $tokenType);
                } else {
                    $this->sendIntegrationRequest(sprintf('%s/data-import/?format=json', $this->container->getParameter('integrations.uri')), 'post', $dataImport);
                }
            } catch (BadRequestHttpException | \Exception $exception) {
                // Data import already exists, ignore
                continue;
            }
        }
    }

    private function importConfigurations($configurations, $connectorId, $tokenType) {
        $token = $this->getTokenFromStorage();
        foreach ($configurations as $configuration) {
            $configuration['connector'] = $connectorId;
            if (array_key_exists('agent', $configuration)) {
                unset($configuration['agent']);
            }
            try {
                if ($token) {
                    $this->sendRequest(sprintf('%s/api/integration/configuration/?format=json', $this->container->getParameter('crudhub.uri')), 'post', $configuration, $tokenType);
                } else {
                    $this->sendIntegrationRequest(sprintf('%s/configuration/?format=json', $this->container->getParameter('integrations.uri')), 'post', $configuration);
                }
            } catch (BadRequestHttpException | \Exception $exception) {
                // Config already exists, ignore
                continue;
            }
        }
    }

    private function importIngestionPlaybooks($ingestionPlaybooks, $tokenType) {
        if ($ingestionPlaybooks) {
            $allData = [
                'playbooks' => $ingestionPlaybooks
            ];
            $playbookOptions = $this->playbookConfig->getImportOptions($allData, 'playbooks', $tokenType);
            $this->playbookConfig->import($allData['playbooks'], $playbookOptions, $tokenType);
        }
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS)) {
            return [];
        }
        $data = [];
        if (count($options) > 0) {
            $connectors = $this->getConnectors($tokenType);
        }
        foreach ($options as $option) {
            foreach ($connectors as $connector) {
                // Load ingestion playbooks
                $rpmName = sprintf('cyops-connector-%s-%s', $connector['name'], $connector['version']);
                if ($option['value'] == $rpmName) {

                    $connectorObject = [
                        'name' => $connector['name'],
                        'label' => $connector['label'],
                        'category' => $connector['category'],
                        'description' => $connector['description'],
                        'publisher' => $connector['publisher'],
                        'version' => $connector['version'],
                        'operation_roles' => [],
                        'rpm_name' => null,
                        'configurations' => [],
                        'dataImports' => []
                    ];
                    foreach ($connector['operations'] as $operation) {
                        if (array_key_exists('roles', $operation)) {
                            $connectorObject['operation_roles'][$operation['operation']] = $operation['roles'];
                        }
                    }
                    if ($option['rpm']) {
                        $connectorObject['rpm_name'] = $rpmName;
                    }
                    if ($option['configurations']) {
                        $connectorObject['configurations'] = $connector['configuration'];
                        if (count($connector['configuration']) > 0) {
                            $configIds = array_column($connector['configuration'], 'config_id');
                            $options = [
                                'collections' => $configIds,
                                'globalVariables' => []
                            ];
                            $ingestionPlaybooks = $this->playbookConfig->export($options, $tokenType);
                            if ($ingestionPlaybooks) {
                                $connectorObject['ingestionPlaybooks'] = $ingestionPlaybooks;
                            }
                            foreach ($configIds as $configId) {
                                $configDataImports = $this->getDataImports($tokenType, $configId);
                                foreach ($configDataImports as $dataImport) {
                                    if ($dataImport['metadata']['scheduleId']) {
                                        $dataImport['schedule'] = $this->loadSchedule($dataImport['metadata']['scheduleId'], $tokenType);
                                    }
                                    $connectorObject['dataImports'][] = $dataImport;
                                }
                            }
                        }
                    }

                    $data[] = $connectorObject;
                    break;
                }
            }
        }
        return $data;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType='JWT')
    {
        if (count($options) === 0 || !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS)) {
            return null;
        }
        $connectorDirectory = $tmpDirectory . $type;
        $connectorsInRepo = $this->loadConnectorRpms();
        $token = $this->getTokenFromStorage();
        $data = [];
        $connectors = $this->getConnectors($tokenType);
        foreach ($options as $option) {
            foreach ($connectors as $connector) {
                // Load ingestion playbooks
                $rpmName = sprintf('cyops-connector-%s-%s', $connector['name'], $connector['version']);
                if ($option['value'] == $rpmName) {
                    $connectorObject = [
                        'name' => $connector['name'],
                        'label' => $connector['label'],
                        'category' => $connector['category'],
                        'description' => $connector['description'],
                        'publisher' => $connector['publisher'],
                        'version' => $connector['version'],
                        'operation_roles' => [],
                        'configurations' => [],
                        'dataImports' => []
                    ];
                    foreach ($connector['operations'] as $operation) {
                        if (array_key_exists('roles', $operation)) {
                            $connectorObject['operation_roles'][$operation['operation']] = $operation['roles'];
                        }
                    }
                    $fullKey = sprintf('%s_%s', $connector['name'], $connector['version']);
                    if (array_key_exists($fullKey, $connectorsInRepo) && $option['rpm']) {
                        $connectorObject['rpm_name'] = $rpmName;
                        $connectorObject['install_mode'] = 'rpm';
                        $connectorObject['installer_path'] = $rpmName;
                    } else {
                        $uri = sprintf('%s/api/integration/connector/development/entity/%s/export/?format=json', $this->container->getParameter('crudhub.uri'), $connector['id']);
                        if ($token) {
                            $response = $this->sendRequest($uri, 'get', [], $tokenType);
                        } else {
                            $response = $this->sendIntegrationRequest(str_replace('/api/integration/', '', $uri), 'get', []);
                        }
                        $tgzFilename = $fullKey . '.tgz';
                        $this->writeBytesToFile(sprintf('%s/%s', $connectorDirectory, $tgzFilename), $response);
                        $connectorObject['install_mode'] = 'tgz';
                        $connectorObject['installer_path'] = $tgzFilename;
                    }
                    if ($option['configurations']) {
                        $connectorObject['configurations'] = $connector['configuration'];
                        if (count($connector['configuration']) > 0) {
                            $configIds = array_column($connector['configuration'], 'config_id');
                            $options = [
                                'collections' => $configIds,
                                'globalVariables' => []
                            ];
                            $ingestionPlaybooks = $this->playbookConfig->export($options, $tokenType);
                            if ($ingestionPlaybooks) {
                                $connectorObject['ingestionPlaybooks'] = $ingestionPlaybooks;
                            }
                            foreach ($configIds as $configId) {
                                $configDataImports = $this->getDataImports($tokenType, $configId);
                                foreach ($configDataImports as $dataImport) {
                                    if ($dataImport['metadata']['scheduleId']) {
                                        $dataImport['schedule'] = $this->loadSchedule($dataImport['metadata']['scheduleId'], $tokenType);
                                    }
                                    $connectorObject['dataImports'][] = $dataImport;
                                }
                            }
                        }
                    }

                    $data[] = $connectorObject;
                    break;
                }
            }
        }
        $filename = $connectorDirectory . '/data.json';
        $this->writeToFile($filename, $data);
        return $filename;
    }

    private function loadConnectorRpms() {
        try {
            $jsonUri = sprintf('https://%s/connectors/info/connectors-all.json', getenv("REPOSERVER"));
            $response = $this->sendRequest($jsonUri, 'GET', null, null, ['connect_timeout' => 15]);
            return json_decode($response, true);
        } catch (RequestException | \Exception $e) {
            $this->io->warning(sprintf('Cannot load connector RPMs from %s: %s', getenv("REPOSERVER"), $e->getMessage()));
            return [];
        }
    }
}
