<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Authorization\Role;
use App\Entity\ViewTemplate\Dashboard;
use App\Providers\PermissionsDictionaryBuilder;
use App\Security\Voter\FsrVoter;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * Class DashboardConfig
 * @package App\Service\ConfigExportImport
 */
class DashboardConfig extends PortingConfig
{
    /** @var fsrVoter */
    private $fsrVoter;

    public function __construct(FsrVoter $fsrVoter, EntityManagerInterface $entityManager, NormalizerInterface $normalizer, 
        TokenStorageInterface $tokenStorage, ContainerInterface $container)
    {
        $this->fsrVoter = $fsrVoter;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container);
    }
    
    /**
     * @param array $allData
     * @param string $key
     * @return void
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.dashboard') || 
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.dashboard') || 
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ. '.dashboard') ||
            !$data) {
            return [];
        }
        $repository = $this->loadRepository(Dashboard::class);
        $options = [];
        foreach ($data as $dashboard) {
            $existingUuids = $repository->findOneBy(['uuid' => $dashboard['uuid']]);
            $existingDisplayNames = $repository->findOneBy(['displayName' => $dashboard['displayName']]);
            $foundRoles = $this->loadRolesFromReport($dashboard, $allData['roles']);
            
            $individualOptions = [];
            $individualOptions['uuid'] = $existingDisplayNames ? $existingDisplayNames->getUuid() : $dashboard['uuid'];
            $individualOptions['name'] = $dashboard['displayName'];
            $individualOptions['roles'] = $foundRoles;
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existingUuids || $existingDisplayNames ? true : false;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType) {
        $dashboardsToAdd = [];
        $dashboardsToUpdate = [];
        foreach ($options as $dashboardOption) {
            if (!$dashboardOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $dashboard) {
                if ($dashboardOption['uuid'] == $dashboard['uuid']) {
                    unset($dashboard['userId']);
                    $foundRoles = $this->loadRolesFromReport($dashboard, []);
                    $dashboard['roles'] = array_column($foundRoles, '@id');
                    $dashboard['filterArray'] = isset($dashboard['config']['filterArray']) ? $dashboard['config']['filterArray'] : $dashboard['filterArray'];
                    unset($dashboard['roleIds']);
                    if (!$dashboardOption['exists']) {
                        $dashboardsToAdd[] = $dashboard;
                        break;
                    }
                    $dashboard['@id'] = sprintf('/api/3/dashboard/%s', $dashboard['uuid']);
                    unset($dashboard['uuid']);
                    $dashboardsToUpdate[] = $dashboard;
                    break;
                }
            }
        }
        $results = [];
        foreach ($dashboardsToAdd as $dashboard) {
            $fullURI = sprintf('%s/api/3/dashboard', $this->container->getParameter('crudhub.uri'));
            $result = $this->sendRequest($fullURI, 'post', $dashboard, $tokenType);
            $results[] = $result;
        }
        if (count($dashboardsToUpdate) > 0) {
            $fullURI = sprintf('%s/api/3/update/dashboard', $this->container->getParameter('crudhub.uri'));
            $body = [
              "data" => $dashboardsToUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }
        
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType='JWT') {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.dashboard')) {
            return [];
        }
        return $this->getResourceFromIRI($options, Dashboard::class, true);
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType='JWT') {
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $filesystem = new Filesystem();
        foreach ($data as $item) {
            $filename = $tmpDirectory . $type . '/' . $this->cleanFileName($item['displayName']) . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
        }
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return void
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType) {
        $defaultJsonPath = $tmpDirectory . $key . '/data.json';
        if (file_exists($defaultJsonPath)) {
            $fileData = file_get_contents($defaultJsonPath);
            $jsonData = json_decode($fileData, true);
        } else if (is_dir($tmpDirectory . $key)) {
            $jsonData = $this->loadDataFromDirectory($key, $tmpDirectory);
        }
        $roleData = [];
        if (is_dir($tmpDirectory . 'roles')) {
            $roleData = $this->loadDataFromDirectory('roles', $tmpDirectory);
        }
        return $this->getImportOptions([$key => $jsonData, 'roles' => $roleData], $key, $tokenType);
    }
}
