<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Core\Tag;
use App\Entity\Workflow\Workflow;
use App\Entity\Workflow\WorkflowCollection;
use App\Providers\PermissionsDictionaryBuilder;
use App\Security\Voter\FsrVoter;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Doctrine\ORM\EntityManagerInterface;
use Ramsey\Uuid\Uuid;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * Class PlaybookConfig
 * @package App\Service\ConfigExportImport
 */
class PlaybookConfig extends PortingConfig
{
    private const STEP_TYPES = [
        'MANUAL_INPUT' => 'fc04082a-d7dc-4299-96fb-6837b1baa0fe',
        'DECISION' => '12254cf5-5db7-4b1a-8cb1-3af081924b28'
    ];
    /** @var FsrVoter */
    private $fsrVoter;

    private $requiredMacros = [];

    public function __construct(
        FsrVoter $fsrVoter,
        ImportOptionsGenerator $importOptionsGenerator,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->importOptionsGenerator = $importOptionsGenerator;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if (
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.workflows') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.workflows') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.workflows')
        ) {
            return [];
        }
        $options = [];
        if (array_key_exists('collections', $data) && count($data['collections']) > 0) {
            $subOptions = $this->generatePlaybookCollectionOptions($data['collections']);
            $options['collections'] = [
                "include" => count($subOptions) > 0,
                "values" => $subOptions
            ];
        }
        if (array_key_exists('globalVariables', $data) && count($data['globalVariables']) > 0) {
            $subOptions = $this->generateGlobalVariableOptions($data['globalVariables'], $tokenType);
            $options['globalVariables'] = [
                "include" => count($subOptions) > 0,
                "values" => $subOptions
            ];
        }
        return $options;
    }

    /**
     * @param array $config
     * @return array
     */
    private function generateGlobalVariableOptions(array $dynamicVariables, $tokenType)
    {
        $existingDynamicVariables = $this->getGlobalVariables($tokenType);
        $existingDynamicVariableNames = array_column($existingDynamicVariables, 'name');
        $options = [];
        foreach ($dynamicVariables as $dynamicVariable) {
            $existing = in_array($dynamicVariable['name'], $existingDynamicVariableNames);
            $individualOptions = [];
            $individualOptions['name'] = $dynamicVariable['name'];
            $individualOptions['include'] = !$existing;
            $individualOptions['required'] = !$existing && in_array($dynamicVariable['name'], $this->requiredMacros);
            $individualOptions['exists'] = $existing;

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function getGlobalVariables($tokenType = 'JWT')
    {
        $token = $this->getTokenFromStorage();
        $workflowMacroEndpoint = '/api/dynamic-variable/?offset=0&limit=2147483647';
        if ($token) {
            $uri = $this->container->getParameter('crudhub.uri') . '/api/wf' . $workflowMacroEndpoint;
            $response = $this->sendRequest($uri, 'get', null, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('hydra:member', $jsonDecodeResponse) ? $jsonDecodeResponse['hydra:member'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('workflow.uri') . $workflowMacroEndpoint;
            return $this->sealabRequest($uri, 'get');
        }
    }

    /**
     * @param array $config
     * @return array
     */
    private function generatePlaybookCollectionOptions(array $playbookCollections)
    {
        $repository = $this->loadRepository(WorkflowCollection::class);
        $options = [];
        foreach ($playbookCollections as $playbookCollection) {
            $existingByUuid = $repository->findOneBy(['uuid' => $playbookCollection['uuid']]);
            $existingByName = $repository->findOneBy(['name' => $playbookCollection['name']]);
            $individualOptions = [];
            $individualOptions['uuid'] = $playbookCollection['uuid'];
            $individualOptions['name'] = $playbookCollection['name'];
            $individualOptions['include'] = true;
            $individualOptions['mergeType'] = 'merge_append';
            $individualOptions['exists_uuid'] = $existingByUuid ? true : false;
            $individualOptions['existing_uuid'] = $existingByName ? $existingByName->getUuid() : $playbookCollection['uuid'];
            $individualOptions['exists_name'] = $existingByName ? true : false;
            $individualOptions['exists'] = $individualOptions['exists_uuid'] || $individualOptions['exists_name'];
            $individualOptions['macros'] = $this->getMacrosFromPlaybooks($playbookCollection['workflows']);
            $playbookCount = 0;
            if ($existingByName) {
                $playbookCount = count($existingByName->getWorkflows());
            } else if ($existingByUuid) {
                $playbookCount = count($existingByUuid->getWorkflows());
            }
            $individualOptions['playbooksOverwritten'] = $playbookCount;
            $this->requiredMacros = array_unique(array_merge($this->requiredMacros, $individualOptions['macros']));

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function getMacrosFromPlaybooks($playbooks)
    {
        $macros = [];
        foreach ($playbooks as $playbook) {
            if (array_key_exists('steps', $playbook)) {
                foreach ($playbook['steps'] as $step) {
                    if (array_key_exists('arguments', $step)) {
                        $argsJson = json_encode($step['arguments']);
                        $regex = '/(?>globalVars|macro)\.(\w+)/m';
                        $matches = [];
                        preg_match_all($regex, $argsJson, $matches, PREG_PATTERN_ORDER, 0);
                        $macros = array_merge($macros, $matches[1]);
                    }
                }
            }
            if ($playbook['versions'] && count($playbook['versions'])) {
                foreach ($playbook['versions'] as $version) {
                    $version = json_decode($version['json'], true);
                    if (array_key_exists('steps', $version)) {
                        foreach ($version['steps'] as $step) {
                            if (array_key_exists('arguments', $step)) {
                                $argsJson = json_encode($step['arguments']);
                                $regex = '/(?>globalVars|macro)\.(\w+)/m';
                                $matches = [];
                                preg_match_all($regex, $argsJson, $matches, PREG_PATTERN_ORDER, 0);
                                $macros = array_merge($macros, $matches[1]);
                            }
                        }
                    }
                }
            }
        }
        return $macros;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @param string $tokenType
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType)
    {
        $tagsResult = true;
        if (array_key_exists('exported_tags', $data)) {
            $tagsResult = $this->importTags($data['exported_tags'], $tokenType);
        }
        $collectionResult = true;
        if (array_key_exists('collections', $options) && $options['collections']['include']) {
            $collectionResult = $this->importPlaybookCollections($data['collections'], $options['collections']['values'], $tokenType);
        }
        $globalVariableResult = true;
        if (array_key_exists('globalVariables', $options) && $options['globalVariables']['include']) {
            $globalVariableResult = $this->importGlobalVariables($data['globalVariables'], $options['globalVariables']['values'], $tokenType);
        } elseif (array_key_exists('globalVariables', $options) && array_key_exists('collections', $options) && $options['collections']['include']) {
            // Import required global variables from playbooks
            $requiredMacros = [];
            foreach ($options['collections']['values'] as $collection) {
                $requiredMacros = array_merge($requiredMacros, $collection['macros']);
            }
            $uniqueRequiredMacros = array_unique($requiredMacros);
            $requiredMacrosToImport = [];
            foreach ($options['globalVariables']['values'] as $globalVariable) {
                if (in_array($globalVariable['name'], $uniqueRequiredMacros)) {
                    $requiredMacrosToImport[] = $globalVariable;
                }
            }
            $globalVariableResult = $this->importGlobalVariables($data['globalVariables'], $requiredMacrosToImport, $tokenType);
        }
        return $collectionResult && $globalVariableResult && $tagsResult;
    }

    private function importTags($tags, $tokenType)
    {
        $repository = $this->loadRepository(Tag::class);
        $tagObjects = [];
        foreach ($tags as $tag) {
            $existing = $repository->findOneBy(['uuid' => $tag]);
            if ($existing) {
                continue;
            }
            $tagObjects[] = [
                'uuid' => $tag
            ];
        }
        if (!empty($tagObjects)) {
            $fullURI = sprintf('%s/api/3/bulkupsert/tags', $this->container->getParameter('crudhub.uri'));
            $body = [
                "__unique" => ['uuid'],
                "__replace" => false,
                "__data" => $tagObjects
            ];
            return $this->sendRequest($fullURI, 'post', $body, $tokenType);
        }
        return true;
    }

    /**
     * Imports playbook collections
     *
     * @param array $collections
     * @param array $collectionOptions
     * @return boolean
     */
    private function importPlaybookCollections(array $collections, array $collectionOptions, $tokenType)
    {
        $toAdd = [];
        $toAppend = [];
        $deleteUuids = [];
        $playbooksToDelete = [];
        $schedulesToImport = [];
        $playbookRepository = $this->loadRepository(Workflow::class);
        $collectionRepository = $this->loadRepository(WorkflowCollection::class);
        foreach ($collectionOptions as $playbookCollectionOption) {
            if (!$playbookCollectionOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($collections as $collection) {
                if ($playbookCollectionOption['uuid'] == $collection['uuid']) {
                    $newPlaybooks = [];
                    foreach ($collection['workflows'] as $playbook) {
                        $remove = false;
                        /** @var Workflow $existingPlaybook */
                        $existingPlaybookByUuid = $playbookRepository->findOneBy(['uuid' => $playbook['uuid']]);
                        if ($existingPlaybookByUuid && ($existingPlaybookByUuid->getCollection()->getUuid() != $collection['uuid'] || $playbookCollectionOption['mergeType'] === 'rename')) {
                            // If a uuid already exists in a different collection, or if we are renaming the current collection, generate a new one
                            $playbook['uuid'] = Uuid::uuid4()->toString();
                        } else if ($existingPlaybookByUuid && $existingPlaybookByUuid->getCollection()->getUuid() == $collection['uuid']) {
                            if ($playbookCollectionOption['mergeType'] === 'merge_replace') {
                                // If we are replacing existing playbooks, we need to remove any found playbooks by UUID in the same collection
                                $playbooksToDelete[] = $existingPlaybookByUuid->getUuid();
                            } else if ($playbookCollectionOption['mergeType'] === 'merge_append') {
                                // If we are appending, we skip any existing playbooks
                                $remove = true;
                            }
                        }
                        if ($playbookCollectionOption['exists']) {
                            $existingPlaybookByName = $playbookRepository->findOneBy(['name' => $playbook['name'], 'collection' => $playbookCollectionOption['existing_uuid']]);
                            if ($existingPlaybookByName) {
                                if ($playbookCollectionOption['mergeType'] === 'merge_replace') {
                                    // If we are replacing existing playbooks, we need to remove any found playbooks by UUID in the same collection
                                    $playbooksToDelete[] = $existingPlaybookByName->getUuid();
                                } else if ($playbookCollectionOption['mergeType'] === 'merge_append') {
                                    // If we are appending, we just need to add the playbook
                                    $remove = true;
                                }
                            }
                        }
                        unset($playbook['collection']);
                        $playbook = $this->fixStepUuids($playbook);
                        $playbook['lastModifyDate'] = time();
                        if (array_key_exists('schedules', $playbook)) {
                            foreach ($playbook['schedules'] as $schedule) {
                                $schedule['kwargs']['wf_iri'] = sprintf('/api/3/workflows/%s', $playbook['uuid']);
                                $schedulesToImport[] = $schedule;
                            }
                        }

                        if (!$remove) {
                            $newPlaybooks[] = $playbook;
                        }
                    }
                    $collection['workflows'] = $newPlaybooks;
                    if (!$playbookCollectionOption['exists_uuid'] && !$playbookCollectionOption['exists_name']) {
                        $toAdd[] = $collection;
                        break;
                    }
                    if ($playbookCollectionOption['mergeType'] === 'rename') {
                        $number = 1;
                        while ($collectionRepository->findOneBy(['name' => $collection['name'] . ' (' . $number . ')'])) {
                            $number++;
                        }
                        $collection['name'] = $collection['name'] . ' (' . $number . ')';
                        if ($playbookCollectionOption['exists_uuid']) {
                            unset($collection['uuid']);
                        }
                        $toAdd[] = $collection;
                        break;
                    } else if ($playbookCollectionOption['mergeType'] === 'replace') {
                        $deleteUuids[] = $playbookCollectionOption['existing_uuid'];
                        $toAdd[] = $collection;
                    } else if (in_array($playbookCollectionOption['mergeType'], ['merge_replace', 'merge_append'])) {
                        // UI Name (Merge Replace Existing Playbook)
                        $collection['uuid'] = $playbookCollectionOption['existing_uuid'];
                        $toAppend[] = $collection;
                    }
                    break;
                }
            }
        }
        $results = [];
        if (!empty($deleteUuids)) {
            $fullURI = sprintf('%s/api/3/delete/workflow_collections', $this->container->getParameter('crudhub.uri'));
            $body = [
                "ids" => array_unique($deleteUuids)
            ];
            $this->sendRequest($fullURI, 'delete', $body, $tokenType);
        }
        if (!empty($playbooksToDelete)) {
            $fullURI = sprintf('%s/api/3/delete/workflows', $this->container->getParameter('crudhub.uri'));
            $body = [
                "ids" => array_unique($playbooksToDelete)
            ];
            $this->sendRequest($fullURI, 'delete', $body, $tokenType);
        }
        foreach ($toAdd as $item) {
            $fullURI = sprintf('%s/api/3/workflow_collections', $this->container->getParameter('crudhub.uri'));
            $result = $this->sendRequest($fullURI, 'post', $item, $tokenType);
            $results[] = $result;
        }
        foreach ($toAppend as $collection) {
            foreach ($collection['workflows'] as $playbook) {
                $playbook['collection'] = sprintf('/api/3/workflow_collections/%s', $collection['uuid']);
                $fullURI = sprintf('%s/api/3/workflows', $this->container->getParameter('crudhub.uri'));
                $result = $this->sendRequest($fullURI, 'post', $playbook, $tokenType);
                $results[] = $result;
            }
        }
        foreach ($schedulesToImport as $schedule) {
            $this->saveSchedule($schedule, $tokenType);
        }
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Replaces the uuids for all playbook steps and points the routes to the new steps
     *
     * @param array $playbook
     * @return array
     */
    private function fixStepUuids(array $playbook)
    {
        $uuidMapping = [];
        // First generate the new uuids and keep track using $uuidMapping
        $steps = [];
        foreach ($playbook['steps'] as $step) {
            $newUuid = Uuid::uuid4()->toString();
            $uuidMapping[$step['uuid']] = $newUuid;
            $step['uuid'] = $newUuid;
            $steps[] = $step;
        }
        $playbook['steps'] = $steps;
        // Next, find all step references and replace them
        if ($playbook['triggerStep']) {
            $playbook['triggerStep'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($playbook['triggerStep'])];
        }
        $routes = [];
        foreach ($playbook['routes'] as $route) {
            $route['uuid'] = Uuid::uuid4()->toString();
            $route['targetStep'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($route['targetStep'])];
            $route['sourceStep'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($route['sourceStep'])];
            $routes[] = $route;
        }
        $playbook['routes'] = $routes;
        $steps = [];
        foreach ($playbook['steps'] as $step) {
            $stepTypeUuid = self::getUuidFromIri($step['stepType']);
            if ($stepTypeUuid == self::STEP_TYPES['MANUAL_INPUT']) {
                if (array_key_exists('timeout', $step['arguments']) && array_key_exists('step_iri', $step['arguments']['timeout'])) {
                    $step['arguments']['timeout']['step_iri'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($step['arguments']['timeout']['step_iri'])];
                }
                $options = [];
                foreach ($step['arguments']['response_mapping']['options'] as $responseMappingOption) {
                    $responseMappingOption['step_iri'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($responseMappingOption['step_iri'])];
                    $options[] = $responseMappingOption;
                }
                $step['arguments']['response_mapping']['options'] = $options;
            } elseif ($stepTypeUuid == self::STEP_TYPES['DECISION']) {
                $conditions = [];
                foreach ($step['arguments']['conditions'] as $condition) {
                    $condition['step_iri'] = '/api/3/workflow_steps/' . $uuidMapping[self::getUuidFromIri($condition['step_iri'])];
                    $conditions[] = $condition;
                }
                $step['arguments']['conditions'] = $conditions;
            }
            $steps[] = $step;
        }
        $playbook['steps'] = $steps;
        return $playbook;
    }

    private function importGlobalVariables(array $globalVariables, array $globalVariableOptions, $tokenType)
    {
        $toAdd = [];
        foreach ($globalVariableOptions as $globalVariableOption) {
            if (!$globalVariableOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($globalVariables as $globalVariable) {
                if ($globalVariableOption['name'] == $globalVariable['name']) {
                    $toAdd[] = $globalVariable;
                    break;
                }
            }
        }
        $results = [];
        foreach ($toAdd as $item) {
            $token = $this->getTokenFromStorage();
            if ($token) {
                $workflowMacroEndpoint = '/api/wf/api/dynamic-variable/';
                $uri = $this->container->getParameter('crudhub.uri') . $workflowMacroEndpoint;
                $results[] = $this->sendRequest($uri, 'post', $item, $tokenType);
            } else {
                $workflowMacroEndpoint = '/api/dynamic-variable/';
                $uri = $this->container->getParameter('workflow.uri') . $workflowMacroEndpoint;
                $results[] = $this->sealabRequest($uri, 'post', $item, $tokenType);
            }
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.workflows')) {
            return [];
        }
        $usedMacros = [];
        if (array_key_exists('collections', $options)) {
            $data['collections'] = $this->getResourceFromIRI($options['collections'], WorkflowCollection::class, true);
            if (array_key_exists('includeVersions', $options) && $options['includeVersions']) {
                $newCollections = [];
                $recordTags = [];
                foreach ($data['collections'] as $collection) {
                    $iris = [];
                    foreach ($collection['workflows'] as $playbook) {
                        $iris[] = '/api/3/workflows/' . $playbook['uuid'];
                        foreach ($playbook['recordTags'] as $recordTag) {
                            $parts = explode('/', $recordTag);
                            $recordTags[] = end($parts);
                        }
                    }
                    $collection['workflows'] = $this->getResourceFromIRI($iris, Workflow::class, true, true, true);
                    $newCollections[] = $collection;
                    foreach ($collection['recordTags'] as $recordTag) {
                        $parts = explode('/', $recordTag);
                        $recordTags[] = end($parts);
                    }
                }
                $data['collections'] = $newCollections;
            }
            $recordTags = [];
            $newCollections = [];
            foreach ($data['collections'] as $collection) {
                foreach ($collection['recordTags'] as $recordTag) {
                    $parts = explode('/', $recordTag);
                    $recordTags[] = end($parts);
                }
                foreach ($collection['workflows'] as $index => $playbook) {
                    $collection['workflows'][$index]['isPrivate'] = false;
                    $collection['workflows'][$index]['owners'] = [];
                    foreach ($playbook['recordTags'] as $recordTag) {
                        $parts = explode('/', $recordTag);
                        $recordTags[] = end($parts);
                    }
                    $schedules = $this->loadSchedulesFromPlaybook($playbook['uuid'], $tokenType);
                    if (count($schedules) > 0) {
                        $collection['workflows'][$index]['schedules'] = $schedules;
                    }
                }
                $usedMacros = array_merge($usedMacros, $this->getMacrosFromPlaybooks($collection['workflows']));
                $newCollections[] = $collection;
            }
            $data['collections'] = $newCollections;
            $data['exported_tags'] = array_values(array_unique($recordTags));
        }
        if (array_key_exists('globalVariables', $options) || count($usedMacros) > 0) {
            $data['globalVariables'] = [];
            $existingDynamicVariables = $this->getGlobalVariables($tokenType);
            $macrosToImport = array_unique(array_merge($options['globalVariables'], $usedMacros));
            foreach ($macrosToImport as $dynamicVariableName) {
                foreach ($existingDynamicVariables as $existingDynamicVariable) {
                    if ($existingDynamicVariable['name'] == $dynamicVariableName) {
                        $data['globalVariables'][] = $existingDynamicVariable;
                        break;
                    }
                }
            }
        }
        return $data;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $directory = $tmpDirectory . $type . '/';
        $filenames = [];
        foreach ($data['collections'] as $collection) {
            $collectionName = $this->cleanFileName($collection['name']);
            $collectionFolder = $directory . $collectionName . '/';
            foreach ($collection['workflows'] as $playbook) {
                $playbookName = $this->cleanFileName($playbook['name']);
                $playbookFilename = $collectionFolder . $playbookName . '.json';
                $this->writeToFile($playbookFilename, $playbook);
                $filenames[] = $playbookFilename;
            }
            unset($collection['workflows']);
            $collectionFilename = $collectionFolder . 'collection.metadata.json';
            $this->writeToFile($collectionFilename, $collection);
            $filenames[] = $collectionFilename;
        }
        $globalVariableFilename = $directory . 'globalVariables.json';
        $this->writeToFile($globalVariableFilename, $data['globalVariables']);
        $filenames[] = $globalVariableFilename;

        $exportedTagsFilename = $directory . 'tags.json';
        $this->writeToFile($exportedTagsFilename, $data['exported_tags']);
        $filenames[] = $exportedTagsFilename;

        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return void
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        $folderPath = $tmpDirectory . $key . '/';
        $items = [];
        $data = [];
        if (file_exists($folderPath)) {
            $items = scandir($folderPath);
        }
        $data['collections'] = [];
        foreach ($items as $item) {
            if ($item == 'globalVariables.json') {
                $fileData = file_get_contents($folderPath . $item);
                $jsonData = json_decode($fileData, true);
                $data['globalVariables'] = $jsonData;
            } else if ($item == 'tags.json') {
                $fileData = file_get_contents($folderPath . $item);
                $jsonData = json_decode($fileData, true);
                $data['exported_tags'] = $jsonData;
            } else if (file_exists($folderPath . $item . '/collection.metadata.json')) {
                // Folder
                $playbooks = scandir($folderPath . $item . '/');
                $collectionFolder = $folderPath . $item . '/';
                $collectionFileData = file_get_contents($collectionFolder . 'collection.metadata.json');
                $collection = json_decode($collectionFileData, true);
                $collection['workflows'] = [];
                foreach ($playbooks as $subFile) {
                    if ($subFile != 'collection.metadata.json' && $subFile != '.' && $subFile != '..') {
                        $pbFileData = file_get_contents($collectionFolder . $subFile);
                        $collection['workflows'][] = json_decode($pbFileData, true);
                    }
                }
                $data['collections'][] = $collection;
            }
        }
        return $data;
    }
}
