<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Authorization\Role;
use GuzzleHttp\Client;
use InvalidArgumentException;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Interface PortingConfig
 * @package App\Service
 */
abstract class PortingConfig
{
    const METHOD_MAP = [
        "post" => "create",
        "put" => "update",
        "delete" => "delete"
    ];

    public const RESERVE_BUFFER_BYTES = 10000000;

    /** @var EntityManagerInterface */
    protected $entityManager;

    /** @var NormalizerInterface */
    protected $normalizer;

    /** @var ContainerInterface */
    protected $container;

    /** @var TokenStorageInterface */
    protected $tokenStorage;

    protected $importUser;

    /** @var SymfonyStyle */
    protected $io;

    public function __construct(
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container
    ) {
        $this->entityManager = $entityManager;
        $this->normalizer = $normalizer;
        $this->tokenStorage = $tokenStorage;
        $this->container = $container;
        $this->io = null;
    }

    /**
     * Generates import options from import data
     *
     * @param array $data
     * @return array
     */
    abstract public function getImportOptions(array $allData, string $key, $tokenType);

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $defaultJsonPath = $tmpDirectory . $key . '/data.json';
        if (file_exists($defaultJsonPath)) {
            $fileData = file_get_contents($defaultJsonPath);
            $jsonData = json_decode($fileData, true);
        } else if (is_dir($tmpDirectory . $key)) {
            $jsonData = $this->loadDataFromDirectory($key, $tmpDirectory);
        } else {
            $jsonData = [];
        }
        return $this->getImportOptions([$key => $jsonData], $key, $tokenType);
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    abstract public function import(array $data, array $options, $tokenType);

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType)
    {
        $defaultJsonPath = $tmpDirectory . $key . '/data.json';

        if (file_exists($defaultJsonPath)) {
            $fileData = file_get_contents($defaultJsonPath);
            $jsonData = json_decode($fileData, true);
        } else if (is_dir($tmpDirectory . $key)) {
            $jsonData = $this->loadDataFromDirectory($key, $tmpDirectory);
        } else {
            $jsonData = [];
        }
        return $this->import($jsonData, $options, $tokenType);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        $folderPath = $tmpDirectory . $key . '/';
        $items = [];
        if (file_exists($folderPath)) {
            $items = scandir($folderPath);
        }
        $data = [];
        foreach ($items as $item) {
            $file = $folderPath . $item;
            if ($item != '.' && $item != '..' && file_exists($file)) {
                $fileData = file_get_contents($file);
                $jsonData = json_decode($fileData, true);
                $data[] = $jsonData;
            }
        }
        return $data;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    abstract public function export(array $options, $tokenType = 'JWT');

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $modulesDirectory = $tmpDirectory . $type . '/';
        $filesystem = new Filesystem();
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $filename = $modulesDirectory . '/data.json';
        $this->writeToFile($filename, $data);
        return $filename;
    }

    public function writeToFile($filename, $data)
    {
        $filesystem = new Filesystem();
        $freeSpace = disk_free_space(sys_get_temp_dir());
        $jsonData = json_encode($data, JSON_PRETTY_PRINT);
        if ($freeSpace - strlen($jsonData) < self::RESERVE_BUFFER_BYTES) {
            throw new \Exception(sprintf('Not enough free space in /tmp to export this file. Available: %.1fMb', $freeSpace / 1000000));
        }
        $filesystem->dumpFile($filename, $jsonData);
    }

    public function writeBytesToFile($filename, $bytes)
    {
        $filesystem = new Filesystem();
        $freeSpace = disk_free_space(sys_get_temp_dir());
        if ($freeSpace - strlen($bytes) < self::RESERVE_BUFFER_BYTES) {
            throw new \Exception(sprintf('Not enough free space in /tmp to export this file. Available: %.1fMb', $freeSpace / 1000000));
        }
        $filesystem->dumpFile($filename, $bytes);
    }

    protected function getTokenFromStorage()
    {
        return $this->tokenStorage->getToken();
    }

    protected function loadRepository($resourceClass)
    {
        return $this->entityManager->getRepository($resourceClass);
    }

    protected function getResourceFromIRI($iris, $className, $relationships = false, $export = true, $versions = false, $ignoreDepth = false)
    {
        $repository = $this->entityManager->getRepository($className);
        if ($versions) {
            $_GET['$versions'] = true;
            $_GET['$includeData'] = true;
        }
        $context = $this->getDefaultNormalizationContext($className, $export, $relationships, $ignoreDepth);
        $exportRecords = [];
        foreach ($iris as $iri) {
            $parts = explode('/', $iri);
            //TODO check if we need team rbac here. accordingly use getItemFromIri instead
            $record = $repository->findOneBy(['uuid' => end($parts)]);
            $normalizedObject = $this->normalizer->normalize($record, 'jsonld', $context);
            if ($normalizedObject) {
                $exportRecords[] = $normalizedObject;
            }
        }
        return $exportRecords;
    }

    protected function sendRequest($uri, $method, $data = null, $tokenType = 'JWT', $params = [])
    {
        $token = $this->tokenStorage->getToken();
        if ($token && $tokenType == 'JWT') {
            $client = new Client();
        } else {
            $client = $this->container->get("app.proxy.client.crudhub");
        }
        $headers = array(
            "Content-type" => "application/json",
            "Authorization" => $token ? sprintf("Bearer %s", $token->getCredentials()) : null
        );
        $request = new GuzzleRequest($method, $uri, $headers, $data ? json_encode($data) : null);
        $params = array_merge($params, ['verify' => false]);
        try {
            $response = $client->send($request, $params);
            $statusCode = $response->getStatusCode();
            if ($statusCode < 400) {
                //$jsonDecodeResponse = json_decode($response->getBody()->getContents(), true);
                return $response->getBody()->getContents();
            } else {
                throw new \Exception($uri . '\n' . $response->getBody());
            }
        } catch (BadRequestHttpException | \Exception $exception) {
            $errCode = $exception->getCode();
            if ($errCode == 403) {
                $iri = str_replace($this->container->getParameter('crudhub.uri'), '', $uri);
                $module = explode('/', $iri)[4];
                if ($module){
                    throw new \Exception(sprintf('You do not have %s access on %s module', self::METHOD_MAP[$method], $module));
                } else {
                    $error = json_decode($exception->getResponse()->getBody(), true);
                    if (array_key_exists('Error', $error)) {
                        throw new \Exception($error['Error']);
                    }
                    throw new \Exception($exception->getMessage());
                }
            } elseif ($errCode == 404) {
                $iri = str_replace($this->container->getParameter('crudhub.uri'), '', $uri);
                $message = $data ? json_encode($data) : "{}";
                throw new \Exception(sprintf('The Requested Resource %s (%s) Does not exists ', $iri, $method));
            } elseif ($errCode == 409) {
                $message = json_decode($exception->getResponse()->getBody(), true);
                $message = $message['hydra:description'] ?? $exception->getMessage();
                throw new \Exception($message);
            }
            $message = $exception->getResponse() ? $exception->getResponse()->getBody() : $exception->getMessage();
            throw new \Exception(sprintf('%d: %s', $errCode, $message));
        }
    }

    protected function sendIntegrationRequest($uri, $method, $data = null)
    {
        $client = $this->container->get("app.proxy.client.integrations");
        $headers = array(
            "Content-type" => "application/json",
            "Authorization" => null
        );
        $request = new GuzzleRequest($method, $uri, $headers, $data ? json_encode($data) : null);
        try {
            $response = $client->send($request, ['verify' => false]);
            $statusCode = $response->getStatusCode();
            if ($statusCode < 400) {
                return $response->getBody();
            } else {
                throw new \Exception($response->getBody());
            }
        } catch (BadRequestHttpException | \Exception $exception) {
            $errCode = $exception->getCode();
            $message = $exception->getResponse() ? $exception->getResponse()->getBody() : $exception->getMessage();
            throw new \Exception(sprintf('%d: %s', $errCode, $message));
        }
    }

    /**
     * Finds the value in the array by key
     *
     * @param array $array
     * @param string $key
     * @param string $value
     * @return mixed
     */
    protected static function findBy(array $array, string $key, string $value, $caseInsensitive = false)
    {
        foreach ($array as $item) {
            if ((!$caseInsensitive && $item[$key] == $value) || ($caseInsensitive && strtolower($item[$key]) == strtolower($value))) {
                return $item;
            }
        }
        return null;
    }

    protected function sealabRequest($uri, $method, $data = null)
    {
        $headers = array(
            "Content-type" => "application/json",
        );
        $workflowRequest = new GuzzleRequest($method, $uri, $headers, $data ? json_encode($data) : null);
        $client = $this->container->get("app.proxy.client.workflow");
        try {
            $response = $client->send($workflowRequest, ['verify' => false]);
            $statusCode = $response->getStatusCode();
            if ($statusCode < 400) {
                $jsonDecodeResponse = json_decode($response->getBody()->getContents(), true);
                return array_key_exists('hydra:member', $jsonDecodeResponse) ? $jsonDecodeResponse['hydra:member'] : $jsonDecodeResponse;
            } else {
                throw new \Exception($uri . '\n' . $response->getBody());
            }
        } catch (BadRequestHttpException | \Exception $exception) {
            $errVar = json_decode($exception->getResponse()->getBody()->read(2048), true);
            $errMsg['message'] = $errVar['Error'] ? isset($errVar['Error']) : $errVar['error'];
            $errCode = $exception->getCode();
            throw (sprintf('%d: %s', $errCode, $errMsg));
        }
    }

    protected function getDefaultNormalizationContext(string $resourceClass, bool $export = true, bool $relationships = false, bool $ignoreDepth = false)
    {
        return [
            'export' => $export,
            'relationships' => $relationships,
            'ignore_field_authorization' => true,
            'ignore_authorization' => true,
            'ignore_depth' => $ignoreDepth,
            'resource_class' => $resourceClass,
            'item_operation_name' => 'get',
            'collection_operation_name' => 'get',
            'showPassword' => true
        ];
    }

    protected function saveSchedule($schedule, $tokenType)
    {
        $endpoint = '/api/scheduled/?format=json';
        $method = 'post';
        // Check if schedule exists
        $getEndpoint = sprintf('/api/scheduled/?format=json&task=workflow.tasks.periodic_task&name=%s', $schedule['name']);
        try {
            $getResult = $this->sendPlaybookRequest($getEndpoint, 'get', null, $tokenType);
            if (count($getResult) > 0) {
                $method = 'put';
                $endpoint = sprintf('/api/scheduled/%s/?format=json', $getResult[0]['id']);
            }
            $result = $this->sendPlaybookRequest($endpoint, $method, $schedule, $tokenType);
            return $result['id'];
        } catch (BadRequestHttpException | \Exception $e) {
            error_log("PortingConfig:saveSchedule " . $e->getMessage());
        }
    }

    protected function loadSchedule($scheduleId, $tokenType)
    {
        $endpoint = sprintf('/api/scheduled/%s/?format=json', $scheduleId);
        try {
            $getResult = $this->sendPlaybookRequest($endpoint, 'get', null, $tokenType);
            return $getResult;
        } catch (InvalidArgumentException | \Exception $e) {
            error_log("PortingConfig:loadSchedule " . $e->getMessage());
        }
        return null;
    }

    protected function loadSchedulesFromPlaybook($playbookUuid, $tokenType)
    {
        $endpoint = sprintf('/api/scheduled/?format=json&limit=1000&ordering=-modified&task=workflow.tasks.periodic_task&wf_iri=/api/3/workflows/%s', $playbookUuid);
        try {
            $getResult = $this->sendPlaybookRequest($endpoint, 'get', null, $tokenType);
            return $getResult;
        } catch (InvalidArgumentException | \Exception $e) {
            error_log($e->getMessage());
        }
        return [];
    }

    private function sendPlaybookRequest($endpoint, $method, $data, $tokenType)
    {
        $tokenStorage = $this->container->get('security.token_storage');
        $token = $tokenStorage->getToken();
        if ($token) {
            $uri = $this->container->getParameter('crudhub.uri') . '/api/wf' . $endpoint;
            $response = $this->sendRequest($uri, $method, $data, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('hydra:member', $jsonDecodeResponse) ? $jsonDecodeResponse['hydra:member'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('workflow.uri') . $endpoint;
            return $this->sealabRequest($uri, 'get');
        }
    }

    protected function cleanFileName(string $name)
    {
        return preg_replace('/[\*\/\\\\\?]/', '_', $name);
    }

    protected function loadRolesFromReport($report, $rolesBeingImported)
    {
        $roleRepository = $this->loadRepository(Role::class);
        if ($report['roleIds']) {
            $roleUuids = $report['roleIds'];
        } else {
            $roleUuids = [];
            foreach ($report['roles'] as $roleIri) {
                $parts = explode('/', $roleIri);
                $roleUuid = end($parts);
                $roleUuids[] = $roleUuid;
            }
        }

        $roles = $roleRepository->findBy(['uuid' => $roleUuids]);
        $foundRoles = [];
        $foundUuids = [];
        foreach ($roles as $role) {
            $foundRoles[] = [
                '@id' => '/api/3/roles/' . $role->getUuid(),
                'name' => $role->getName()
            ];
            $foundUuids[] = $role->getUuid();
        }
        foreach ($rolesBeingImported as $roleBeingImported) {
            if (in_array($roleBeingImported['uuid'], $roleUuids) && !in_array($roleBeingImported['uuid'], $foundUuids)) {
                $foundRoles[] = [
                    '@id' => '/api/3/roles/' . $roleBeingImported['uuid'],
                    'name' => $roleBeingImported['name']
                ];
            }
        }
        return $foundRoles;
    }

    /**
     * Converts IRIs to UUIds
     *
     * @param string $iri
     * @return string
     */
    protected static function getUuidFromIri($iri)
    {
        $iriParts = explode('/', $iri);
        return end($iriParts);
    }

    public function setImportUser($user)
    {
        $this->importUser = $user;
    }

    public function setIO(SymfonyStyle $io)
    {
        $this->io = $io;
    }
}
