<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */


namespace App\Service\ConfigExportImport;

use App\Entity\ViewTemplate\Reporting;
use App\Providers\PermissionsDictionaryBuilder;
use App\Security\Voter\FsrVoter;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * Class ReportConfig
 * @package App\Service\ConfigExportImport
 */
class ReportConfig extends PortingConfig
{
    /** @var fsrVoter */
    private $fsrVoter;

    public function __construct(FsrVoter $fsrVoter, EntityManagerInterface $entityManager, NormalizerInterface $normalizer, 
        TokenStorageInterface $tokenStorage, ContainerInterface $container)
    {
        $this->fsrVoter = $fsrVoter;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return void
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.reporting') ||
         !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.reporting') || 
         !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.reporting') ||
         !$data) {
            return [];
        }
        $repository = $this->loadRepository(Reporting::class);
        $options = [];
        foreach ($data as $report) {
            $existing = $repository->findOneBy(['uuid' => $report['uuid']]);
            $foundRoles = $this->loadRolesFromReport($report, $allData['roles']);
            $individualOptions = [];
            $individualOptions['uuid'] = $report['uuid'];
            $individualOptions['name'] = $report['displayName'];
            $individualOptions['roles'] = $foundRoles;
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existing ? true : false;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType) {
        $reportsToAdd = [];
        $reportsToUpdate = [];
        foreach ($options as $reportOption) {
            if (!$reportOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $report) {
                if ($reportOption['uuid'] == $report['uuid']) {
                    unset($report['userId']);
                    $foundRoles = $this->loadRolesFromReport($report, []);
                    $report['roles'] = array_column($foundRoles, '@id');
                    $report['filterArray'] = isset($report['config']['filterArray']) ? $report['config']['filterArray'] : $report['filterArray'];

                    unset($report['roleIds']);
                    if (!$reportOption['exists']) {
                        $reportsToAdd[] = $report;
                        break;
                    }
                    $report['@id'] = sprintf('/api/3/reporting/%s', $report['uuid']);
                    unset($report['uuid']);
                    $reportsToUpdate[] = $report;
                    break;
                }
            }
        }
        $results = [];
        foreach ($reportsToAdd as $report) {
            $fullURI = sprintf('%s/api/3/reporting', $this->container->getParameter('crudhub.uri'));
            $result = $this->sendRequest($fullURI, 'post', $report, $tokenType);
            $results[] = $result;
        }
        if (count($reportsToUpdate) > 0) {
            $fullURI = sprintf('%s/api/3/update/reporting', $this->container->getParameter('crudhub.uri'));
            $body = [
              "data" => $reportsToUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }
        
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType='JWT') {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.reporting')) {
            return [];
        }
        return $this->getResourceFromIRI($options, Reporting::class, true);
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType='JWT') {
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $filesystem = new Filesystem();
        foreach ($data as $item) {
            $filename = $tmpDirectory . $type . '/' . $this->cleanFileName($item['displayName']) . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
        }
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return void
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType) {
        $defaultJsonPath = $tmpDirectory . $key . '/data.json';
        if (file_exists($defaultJsonPath)) {
            $fileData = file_get_contents($defaultJsonPath);
            $jsonData = json_decode($fileData, true);
        } else if (is_dir($tmpDirectory . $key)) {
            $jsonData = $this->loadDataFromDirectory($key, $tmpDirectory);
        }
        $roleData = [];
        if (is_dir($tmpDirectory . 'roles')) {
            $roleData = $this->loadDataFromDirectory('roles', $tmpDirectory);
        }
        return $this->getImportOptions([$key => $jsonData, 'roles' => $roleData], $key, $tokenType);
    }
}
