<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Authorization\Role;
use App\Entity\Core\Module;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Class RoleConfig
 * @package App\Service\ConfigExportImport
 */
class RoleConfig extends PortingConfig
{
    /**
     * @param array $allData
     * @param string $key
     * @return void
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $repository = $this->loadRepository(Role::class);
        $options = [];
        foreach ($data as $role) {
            $iriParts = explode('/', $role['@id']);
            $uuid = end($iriParts);
            $existing = $repository->findOneBy(['uuid' => $uuid]);
            $individualOptions = [];
            $individualOptions['uuid'] = $uuid;
            $individualOptions['name'] = $role['name'];
            $individualOptions['description'] = $role['description'];
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existing ? true : false;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType) {
        $toAdd = [];
        $toUpdate = [];
        $moduleRepository = $this->loadRepository(Module::class);
        foreach ($options as $roleOption) {
            if (!$roleOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $role) {
                $iriParts = explode('/', $role['@id']);
                $uuid = end($iriParts);
                if ($roleOption['uuid'] == $uuid) {
                    $existingModulePermissions = [];
                    $moduleUuids = [];
                    foreach ($role['modulePermissions'] as $modulePermission) {
                        $module = $moduleRepository->findOneBy(['type' => $modulePermission['module']]);
                        // Only include permissions for existing modules
                        if ($module) {
                            $modulePermission['module'] = '/api/3/modules/' . $module->getUuid();
                            $moduleUuids[] = $module->getUuid();
                            unset($modulePermission['@id']);
                            unset($modulePermission['uuid']);
                            if (array_key_exists('fieldPermissions', $modulePermission)) {
                                $newFieldPermissions = [];
                                foreach ($modulePermission['fieldPermissions'] as $fieldPermission) {
                                    unset($fieldPermission['@id']);
                                    unset($fieldPermission['uuid']);
                                    $newFieldPermissions[] = $fieldPermission;
                                }
                                $modulePermission['fieldPermissions'] = $newFieldPermissions;
                            }
                            $existingModulePermissions[] = $modulePermission;
                        }
                    }
                    $role['modulePermissions'] = $existingModulePermissions;
                    if ($roleOption['exists']) {
                        /** @var Role $existing */
                        $roleOption = [
                            'value' => '/api/3/roles/' . $roleOption['uuid']
                        ];
                        $existingRole = $this->export([$roleOption], 'JWT', '@id')[0];
                        // Automatically merge existing permissions
                        foreach ($existingRole['modulePermissions'] as $modulePermission) {
                            $iriParts = explode('/', $modulePermission['module']);
                            $moduleUuid = end($iriParts);
                            if (!in_array($moduleUuid, $moduleUuids)) {
                                $role['modulePermissions'][] = $modulePermission;
                            }
                        }
                        $toUpdate[] = $role;
                    } else {
                        unset($role['@id']);
                        $role['uuid'] = $uuid;
                        $toAdd[] = $role;
                    }
                    break;
                }
            }
        }
        $results = [];
        foreach ($toAdd as $item) {
            $fullURI = sprintf('%s/api/3/roles', $this->container->getParameter('crudhub.uri'));
            $result = $this->sendRequest($fullURI, 'post', $item, $tokenType);
            $results[] = $result;
        }
        if (count($toUpdate) > 0) {
            $fullURI = sprintf('%s/api/3/update/roles', $this->container->getParameter('crudhub.uri'));
            $body = [
              "data" => $toUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }
        
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType='JWT', $moduleField = 'type') {
        $iris = array_column($options, 'value');
        $roles = $this->getResourceFromIRI($iris, Role::class, true, false, false, true);
        foreach ($roles as $index => $role) {
            $option = $this->findBy($options, 'value', $role['@id']);
            foreach ($role['modulePermissions'] as $modulePermissionIndex => $modulePermission) {
                if (array_key_exists('module', $modulePermission) && is_array($modulePermission['module'])) {
                    $moduleType = $modulePermission['module'][$moduleField];
                    $roles[$index]['modulePermissions'][$modulePermissionIndex]['module'] = $moduleType;
                    if (array_key_exists('modulePermissions', $option) && array_key_exists($moduleType, $option['modulePermissions']) && !$option['modulePermissions'][$moduleType]) {
                        unset($roles[$index]['modulePermissions'][$modulePermissionIndex]);
                    }
                }
            }
            
        }
        return $roles;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType='JWT') {
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $filesystem = new Filesystem();
        foreach ($data as $item) {
            $filename = $tmpDirectory . $type . '/' . $this->cleanFileName($item['name']) . '-' . $item['uuid'] . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
        }
        return $filenames;
    }
}
