<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Actors\Actor;
use App\Entity\Authorization\Team;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Class TeamConfig
 * @package App\Service\ConfigExportImport
 */
class TeamConfig extends PortingConfig
{
    /**
     * 
     * @param array $data List of teams
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType) {
        $data = $allData[$key];
        $repository = $this->loadRepository(Team::class);
        $options = [];
        foreach ($data as $team) {
            $iriParts = explode('/', $team['@id']);
            $uuid = end($iriParts);
            $existing = $repository->findOneBy(['uuid' => $uuid]);
            $existingByName = $repository->findOneBy(['name' => $team['name']]);
            $individualOptions = [];
            $individualOptions['uuid'] = $uuid;
            $individualOptions['existingUuid'] = $existingByName ? $existingByName->getUuid() : $uuid;
            $individualOptions['name'] = $team['name'];
            $individualOptions['description'] = $team['description'];
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existing || $existingByName ? true : false;

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function filterExistingTeams(array $teams, array $allTeams) {
        $existing = [];
        foreach ($teams as $team) {
            if (is_array($team)) {
                $team = $team['@id'];
            }
            $iriParts = explode('/', $team);
            $uuid = end($iriParts);
            $existingTeam = in_array($uuid, $allTeams);
            if ($existingTeam) {
                $existing[] = $team;
            }
        }
        return $existing;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType) {
        $results = [];
        $toUpdate = [];
        $actorRepository = $this->loadRepository(Actor::class);
        $teamRepository = $this->loadRepository(Team::class);
        $allTeams = $teamRepository->findAll();
        $allTeamUuids = array_map(function ($team) {
            return $team->getUuid();
        }, $allTeams);
        foreach ($options as $teamOption) {
            if (!$teamOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $team) {
                if (!$team) {
                    continue;
                }
                $iriParts = explode('/', $team['@id']);
                $uuid = end($iriParts);
                if ($teamOption['uuid'] == $uuid) {
                    $existingActors = [];
                    foreach ($team['actors'] as $actor) {
                        $actorUuid = $actor['uuid'];
                        $existingActor = $actorRepository->findOneBy(['uuid' => $actorUuid]);
                        // Only include permissions for existing modules
                        if ($existingActor) {
                            $existingActors[] = $actor['@id'];
                        }
                    }
                    $team['actors'] = $existingActors;
                    if (!$teamOption['exists']) {
                        $team['parents'] = $this->filterExistingTeams($team['parents'], $allTeamUuids);
                        $team['children'] = $this->filterExistingTeams($team['children'], $allTeamUuids);
                        $team['siblings'] = $this->filterExistingTeams($team['siblings'], $allTeamUuids);
                        unset($team['@id']);
                        $team['uuid'] = $uuid;
                        $allTeamUuids[] = $uuid;
                        // Add immediately so that future teams can link to it
                        $fullURI = sprintf('%s/api/3/teams', $this->container->getParameter('crudhub.uri'));
                        $result = $this->sendRequest($fullURI, 'post', $team, $tokenType);
                        $results[] = $result;
                        break;
                    }
                    // Add existing actors back to the team
                    $existingTeam = $teamRepository->findOneBy(['uuid' => $teamOption['existingUuid']]);
                    foreach ($existingTeam->getActors() as $actor) {
                        $iri = sprintf('/api/3/actors/%s', $actor->getUuid());
                        if (!in_array($iri, $team['actors'])) {
                            $team['actors'][] = $iri;
                        }
                    }
                    unset($team['uuid']);
                    $team['@id'] = sprintf('/api/3/teams/%s', $teamOption['existingUuid']);
                    $toUpdate[] = $team;
                    break;
                }
            }
        }
        if (count($toUpdate) > 0) {
            $fixedTeams = [];
            foreach ($toUpdate as $team) {
                // Verify that teams exist again
                $team['parents'] = $this->filterExistingTeams($team['parents'], $allTeamUuids);
                $team['children'] = $this->filterExistingTeams($team['children'], $allTeamUuids);
                $team['siblings'] = $this->filterExistingTeams($team['siblings'], $allTeamUuids);
                $fixedTeams[] = $team;
            }
            $fullURI = sprintf('%s/api/3/update/teams', $this->container->getParameter('crudhub.uri'));
            $body = [
              "data" => $fixedTeams
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }
        
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType='JWT') {
        $teams = $this->getResourceFromIRI($options, Team::class, true, false);
        return $teams;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType='JWT') {
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $filesystem = new Filesystem();
        foreach ($data as $item) {
            $filename = $tmpDirectory . $type . '/' . $this->cleanFileName($item['name']) . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
        }
        return $filenames;
    }
}
