<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Core\Widget;
use App\Security\Voter\FsrVoter;
use Doctrine\ORM\EntityManagerInterface;
use App\Service\ConfigExportImport\PortingConfig;
use GuzzleHttp\Exception\RequestException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Filesystem\Exception\IOException;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class WidgetConfig
 */
class WidgetConfig extends PortingConfig
{
    private $fsrVoter;

    public function __construct(EntityManagerInterface $entityManager, NormalizerInterface $normalizer, 
        TokenStorageInterface $tokenStorage, ContainerInterface $container, FsrVoter $fsrVoter)
    {
        $this->fsrVoter = $fsrVoter;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container);
    }

    /**
     *
     * @param array $data Playbook object containing collections & global vars
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if ($tokenType && ! ($this->fsrVoter->isGranted('update.widgets') || $this->fsrVoter->isGranted('create.widgets') || $this->fsrVoter->isGranted('read.widgets'))) {
            return [];
        }
        $options = [];
        $repository = $this->loadRepository(Widget::class);
        $existingWidgets = $repository->findAll();
        $existingWidgetNames = array_map(function($item) {
            return [$item->getName(),$item->getVersion()];
        }, $existingWidgets);
        $options = [];
        foreach ($data as $widget) {
            $existing = in_array([$widget['name'],$widget['version']], $existingWidgetNames);
            $individualOptions = [];
            $individualOptions['name'] = $widget['name'];
            $individualOptions['version'] = $widget['version'];
            $individualOptions['title'] = $widget['title'];
            $fullKey = sprintf('%s-%s', $widget['name'], $widget['version']);
            if (array_key_exists('install_mode', $widget) && array_key_exists('installer_path', $widget)) {
                $individualOptions['install_mode'] = $widget['install_mode'];
                $individualOptions['installer_path'] = $widget['installer_path'];
            } else {
                $individualOptions['install_mode'] = 'rpm';
                $individualOptions['installer_path'] = $fullKey;
            }
            $individualOptions['exists'] = $existing;
            $individualOptions['include'] = !$existing;

            $options[] = $individualOptions;
        }
        return $options;
    }
    
    
    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType)
    {
        foreach ($options as $option) {
            if ($option['exists'] || !$option['include']) {
                // Skip existing widgets
                continue;
            }
            // Install & wait if not exists
            $params = [
                'name' => $option['name'],
                'version' => $option['version'],
                'replace' => true
            ];
            $this->sendRequest(sprintf('%s/api/3/widgets/install', $this->container->getParameter('crudhub.uri')), 'post', $params, $tokenType);
        }
        return true;
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType) {
        $widgetsPath = $tmpDirectory . $key . '/';
        $widgetsInRepo = $this->loadRepoWidgets();
        
        foreach ($options as $option) {
            if ($option['exists'] || !$option['include']) {
                // Skip existing widgets
                continue;
            }
            // Install & wait if not exists
            if ($option['install_mode'] == 'rpm') {
                if (!array_key_exists($option['installer_path'], $widgetsInRepo)) {
                    continue;
                }
                $params = [
                    'name' => $option['name'],
                    'version' => $option['version'],
                    'replace' => true
                ];
                $this->sendRequest(sprintf('%s/api/3/widgets/install', $this->container->getParameter('crudhub.uri')), 'post', $params, $tokenType);
            } else if ($option['install_mode'] == 'tgz') {
                $uri = sprintf('%s/api/3/widgets/import', $this->container->getParameter('crudhub.uri'));
                $fileStream = fopen($widgetsPath . $option['installer_path'], 'r');
                $token = $this->tokenStorage->getToken();
                $authorization = $token ? sprintf("Bearer %s", $token->getCredentials()) : null;
                $client = new \GuzzleHttp\Client();
                $params = [
                    'verify' => false,
                    'multipart' => [
                        [
                            'name' => 'body',
                            'contents' => json_encode(['replace' => true]),
                            'headers' => ['Content-Type' => 'application/json']
                        ], 
                        [
                            'name' => 'file',
                            'contents' => $fileStream,
                            'filename' => $option['installer_path'],
                            'headers' => ['Content-Type' => 'application/gzip']
                        ]
                    ],
                    'headers' => [
                        'Authorization' => $authorization
                    ]
                ];
                try {
                    $client->request('POST', $uri, $params);
                } catch (IOException | \Exception $e) {
                    $this->io->warning(sprintf('Skipping import for %s (from file %s) since the file is invalid or incompatible.', $option['name'], $option['installer_path']));
                    continue;
                }
            }
            
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType='JWT')
    {
        if (!$this->fsrVoter->isGranted('read.widgets')) {
            return [];
        }
        $data = [];
        if (count($options) > 0) {
            $repository = $this->loadRepository(Widget::class);
            $widgets = $repository->findAll();
        }
        foreach ($options as $option) {
            foreach ($widgets as $widget) {
                if ($option == $widget->getName()) {
                    $widgetObject = [
                        'name' => $widget->getName(),
                        'version' => $widget->getVersion(),
                        'title' => $widget->getTitle()
                    ];
                    
                    $data[] = $widgetObject;
                    break;
                }
            }
        }
        return $data;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType='JWT')
    {
        if (!$this->fsrVoter->isGranted('read.widgets') || count($options) == 0) {
            return null;
        }
        $widgetDirectory = $tmpDirectory . $type;
        $widgetObjects = [];
        $repository = $this->loadRepository(Widget::class);
        $widgets = $repository->findAll();
        $widgetsInRepo = $this->loadRepoWidgets();
        foreach ($options as $option) {
            foreach ($widgets as $widget) {
                if ($option == $widget->getName()) {
                    // Verify the existence of the widget in the repo
                    $rpmName = sprintf('%s-%s', $widget->getName(), $widget->getVersion());
                    $widgetObject = [
                        'name' => $widget->getName(),
                        'version' => $widget->getVersion(),
                        'title' => $widget->getTitle()
                    ];
                    if (array_key_exists($rpmName, $widgetsInRepo)) {
                        $widgetObject['install_mode'] = 'rpm';
                        $widgetObject['installer_path'] = $rpmName;
                    } else {
                        // Download tgz
                        $params = [
                            'development' => false
                        ];
                        $response = $this->sendRequest(sprintf('%s/api/3/widgets/export/%s', $this->container->getParameter('crudhub.uri'), $widget->getUuid()), 'post', $params, $tokenType);
                        $tgzFilename = $rpmName . '.tgz';
                        $this->writeBytesToFile(sprintf('%s/%s', $widgetDirectory, $tgzFilename), $response);
                        $widgetObject['install_mode'] = 'tgz';
                        $widgetObject['installer_path'] = $tgzFilename;
                    }
                    $widgetObjects[] = $widgetObject;
                    break;
                }
            }
        }
        $filename = $widgetDirectory . '/data.json';
        $this->writeToFile($filename, $widgetObjects);
        return $filename;
    }

    private function loadRepoWidgets() {
        try {
            $widgetJsonUri = sprintf('https://%s/fsr-widgets/widgets.json', getenv("REPOSERVER"));
            $response = $this->sendRequest($widgetJsonUri, 'GET', null, null, ['connect_timeout' => 15]);
            return json_decode($response, true);
        } catch (RequestException | \Exception $e) {
            $this->io->warning(sprintf('Cannot load widget RPMs from %s: %s', getenv("REPOSERVER"), $e->getMessage()));
            return [];
        }
    }
}
