<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use App\Entity\Actors\Actor;
use Doctrine\ORM\EntityManager;
use App\Entity\Settings\UserSettings;
use App\Entity\Settings\SystemSettings;
use App\Service\AbstractUserDataProvider;
use App\Repository\UserSettingsRepository;
use App\Repository\SystemSettingsRepository;
use ApiPlatform\Core\Api\IriConverterInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Lexik\Bundle\JWTAuthenticationBundle\Security\Authentication\Token\JWTUserToken;
use App\Encoder\Hmac\HmacUserToken;
use Symfony\Component\Security\Core\Authentication\AuthenticationManagerInterface;

/**
 * Data provider for the Settings document.
 *
 */
class SettingsHelper extends AbstractUserDataProvider
{

    protected $iriConverter;

    /** @var EntityManager */
    protected $entityManager;

    protected $containers;

    /** @var AuthenticationManagerInterface */
    private $authenticationManager;

    public function __construct(IriConverterInterface $iriConverter, ContainerInterface $containers, 
        TokenStorageInterface $tokenStorage, AuthenticationManagerInterface $authenticationManager)
    {
        $this->iriConverter = $iriConverter;
        $this->containers = $containers;
        $this->entityManager = $this->containers->get('doctrine.orm.entity_manager');
        $this->authenticationManager = $authenticationManager;
        parent::__construct($tokenStorage);
    }

    /**
     * @inheritdoc
     * @return array
     */
    public function getDataForActor(Actor $actor)
    {
        $userIri = $this->iriConverter->getIriFromItem($actor);
        $userSettings = $this->getUserSettingsFromIRI($userIri);
        /** @var SystemSettingsRepository $systemSettingsRepo */
        $systemSettingsRepo = $this->entityManager->getRepository(SystemSettings::class);
        $systemSettings = $systemSettingsRepo->findAll()[0];
        return [$userSettings, $systemSettings];
    }

    public function getUserSettingsFromIRI($userIri) {
         /** @var UserSettingsRepository $userSettingsRepo */
         $userSettingsRepo = $this->entityManager->getRepository(UserSettings::class);
         $userSettings = $userSettingsRepo->findBy(['owner' => $userIri]);
         if (count($userSettings)) {
             $userSettings = $userSettings[0];
         }

         if (!$userSettings || is_null($userSettings)) {
            $userSettings = new UserSettings();
            $userSettings->setOwner($userIri);
            $userSettings->setPublicValues([]);
        }
        return $userSettings;
    }

    public function getCurrentUser(){
        if ($this->tokenStorage->getToken()){
            return $this->tokenStorage->getToken()->getUser();
        }
        return null;
    }

    /**
     * @return string
     */
    public static function extractUserName($user)
    {
        if (method_exists($user, "getfirstname")) {
            if (method_exists($user, "getlastname")) {
                $userName = $user->getfirstname() . ' ' . $user->getlastname();
            } else {
                $userName = $user->getfirstname();
            }
        } else {
            $userName = $user->getname();
        }
        return $userName;
    }

    /**
     * @inheritdoc
     * @return UserSettings
     */
    public function getDataForToken(TokenInterface $token)
    {
        return parent::getDataForToken($token);
    }

    /**
     * @inheritdoc
     * @return array
     */
    public function getDataForCurrentUser()
    {
        return parent::getDataForCurrentUser();
    }

    /**
     * @param UserSettings $userSettings
     * @param array $newValues
     */
    public function saveUserSettings(UserSettings $userSettings, $newValues = [])
    {
        $userSettings->setPublicValues($newValues);

        $this->entityManager->persist($userSettings);
        $this->entityManager->flush();
    }

    /**
     * @param EntityManager $entityManager
     */
    public function setEntityManager($entityManager)
    {
        $this->entityManager = $entityManager;
    }

    /**
     * @param IriConverterInterface $iriConverter
     */
    public function setIriConverter($iriConverter)
    {
        $this->iriConverter = $iriConverter;
    }

    /**
     * @param UserSettings $userSettings
     * @param SystemSettings $systemSettings
     *
     * @return array
     */
    public function getMergedSettings(UserSettings $userSettings, SystemSettings $systemSettings)
    {
        $mergedValues = array_replace_recursive(
            is_null($systemSettings) || is_null($systemSettings->getPublicValues()) ? [] : $systemSettings->getPublicValues(),
            is_null($userSettings) || is_null($userSettings->getPublicValues()) ? [] : $userSettings->getPublicValues(),
            is_null($systemSettings) || is_null($systemSettings->getPrivateValues()) ? [] : $systemSettings->getPrivateValues()
        );

        return $mergedValues;
    }

    public function setRefreshedToken($rawToken, $tokenType='JWT')
    {
        $token = $this->tokenStorage->getToken();
        if ($tokenType == 'JWT') {
            $token = new JWTUserToken();
            $token->setRawToken($rawToken);
        } else {
            $token = new HmacUserToken();
            $token->setPayload($rawToken);
        }
        $authToken = $this->authenticationManager->authenticate($token);
        $this->tokenStorage->setToken($authToken);
    }
}
