<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use InvalidArgumentException;
use Symfony\Component\Yaml\Yaml;
use App\Constants\CacheConstants;
use App\Entity\Core\EmailTemplate;
use App\Service\AccessibleTeamIdsProvider;
use DateTime;
use Symfony\Component\Filesystem\Filesystem;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use ApiPlatform\Core\DataProvider\ItemDataProviderInterface;
use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class UtilityService
{

    const WIDGET_PREVIEW_PATH = '/opt/cyops-ui/widgets_preview/';
    const WIDGET_PATH = '/opt/cyops-ui/widgets/';
    private const MASTER_ROLE_NAME = 'master';
    private const IS_DISTRIBUTED_INSTANCE = 'isDistributedInstance';
    private const IS_INSTANCE_ROLE = 'instanceRole';

    private $headers = array("Content-type" => "application/json");

    protected $accessibleTeamIdsProvider;
    protected $tokenStorage;
    protected $containers;

    protected $dasClient;

    protected $dasUri;

    private $itemDataProvider;
    private $metadataHelper;

    public function __construct(
        AccessibleTeamIdsProvider $accessibleTeamIdsProvider,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $containers,
        TagAwareCacheInterface $cache,
        ItemDataProviderInterface $itemDataProvider,
        MetadataHelper $metadataHelper
    ) {
        $this->accessibleTeamIdsProvider = $accessibleTeamIdsProvider;
        $this->tokenStorage = $tokenStorage;
        $this->containers = $containers;
        $this->cache = $cache;
        $this->dasClient = $this->containers->get("app.proxy.client.das");
        $this->dasUri = $this->containers->getParameter('das.uri');
        $this->itemDataProvider = $itemDataProvider;
        $this->metadataHelper = $metadataHelper;
    }

    public function startsWith($string, $startString)
    {
        $len = strlen($startString);
        return (substr($string, 0, $len) === $startString);
    }

    public function getUniqueResourceFields($request_data, $resourceShortName)
    {
        $permission = $this->getModulePermission($resourceShortName);
        if (isset($request_data) && isset($request_data->{'__unique'}) && $request_data->{'__unique'}) {
            return [$request_data->{'__unique'}, $permission];
        }
        $cacheName = 'constraintField_' . $resourceShortName;
        $item = $this->cache->getItem($cacheName);
        if ($item->isHit()) {
            return $item->get();
        }
        $managerRegistry = $this->containers->get('doctrine');
        $manager = $managerRegistry->getManagerForClass(ModelMetadata::class);
        $modelMetaDataRepository = $manager->getRepository(ModelMetadata::class);
        try {
            $modelMetaData = $modelMetaDataRepository->findOneBy(array('type' => $resourceShortName));
        } catch (InvalidArgumentException | \Exception $e) {
            error_log($e->getMessage());
        }
        if (!$modelMetaData) {
            return [[], $permission];
        }
        $uniqueConstraint = count($modelMetaData->getUniqueConstraint()) ? $modelMetaData->getUniqueConstraint()[0] : [];
        $module = $modelMetaData->getType();
        $columns = [];
        foreach ($uniqueConstraint as $key => $value) {
            $columns = isset($value['columns']) ? $value['columns'] : [];
        }
        $item->set([$columns, $module]);
        $item->expiresAfter(3000);
        $item->tag(CacheConstants::UPSERT_MODULE_FIELDS);
        $this->cache->save($item);
        return [$columns, $module];
    }

    public function getModulePermission($resourceShortName)
    {
        $typeClassMap = $this->cache->getItem(CacheConstants::ENTITY_TYPE_CLASS_MODEL_MAP);
        if (!$typeClassMap->isHit()) {
            $this->metadataHelper->createModuleMapCache();
            $typeClassMap = $this->cache->getItem(CacheConstants::ENTITY_TYPE_CLASS_MODEL_MAP);
        }
        [$classPath, $module] = $typeClassMap->get()[$resourceShortName];
        return $module;
    }

    public function prepareFilterData($requestBody, $filterItems)
    {
        $filter = [];
        foreach ($filterItems as $item) {
            if (is_string($item)) {
                if (is_array($requestBody)) {
                    $filter[$item] = $requestBody[$item] ?? null;
                } else {
                    $filter[$item] = $requestBody->$item ?? null;
                }
            } else {
                $filter = array_merge($filter, get_object_vars($item));
            }
        }
        return $filter;
    }

    public function getItemWithWhere($repository, $filters)
    {
        $queryBuilder = $repository->createQueryBuilder('o');

        foreach ($filters as $key => $value) {
            if (gettype($value) == 'array') {
                $queryBuilder->andwhere('o.' . $key . ' in (:' . $key . ')');
                $queryBuilder->setParameter($key, $value);
                continue;
            }
            if (gettype($value) != 'boolean') {
                if ($this->startsWith($value, '/api/3/')) {
                    $explodedValue = explode('/', $value);
                    $value = end($explodedValue);
                }
                $queryBuilder->andwhere('o.' . $key . '= :' . $key);
                $queryBuilder->setParameter($key, $value);
            } else {
                $queryBuilder->andwhere('o.' . $key . '= :value');
                $queryBuilder->setParameter('value', $value);
            }
        }
        return $queryBuilder->getQuery()->execute();
    }

    /**
     * @param $item
     * @param $normalizedTeamData
     * @param $normalizedItemData
     */
    protected function isRecordInList($accessibleRecordTeamUuid, $loggedInUserTeamUuid)
    {
        $returnFlag = true;
        if (count($accessibleRecordTeamUuid) == 0) {
            // Public record
            $returnFlag =  true;
        }
        if (count($loggedInUserTeamUuid) == 0) {
            // User not on any teams
            $returnFlag =  false;
        }
        if (count($accessibleRecordTeamUuid) && count($loggedInUserTeamUuid) && (!count(array_intersect($accessibleRecordTeamUuid, $loggedInUserTeamUuid)))) {
            // Checking if Record owning teams and logged in user teams have common teams
            $returnFlag =  false;
        }
        return $returnFlag;
    }

    protected function getRecordTeamUuid($object)
    {
        $record_team = [];
        if (method_exists($object, 'getOwners') && $object->getOwners() != null && $object->getOwners() != []) {
            $record_team = $this->accessibleTeamIdsProvider->getParentDataForTeam($object->getOwners());
        }
        return $record_team;
    }

    protected function getLoggedInUserTeamUuid($object)
    {
        $record_team = [];
        if (method_exists($object, 'getTeams') && $object->getTeams() != null && $object->getTeams() != []) {
            foreach ($object->getTeams() as $team) {
                array_push($record_team, $team->getUuid());
            }
        }
        return $record_team;
    }

    private function getUser()
    {
        return $this->tokenStorage->getToken()->getUser();
    }

    public function ValidateOwnerShip($request, $item, $resource = null)
    {
        if ($request && $item) {
            $accessibleRecordTeamUuid = $this->getRecordTeamUuid($item);
            $loggedInUser = $this->getUser();
            $loggedInUserTeamUuid = $this->getLoggedInUserTeamUuid($loggedInUser);
            if (!count($loggedInUserTeamUuid)) {
                // User with no team should not be allowed to view the resource
                throw new AccessDeniedException();
            }
            // Check if Record is ownable
            $teamOwned = $this->isRecordInList($accessibleRecordTeamUuid, $loggedInUserTeamUuid);
            $userOwned = $this->doesUserOwnRecord($item);
            if (!$teamOwned && !$userOwned) {
                throw new AccessDeniedException();
            }
        }
    }

    protected function doesUserOwnRecord($item)
    {
        if (
            method_exists($item, 'getUserOwners')
            && $item->getUserOwners() != null
            && $item->getUserOwners() != []
            && $item->getUserOwners()->contains($this->getUser())
        ) {
            return true;
        }
        return false;
    }

    public function filePutContent($filePath, $fileContent)
    {
        file_put_contents($filePath, $fileContent);
    }

    protected function createWidgetControllerFiles($src, $dst, $matchString, $replaceString)
    {
        $fileContent = preg_replace('/' . $matchString . '/', $replaceString, file_get_contents($src));

        $file = fopen($dst, "w");
        fwrite($file, $fileContent);
        fclose($file);
        if (strpos($src, "widgets/installed/") && strpos($dst, "widgets/development/")) {
            $this->filePutContent($dst, str_replace("widgets/installed/", "widgets/development/", file_get_contents($dst)));
        }
        if (strpos(file_get_contents($src), "widgets/development/") && strpos(file_get_contents($src), "widgets/installed/")) {
            $this->filePutContent($dst, str_replace("widgets/development/", "widgets/installed/", file_get_contents($dst)));
        }
    }

    protected function widgetCustomCopy($src, $dst, $widgetName, $widgetVersion, $removeDevString)
    {
        $filePathArray = explode('/', $src);
        $filePathName = end($filePathArray);
        $widgetsFileSystem = new Filesystem();

        if (count($filePathArray) == 7 && $filePathName == 'view.controller.js') {
            $widgetVersionString = str_replace('.', '', $widgetVersion);
            $matchString = $removeDevString ? ($widgetName . $widgetVersionString . 'DevCtrl') : ($widgetName . $widgetVersionString . 'Ctrl');
            $replaceString = $removeDevString ? ($widgetName . $widgetVersionString . 'Ctrl') : ($widgetName . $widgetVersionString . 'DevCtrl');

            $this->createWidgetControllerFiles($src, $dst, $matchString, $replaceString);
        } else if (count($filePathArray) == 7 && $filePathName == 'edit.controller.js') {
            $widgetVersionString = str_replace('.', '', $widgetVersion);
            $matchString = $removeDevString ? ('edit' . ucfirst($widgetName) . $widgetVersionString . 'DevCtrl') : ('edit' . ucfirst($widgetName) . $widgetVersionString . 'Ctrl');
            $replaceString = $removeDevString ? ('edit' . ucfirst($widgetName) . $widgetVersionString . 'Ctrl') : ('edit' . ucfirst($widgetName) . $widgetVersionString . 'DevCtrl');

            $this->createWidgetControllerFiles($src, $dst, $matchString, $replaceString);
        } else {
            $widgetsFileSystem->copy($src, $dst);
        }
    }

    public function copyWidgetFiles($src, $dst, $widgetName, $widgetVersion, $removeDevString)
    {
        $widgetsFileSystem = new Filesystem();
        if (!is_dir($dst)) {
            $widgetsFileSystem->mkdir($dst);
        }
        if (is_dir($src)) {
            $dir = opendir($src);
            while ($file = readdir($dir)) {
                if ($file != "." && $file != "..") {
                    if (is_dir($src . "/" . $file)) {
                        $this->copyWidgetFiles($src . "/" . $file, $dst . "/" . $file, $widgetName, $widgetVersion, $removeDevString);
                    } else {
                        $this->widgetCustomCopy($src . "/" . $file, $dst . "/" . $file, $widgetName, $widgetVersion, $removeDevString);
                    }
                }
            }
            closedir($dir);
        } else {
            $this->widgetCustomCopy($src, $dst, $widgetName, $widgetVersion, $removeDevString);
        }

        return true;
    }

    public function removeWidgetDir($widgetPath)
    {
        $widgetsFileSystem = new Filesystem();
        if (!file_exists($widgetPath)) {
            return;
        }
        try {
            $files = scandir($widgetPath);
            foreach ($files as $file) {
                if ($file === '.' || $file === '..') {
                    continue;
                }
                $path = $widgetPath . '/' . $file;
                if (is_dir($path)) {
                    $this->removeWidgetDir($path);
                } else {
                    unlink($path);
                }
            }
            $widgetsFileSystem->remove($widgetPath);
        } catch (\Exception $th) {
            error_log("Error: Widget directory cleanup operation not completed.");
        }
    }

    public function removeWidgetFiles($widgetPath)
    {
        $widgetsFileSystem = new Filesystem();
        if ($widgetsFileSystem->exists($widgetPath)) {
            $widgetsFileSystem->remove($widgetPath);
        }
        return true;
    }

    public function syncHANodes($path, $haSynCommandType)
    {
        if ($haSynCommandType === 'delete') {
            $payload =  array('paths' => [$path]);
        } elseif ($haSynCommandType === 'copy') {
            $payload = array(
                'source' => array(
                    'paths' => [$path]
                )
            );
        } elseif ($haSynCommandType === 'rename') {
            $payload = array(
                'source' => $path['source'],
                'target' => $path['target']
            );
        }
        $objectData = array(
            'handlerType' => 'file_handler',
            'commandType' => $haSynCommandType,
            'nodeIds' => 'OTHER_NODES',
            'payload' => $payload
        );
        try {
            $dasRequests = new GuzzleRequest('post', $this->dasUri . '/clustercommand/', $this->headers, json_encode($objectData));
            $dasResponse = $this->dasClient->send($dasRequests);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode != 200) {
                return -1;
            }
        } catch (\Exception $error) {
            error_log($error->getMessage());
            return -1;
        }
    }

    public function broadcastToHA($commandId, $command)
    {
        $objectData = array(
            "commandId" => $commandId,
            "command" => $command,
        );

        $dasRequests = new GuzzleRequest('post', $this->dasUri . '/clustercommand/', $this->headers, json_encode($objectData));
        try {
            $dasResponse = $this->dasClient->send($dasRequests);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode > 299) {
                return 0;
            }
        } catch (BadRequestHttpException | \Exception $error) {
            error_log($error->getMessage());
        }
        return true;
    }

    public static function identifyChangeFields($requestBody)
    {
        $keys = ["fsr_primary", "password"];
        foreach ($requestBody as $key => $value) {
            $keys[] = $key;
            if ($key == '__link') {
                foreach ($requestBody->{'__link'} as $innerKey => $innerValue) {
                    $keys[] = $innerKey;
                }
            }
            if ($key == '__unlink') {
                foreach ($requestBody->{'__unlink'} as $innerKey => $innerValue) {
                    $keys[] = $innerKey;
                }
            }
        }
        return $keys;
    }


    public function getDetailsFromDas()
    {
        $instanceRole = self::MASTER_ROLE_NAME;
        $isDistributedInstance = false;
        $dasRequests = new GuzzleRequest('get', $this->dasUri . '/license/?param=license_details', $this->headers);
        try {
            $dasResponse = $this->dasClient->send($dasRequests);
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode < 300) {
                $isDistributedInstance = $jsonDecodeResponse['details']['is_distributed'];
                $instanceRole = $isDistributedInstance ? $jsonDecodeResponse['details']['role'][0] : null;
                return [$isDistributedInstance, $instanceRole, true];
            }
        } catch (BadRequestHttpException | \Exception $error) {
            error_log($error->getMessage());
        }
        return [$isDistributedInstance, $instanceRole, false];
    }

    public function checkForMsspInstance()
    {
        $item = $this->cache->getItem(self::IS_DISTRIBUTED_INSTANCE);
        if ($item->isHit()) {
            return $item->get();
        } else {
            [$isDistributedInstance, $instanceRole, $lic_deployed] = $this->getDetailsFromDas();
        }
        if ($lic_deployed) {
            $item->set($isDistributedInstance);
            $item->tag(CacheConstants::MSSP_TYPE_ROLE);
            $this->cache->save($item);

            $item = $this->cache->getItem(self::IS_INSTANCE_ROLE);
            $item->set($instanceRole);
            $item->tag(CacheConstants::MSSP_TYPE_ROLE);
            $this->cache->save($item);
        }
        return $isDistributedInstance;
    }

    public function getInstanceRole()
    {
        $item = $this->cache->getItem(self::IS_INSTANCE_ROLE);
        if ($item->isHit()) {
            return $item->get();
        } else {
            [$isDistributedInstance, $instanceRole, $lic_deployed] = $this->getDetailsFromDas();
        }
        if ($lic_deployed) {
            $item->set($instanceRole);
            $item->tag(CacheConstants::MSSP_TYPE_ROLE);
            $this->cache->save($item);

            $item = $this->cache->getItem(self::IS_DISTRIBUTED_INSTANCE);
            $item->set($isDistributedInstance);
            $item->tag(CacheConstants::MSSP_TYPE_ROLE);
            $this->cache->save($item);
        }

        return $instanceRole;
    }

    /**
     * @param $val
     * @return DateTime
     */
    public function parseDateField($val)
    {
        if (strpos($val, '{{') !== false) {
            $regex = '/{{\s*getRelativeDate\(\s*(?\'years\'[^,]*)\s*,\s*(?\'months\'[^,]*)\s*,\s*(?\'days\'[^,]*)\s*,\s*(?\'hours\'[^,]*)\s*,\s*(?\'minutes\'[^,]*)\s*,\s*(?\'seconds\'[^,]*)\s*\)\s*}}/m';
            preg_match_all($regex, $val, $matches, PREG_SET_ORDER, 0);
            if (count($matches) != 1) {
                return null;
            }
            $years = $matches[0]['years'];
            $relativeDate = new DateTime();
            $relativeDate->modify($years . ' years');

            $months = $matches[0]['months'];
            if ($months == '"start"') {
                $relativeDate->setDate($relativeDate->format('Y'), 1, $relativeDate->format('d'));
            } else if ($months == '"end"') {
                $relativeDate->setDate($relativeDate->format('Y'), 12, $relativeDate->format('d'));
            } else if (intval($months)) {
                $relativeDate->modify($months . ' months');
            }

            $days = $matches[0]['days'];
            if ($days == '"start"') {
                $relativeDate->setDate($relativeDate->format('Y'), $relativeDate->format('m'), 1);
            } else if ($days == '"end"') {
                $relativeDate->modify('last day of this month');
            } else if (intval($days)) {
                $relativeDate->modify($days . ' days');
            }

            $hours = $matches[0]['hours'];
            if ($hours == '"start"') {
                $relativeDate->setTime(0, $relativeDate->format('i'));
            } else if ($hours == '"end"') {
                $relativeDate->setTime(23, $relativeDate->format('i'));
            } else if (intval($hours)) {
                $relativeDate->modify($hours . ' hours');
            }

            $minutes = $matches[0]['minutes'];
            if ($minutes == '"start"') {
                $relativeDate->setTime($relativeDate->format('H'), 0);
            } else if ($minutes == '"end"') {
                $relativeDate->setTime($relativeDate->format('H'), 59);
            } else if (intval($minutes)) {
                $relativeDate->modify($minutes . ' minutes');
            }

            $seconds = $matches[0]['seconds'];
            if ($seconds == '"start"') {
                $relativeDate->setTime($relativeDate->format('H'), $relativeDate->format('i'), 0);
            } else if ($seconds == '"end"') {
                $relativeDate->setTime($relativeDate->format('H'), $relativeDate->format('i'), 59);
            } else if (intval($seconds)) {
                $relativeDate->modify($seconds . ' seconds');
            }
            return $relativeDate;
        }
        return new DateTime($val);
    }

    public function sendMail($templateId, $dataToReplace, $mailDetails, $integrationClient, $defaultMessage)
    {
        $resource = EmailTemplate::class;
        $email_template = [];
        try {

            $object = $this->itemDataProvider->getItem($resource, ["uuid" => $templateId]);
            $subject = $object->getSubject();
            $content = $object->getContent();
        } catch (InvalidArgumentException | \Exception $error) {
            $object = false;
        }

        if (!$object) {
            $email_options = Yaml::parse(file_get_contents('/opt/cyops/configs/cyops-api/mailtemplate.yml'));
            $email_template = $email_options[$defaultMessage];
            $subject = $email_template['subject'];
            $content = $email_template['content'];
        }

        if (strpos($content, 'https://{{@Server_fqhn}}')) {
            $content = str_replace('https://{{@Server_fqhn}}', '{{@Server_fqhn}}', $content);
        }
        if (strpos($content, 'https://{{@cyops_hostname}}')) {
            $content = str_replace('https://{{@cyops_hostname}}', '{{@cyops_hostname}}', $content);
        }
        foreach ($dataToReplace as $key => $value) {
            $content = str_replace($key, $value, $content);
        }

        $smtpObject = '{
                            "connector": "smtp",
                            "operation": "send_richtext_email",
                            "params": {
                                "ccc": "",
                                "bcc": "",
                                "content_type":"text/html"
                            }
                        }';

        $jsonDecodedSmtp = json_decode($smtpObject, true);
        $jsonDecodedSmtp['params']['subject'] = $subject;
        $jsonDecodedSmtp['params']['content'] = $content;
        $jsonDecodedSmtp['params']['to'] = $mailDetails['to'];
        $smtpUri = $this->containers->getParameter('integrations.uri');
        $smtpUri = $smtpUri . '/execute/?format=json';
        $smtpRequest = new GuzzleRequest('post', $smtpUri, $this->headers, json_encode($jsonDecodedSmtp));
        try {
            $integrationClient->send($smtpRequest);
        } catch (BadRequestHttpException | \Exception $exception) {
            // error_log($exception->getResponse()->getBody()->read(2048));
            error_log($exception->getMessage());
            error_log($exception->getCode());
            return false;
        }
        return true;
    }

    // https://stackoverflow.com/questions/20522605/what-is-the-best-way-to-resolve-a-relative-path-like-realpath-for-non-existing
    // Adding custom Function to resolve as realpath might not work with nonexisting file paths.
    public function normalize_path($str)
    {
        $N = 0;
        $A = explode("/", preg_replace("/\/\.\//", '/', $str));  // remove current_location
        $B = [];
        for ($i = sizeof($A) - 1; $i >= 0; --$i) {
            if (trim($A[$i]) === "..") {
                $N++;
            } else {
                if ($N > 0) {
                    $N--;
                } else {
                    $B[] = $A[$i];
                }
            }
        }
        return implode("/", array_reverse($B));
    }
}
