<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Utility;

use App\Entity\Custom\Tenant;
use App\Service\UtilityService;
use App\Traits\PeerReplicable;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Contracts\Cache\TagAwareCacheInterface;

class MsspUtility
{

    const DATA_ROUTING_KEY = 'route.crudhub.data.remoterequest';
    const PLAYBOOK_ALIAS_ROUTING_KEY = 'route.crudhub.instructions.remoterequest.playbookaliaspush';
    const MMD_ROUTING_KEY = 'route.crudhub.instructions.remoterequest.stagingmmdupdate';
    const IS_DISTRIBUTED_INSTANCE = 'isDistributedInstance';

    protected $cache;

    protected $utilityService;

    protected $entityManager;

    protected $container;

    public function __construct(UtilityService $utilityService, EntityManagerInterface $entityManager, ContainerInterface $container, TagAwareCacheInterface $cache)
    {
        $this->utilityService = $utilityService;
        $this->entityManager = $entityManager;
        $this->cache = $cache;
        $this->container = $container;
    }

    public function prepareReplicationDiffData($currentData, $replicationOp, $previousData = [], $changeData = [])
    {
        if ($replicationOp == 'PUT' && isset($previousData['tenant']) && $previousData['tenant'] != $currentData['tenant']) {
            $currentData['__sync'] = true;
        }
        return (object) [
            "old_data" => $previousData,
            "new_data" => $currentData,
            // change_data matters only for a PUT operation. for POST or DELETE, old_data or new_data have everything
            "change_data" => $changeData,
            "method" => $replicationOp
        ];
    }

    public function isPeerReplicable($entity_class)
    {
        return in_array(PeerReplicable::class, class_uses($entity_class));
    }

    public function replicateData($entity, $requestBody, $objects, $method, $conflictingData = false)
    {
        $item = $this->cache->getItem(self::IS_DISTRIBUTED_INSTANCE);
        $isDistributedInstance = $item->get();
        if (!$isDistributedInstance) {
            return;
        }
        $replicateDataList = [];
        $sync = false;
        $replicateData = true;
        $routing_key = '';
        $entityClass = $entity;
        $entityClassName = explode('\\', $entityClass);

        if (isset($requestBody->data) && gettype($requestBody->data) == gettype(array())) {
            if (isset($requestBody->data[0]->{'__sync'})) {
                $sync = $requestBody->data[0]->{'__sync'};
            }
        } elseif (isset($requestBody->{'__sync'})) {
            $sync = $requestBody->{'__sync'};
        }

        if (isset($requestBody->{'__replicateData'})) {
            $replicateData = $requestBody->{'__replicateData'};
            if ($conflictingData) {
                $replicateData = true;
            }
        }
        if ($this->isPeerReplicable($entityClass)) {
            $routing_key = self::DATA_ROUTING_KEY;
            $replicateDataList = $objects;
            if ($sync) {
                $replicateDataList = $this->processSyncRequest($objects);
            }
        } elseif (end($entityClassName) == 'Workflow') {
            $routing_key = self::PLAYBOOK_ALIAS_ROUTING_KEY;
            $replicateDataList = $this->processPlaybook($objects, $replicateDataList);
        } elseif (end($entityClassName) == 'StagingModelMetadata') {
            $routing_key = self::MMD_ROUTING_KEY;
            $replicateDataList = $objects;
        } elseif (end($entityClassName) == 'WorkflowCollection') {
            $routing_key = self::PLAYBOOK_ALIAS_ROUTING_KEY;
            $replicateDataList = $this->processPlaybookCollections($objects, $replicateDataList, $method);
        }
        if ($replicateData && count($replicateDataList) > 0) {
            return $this->sendReplicateData($replicateDataList, $method, $routing_key);
        }
    }

    public function sendReplicateData($object, $method, $routing_key)
    {
        $groupedObj = (object) null;
        foreach ($object as $obj) {
            $objData = is_array($obj->new_data) ? $obj->new_data: $obj->old_data;
            $destinationIds = $this->getDestinationIds($objData);
            if (!$destinationIds) {
                continue;
            }
            foreach ($destinationIds as $id) {
                if (isset($groupedObj->$id)) {
                    $merge = array_merge($groupedObj->$id, [$obj]);
                    $groupedObj->$id = $merge;
                } else {
                    $groupedObj->$id = [$obj];
                }
            }
        }
        foreach ($groupedObj as $key => $value) {
            $replicationData = $this->prepareReplicationData($value, $method, $this->getTenantObject()->getTenantId(), $key);
            $this->sendReplicationData($replicationData, $routing_key);
        }
    }

    public function getTenantObject($tenantUuid=null)
    {
        if ($tenantUuid) {
            $tenantFilter = ['uuid' => $tenantUuid];
        } else {
            $tenantFilter = ['role' => 'self'];
        }
        $tenantResource = Tenant::class;
        $entityRepo = $this->entityManager->getRepository($tenantResource);
        $tenantObj = $this->utilityService->getItemWithWhere($entityRepo, $tenantFilter);
        if ($tenantObj) {
            return $tenantObj[0];
        }
        return null;
    }

    /**
     * @param $contextRef
     * @param $Object
     * @param $method
     * @return array
     */
    protected function getDestinationIds($Object)
    {
        $ids = array();
        $tenantUuid = null;
        $tenantResource = Tenant::class;
        if (isset($Object['tenant']) && $Object['tenant']) {
            $tenant = $Object['tenant'];
            $tenantArray = explode('/', $tenant);
            $tenantUuid = end($tenantArray);
        }
        $tenantObj = $this->getTenantObject($tenantUuid);
        if ($tenantObj) {
            if ($tenantObj->getRole() == 'self' || $Object['@type'] == 'Workflow') {
                $tenantFilter = ['role' => 'master', 'isDedicated' => true];
                $entityRepo = $this->entityManager->getRepository($tenantResource);
                $destinationTenants = $this->utilityService->getItemWithWhere($entityRepo, $tenantFilter);
                if ($destinationTenants) {
                    foreach ($destinationTenants as $destinationTenant) {
                        $masterId = $destinationTenant->getTenantId();
                        array_push($ids, $masterId);
                    }
                }
            } elseif ($tenantObj->getTenantId() && $tenantObj->getIsDedicated()) {
                array_push($ids, $tenantObj->getTenantId());
            }
        }
        return $ids;
    }

    public function prepareReplicationData($object, $method, $sourceId, $destinationId)
    {
        return [
            'data' => $object,
            'method' => $method,
            'sourceId' => $sourceId,
            'destinationId' => $destinationId,
        ];
    }

    protected function processSyncRequest($objects)
    {
        foreach ($objects as $obj) {
            $obj->new_data['__sync'] = true;
        }
        return $objects;
    }

    protected function processPlaybook($objects, $replicateDataList)
    {
        $remoteExecutableNew = false;
        $remoteExecutableOld = false;
        foreach ($objects as $obj) {
            if (is_array($obj->new_data)) {
                $remoteExecutableNew = $obj->new_data['remoteExecutableFlag'];
            }
            if (is_array($obj->old_data)) {
                $remoteExecutableOld = $obj->old_data['remoteExecutableFlag'];
            }
            if ($remoteExecutableNew == true || $remoteExecutableOld == true) {
                if (is_array($obj->new_data)) {
                    unset($obj->new_data['steps']);
                    unset($obj->new_data['routes']);
                    unset($obj->new_data['versions']);
                }
                if (is_array($obj->old_data)) {
                    unset($obj->old_data['steps']);
                    unset($obj->old_data['routes']);
                    unset($obj->new_data['versions']);
                }
                array_push($replicateDataList, $obj);
            }
        }
        return $replicateDataList;
    }

    protected function processPlaybookCollections($objects, $replicateDataList, $method)
    {
        if ($method != 'PUT') {
            foreach ($objects as $wf_collection_obj) {
                if (is_array($wf_collection_obj->new_data) && count($wf_collection_obj->new_data)) {
                    $processObject = $wf_collection_obj->new_data;
                } elseif (is_array($wf_collection_obj->old_data) && count($wf_collection_obj->old_data)) {
                    $processObject = $wf_collection_obj->old_data;
                } else {
                    return [];
                }
                if (isset($processObject['workflows']) && count($processObject['workflows'])) {
                    foreach ($processObject['workflows'] as $wf_object) {
                        if ($wf_object['remoteExecutableFlag']) {
                            unset($wf_object['steps']);
                            unset($wf_object['routes']);
                            unset($wf_object['versions']);
                            if ($method == 'POST') {
                                $diffObj = $this->prepareReplicationDiffData($wf_object, $method);
                            } elseif ($method == 'DELETE') {
                                $diffObj = $this->prepareReplicationDiffData([], $method, $wf_object);
                            }
                            array_push($replicateDataList, $diffObj);
                        }
                    }
                }
            }
        }
        return $replicateDataList;
    }

    public function sendReplicationData(array $data, $routing_key)
    {
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode($data), $routing_key);
    }

}
