<?php
/* Copyright start
  Copyright (C) 2008 - 2021 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Validator;

use App\Entity\ModelMetadata\ModelMetadata;
use App\Validator\ModelMetadataCollectionConstraint;
use Doctrine\Inflector\InflectorFactory;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;
use Symfony\Component\Validator\Context\ExecutionContextInterface;

class ModelMetadataCollectionConstraintValidator extends ConstraintValidator
{
    /** @var array */
    protected $coreClassTypeMap;
   
    protected $container;

    protected $inflector;

    public function __construct(ContainerInterface $container) {
        $this->container = $container;
        $this->coreClassTypeMap = $this->container->getParameter("type_map");
        $this->inflector = InflectorFactory::create()->build();
    }

    /** @var array */
    protected $jsonPrimitives = [
        'array',
        'boolean',
        'integer',
        'number',
        'null',
        'object',
        'string'
    ];


    /**
     * @inheritdoc
     */
    public function validate($value, Constraint $constraint)
    {
        if (!$constraint instanceof ModelMetadataCollectionConstraint) {
            return;
        }

        if (!$this->context instanceof ExecutionContextInterface) {
            return;
        }

        /** @var ModelMetadata[] $modelMetadatasTypeMap */
        $modelMetadatasTypeMap = [];

        /** @var ModelMetadata[] $value */
        foreach ($value as $modelMetadata) {
            $attributeMetadatas = [];
            foreach ($modelMetadata->getAttributes()->getValues() as $attributeMetadata) {
                $attributeMetadatas[$attributeMetadata->getName()] = $attributeMetadata;
            }
            foreach ($attributeMetadatas as $attributeMetadata) {
                $modelMetadata->addAttribute($attributeMetadata);
            }

            $modelMetadatasTypeMap[$modelMetadata->getType()] = $modelMetadata;
        }

        foreach ($modelMetadatasTypeMap as $modelType => $modelMetadata) {
            $isCoreType = array_key_exists($modelType, $this->coreClassTypeMap);

            if ($isCoreType) {
                $this->context->setNode(
                    $modelMetadata,
                    $modelMetadatasTypeMap,
                    null,
                    sprintf('modelMetadatas[%s]', $modelMetadata->getUuid())
                );
                $this->context->buildViolation($constraint::CORETYPEVIOLATION)
                    ->setInvalidValue($modelType)
                    ->setParameter('{{ value }}', $modelType)
                    ->addViolation();
                continue;
            }

            $attributeMetadatas = $modelMetadata->getAttributes()->getValues();
            $attributeList = [];

            //Validate relationships and field name duplication
            foreach ($attributeMetadatas as $fieldName => $attributeMetadata) {
                if (is_numeric($attributeMetadata->getName())) {
                    $this->context->buildViolation($constraint::ATTRIBUTENAMEVIOLATION)
                        ->setInvalidValue($attributeMetadata->getName())
                        ->setParameter('{{ value }}', $attributeMetadata->getName())
                        ->setInvalidValue($modelType)
                        ->setParameter('{{ model }}', $modelType)
                        ->addViolation();
                    continue;
                }

                if (in_array(\strtolower($attributeMetadata->getName()), $attributeList) || in_array(\strtolower($this->inflector->pluralize($attributeMetadata->getName())), $attributeList)) {
                    $this->context->buildViolation($constraint::DUPLICATEFIELDVIOLATION)
                        ->atPath('name')
                        ->setInvalidValue($attributeMetadata->getName())
                        ->setParameter('{{ value }}', $attributeMetadata->getName())
                        ->setInvalidValue($modelType)
                        ->setParameter('{{ model }}', $modelType)
                        ->addViolation();
                } else {
                    array_push($attributeList, \strtolower($attributeMetadata->getName()));
                }

                $this->context->setNode($attributeMetadata, $attributeMetadatas, null, sprintf(
                    'modelMetadatas[%s].attributes[%s]',
                    $modelMetadata->getUuid(),
                    $attributeMetadata->getUuid()
                ));
                $attributeType = $attributeMetadata->getType();
                if (in_array($attributeType, $this->jsonPrimitives)) {
                    //Not a relationship
                    continue;
                }

                $isCoreType = array_key_exists($attributeType, $this->coreClassTypeMap);

                if (!$isCoreType && !array_key_exists($attributeType, $modelMetadatasTypeMap)) {
                    if (method_exists($modelMetadata, "gettype") && method_exists($attributeMetadata, "gettype")) {
                        $this->context->setNode($attributeMetadata, $attributeMetadatas, null, sprintf(
                            'modelMetadatas[%s].attributes[%s]',
                            $modelMetadata->gettype(),
                            $attributeMetadata->gettype()
                        ));
                    }
                    $this->context->buildViolation($constraint::ATTRIBUTETYPEVIOLATION)
                        ->atPath('type')
                        ->setInvalidValue($attributeType)
                        ->setParameter('{{ value }}', $attributeType)
                        ->addViolation();

                    $this->context->setNode($attributeMetadata, $attributeMetadatas, null, sprintf(
                        'modelMetadatas[%s].attributes[%s]',
                        $modelMetadata->getUuid(),
                        $attributeMetadata->getUuid()
                    ));
                    continue;
                }


                if ($isCoreType) {
                    //No further validation
                    continue;
                }

                $relatedModelMetadata = $modelMetadatasTypeMap[$attributeType];

                $inversedField = $attributeMetadata->getInversedField();
                if (empty($inversedField)) {
                    continue;
                }

                $sampleAttributeValues = [];
                foreach ($relatedModelMetadata->getAttributes()->getValues() as $attributeValue) {
                    $sampleAttributeValues[$attributeValue->getName()] = $attributeValue;
                }

                if (!$attributeMetadata->getOwnsRelationship() && !array_key_exists($inversedField, $sampleAttributeValues)) {
                    $this->context->buildViolation($constraint::INVERSEDFIELDVIOLATION)
                        ->atPath('inversedField')
                        ->setInvalidValue($inversedField)
                        ->setParameter('{{ value }}', $inversedField)
                        ->addViolation();
                }
            }
        }
    }

}
